(* NIST method for generating DSA primes: (from Handbook of Applied Crypto)
   INPUT: an integer l, 0 <= l <= 8.
   OUTPUT: a 160-bit prime q and an L-bit prime p, where L = 512 + 64l and
           q|(p-1).
   1. Compute L = 512 + 64l.  Using long division of (L - 1) by 160,
      find n, b such that L - 1 = 160n + b, where 0 <= b < 160.
   2. Repeat the following:
      2.1 Choose a random seed s (not necessarily secret) of bitlength
          g >= 160.
      2.2 Compute U = H(s) xor H((s+1) mod 2^g).
      2.3 Form q from U by setting to 1 the most significant and least
          significant bits of U.  (Note that q is a 160-bit odd integer.)
      2.4 Test q for primality using MILLER-RABIN(q,t) for t >= 18 (see
           Note 4.57).
      Until q is found to be a (probable) prime.
   3. Set i = 0, j = 2.
   4. While i < 4096 do the following:
      4.1 For k from 0 to n do the following:
   		set V[k] = H((s + j + k) mod 2^g).
      4.2 For the integer W defined below, let X = W + 2^(L-1).
          (X is an L-bit integer.)
   		W = V[0] + V[1]*2^160 + V[2]*2^320 + ...
   		    + V[n-1]*2^(160*(n-1)) + (V[n] mod 2^b)*2^(160n).
      4.3 Compute c = X mod 2q and set p = X - (c - 1).  (Note that
          p == 1 (mod 2q).)
      4.4 If p >= 2^(L-1) the do the following:
   		- Test p for primality using MILLER-RABIN(p,t) for t >= 5
   		  (See Note 4.75).
   		- If p is a (probable) prime then return(q,p).
      4.5 Set i = i + 1, j = j + n + 1.
   5. Go to step 2.

Note 4.57 (choice of primality test in [above algorithm])
   (i) The FIPS 186 document where [the algorithm] was originally described
       only specifies that a *robust* primality test be used in steps 2.4
       and 4.4, i.e., a primality test where the probablity of a composite
       integer being declared prime is at most (1/2)^80.  If the heuristic
       assumption is made that q is a randomly chosen 160-bit integer then,
       by Table 4.4, MILLER-RABIN(q,18) is a robust test for the primality
       of q.  If p is assumed to be a randomly chosen L-bit integer, then
       by Table 4.4, MILLER-RABIN(p,5) is a robust test for the primality
       of p.  Since the Miller-Rabin test is a probablistic primality
       test, the output of [the algorithm] is a probable prime.
  (ii) To improve performance, candidate primes q and p should be subjected
       to trial division by all odd primes less than some bound B before
       invoking the Miller-Rabin test.  See Note 4.45 for guidance on
       selecting B.

Note 4.45 (optimal trial division bound B)  Let E denote the time for a
full k-bit modular exponentiation, and let D denote the time required for
ruling out one small prime as divisor of a k-bit integer.  (The values E
and D depend on the particular implementation of long-integer arithmetic.)
Then the trial division bound B that minimizes the expected running time of
[a Miller-Rabin search] for generating a k-bit prime is roughly B = E/D.
A more accurate estimate of the optimum choice for B can be obtained
experimentally.  The odd primes up to B can be precomputed and stored in
a table.  If memory is scarce, a value of B that is smaller than the
optimum may be used.
========================================================================

   Miller-Rabin probabilistic primality test:  (same source)
   MILLER-RABIN(n,t)
   INPUT: an odd integer n >= 3 and security parameter t >= 1.
   OUTPUT: an answer "prime" or "composite" to the question: "Is n prime?"
   1. Write n - 1 = (2^s)*r where r is odd.
   2. For i from 1 to t do the following:
   	2.1 Choose a random integer a, 2 <= a <= n-2.
	2.2 Compute y = a^r mod n using Algorithm 2.143.
	2.3 If y <> 1 and y <> n - 1 then do the following
		j = 1
   		While j <= s - 1 and y <> n - 1 do the following:
   			Compute y = y^2 mod n
   			If y = 1 then return("composite").
			j = j + 1.
   		If y <> n then return("composite").
   3. Return("prime").

Algorithm 2.143 Repeated square-and-multiply algorithm for exponentiation
   in *Z*[n].
   INPUT: a in *Z*[n], and integer 0 <= k < n whose binary representation
          is k = \sum(t,i=0)(k[i]*2^i).
   OUTPUT: a^k mod n
   1. Set b = 1.  If k = 0 then return(b).
   2. Set A = a.
   3. If k[0] = 1 then set b = a.
   4. For i from 1 to t do the following:
	4.1 Set A = A^2 mod n.
	4.2 If k[i] = 1 then set b = A * b mod n. (???)
   5. Return(b).

===========================================================================

*)

open Safeloader
open Nat
open Sha1
open Cryptomath

let someprimes = Prime_list.someprimes

let g = 160
let two_pow_g =
  let n = make_nat 6 in
  set_digit_nat n 5 1;
  n
let waste_digit = create_nat 1
let bytes_per_digit = length_of_digit / 8
let digits_from_bytes i = (i + (bytes_per_digit - 1)) / bytes_per_digit
let sha1_digits = digits_from_bytes 20

let nat_to_bytestring n =
  let len_n = length_nat n in
  let res = String.make (len_n * bytes_per_digit) ' ' in
  let byte = create_nat 1 in
  for i = 0 to len_n - 1 do
    let dig_copy = copy_nat n i 1 in
    for j = bytes_per_digit downto 1 do
      shift_left_nat dig_copy 0 1 byte 0 8;
      res.[((len_n - i) * bytes_per_digit) - j] <- Char.chr (int_of_nat byte)
    done
  done;
  res

let bytestring_to_nat s =
  let slen = String.length s in
  let rlen = digits_from_bytes slen in
  let res = create_nat rlen in
  let tmp = create_nat 1 in
  for i = 0 to rlen - 2 do
    set_digit_nat res i
      (Char.code s.[slen - (bytes_per_digit * i) - bytes_per_digit]);
    for j = bytes_per_digit - 1 downto 1 do
      shift_left_nat res i 1 waste_digit 0 8;
      lor_digit_nat res i
	(nat_of_int (Char.code s.[slen - (bytes_per_digit * i) - j])) 0
    done
  done;
  let leftover =
    let rem = slen mod bytes_per_digit in
    if rem > 0 then rem else bytes_per_digit in
  set_digit_nat res (rlen - 1) (Char.code s.[0]);
  for i = 1 to leftover - 1 do
    shift_left_nat res (rlen - 1) 1 waste_digit 0 8;
    lor_digit_nat res (rlen - 1) (nat_of_int (Char.code s.[i])) 0
  done;
  res

let lxor_nat n nofs m mofs len =
  if len + nofs > (length_nat n) or len + mofs > (length_nat m)
  then (failwith "bounds check");
  for i = 0 to len - 1 do
    lxor_digit_nat n (nofs + i) m (mofs + i)
  done
(*  let rec aux noff moff l =
    if l = len then ()
    else begin
      lxor_digit_nat n noff m moff;
      aux (noff + 1) (moff + 1) (l + 1)
    end in
  aux nofs mofs 0*)

let nat_to_hex n =
  let len_n = length_nat n in
  let res = String.make (len_n * bytes_per_digit * 2) ' ' in
  let nybble = create_nat 1 in
  for i = 0 to len_n - 1 do
    let dig_copy = copy_nat n i 1 in
    for j = (bytes_per_digit * 2)  downto 1 do
      shift_left_nat dig_copy 0 1 nybble 0 4;
      let s = Printf.sprintf "%01X" (int_of_nat nybble) in
      res.[((len_n - i) * bytes_per_digit * 2) - j] <- s.[0]
    done
  done;
  res

let bytestring_to_hex s =
  let rec aux ofs len =
    if len = 0 then ""
    else
      (Printf.sprintf "%02X" (Char.code s.[ofs])) ^ (aux (ofs + 1) (len - 1))
  in
  aux 0 (String.length s)

let check_low_primes n =
  let len_n = num_digits_nat n 0 (length_nat n) in
  let len_tmp = 2 * len_n in
  let tmp_n = create_nat len_tmp in
  let rec aux l =
    match l with
      [] -> begin
	Log.log_msg ".";
	true
      end
    | h :: t ->
	blit_nat tmp_n 0 n 0 len_n;
	set_to_zero_nat tmp_n len_n len_n;
	div_nat tmp_n 0 len_tmp (nat_of_int h) 0 1;
	if is_zero_nat tmp_n 0 1 then false
	else aux t
  in
  aux someprimes

(* This routine generates a random value in the range lo (inclusive)
 * through hi (exclusive).
 *)
let rand_range lo hi digits =
  let len_diff = length_nat hi in
  if digits < len_diff then invalid_arg "rand_range";
  let diff = copy_nat hi 0 len_diff in
  let _ = sub_nat diff 0 len_diff lo 0 (length_nat lo) 1 in
  let tmp = Rand.rand_bytes (digits * bytes_per_digit) in
  let raw = create_nat (digits * 2) in
  blit_nat raw 0 tmp 0 digits;		(*leave space for division*)
  set_to_zero_nat raw digits digits;
  let len_diff = length_nat diff in
  div_nat raw 0 (2*digits) diff 0 len_diff;
  set_to_zero_nat raw len_diff ((digits * 2) - len_diff);
  let _ = add_nat raw 0 (2*digits) lo 0 (length_nat lo) 0 in
  copy_nat raw 0 (2*digits)
(*  let x = min len_diff digits in
  let len_res = (max (length_nat lo) x) + 1 in
  let res = create_nat len_res in
  blit_nat res 0 raw 0 x;
  set_to_zero_nat res x (len_res - x);
  let _ = add_nat res 0 len_res lo 0 (length_nat lo) 0 in
  res*)

(* n must be odd and greater than 2 *)
let rec miller_rabin n reps =
  let n_digits = num_digits_nat n 0 (length_nat n) in
  let pred_n = copy_nat n 0 n_digits in
  decr_nat pred_n 0 n_digits 0;
  let num_digits_pred_n = num_digits_nat pred_n 0 (length_nat pred_n) in
  let rec cleanup v ofs len =
    copy_nat v ofs (num_digits_nat v ofs len) in
  let find_twos n =			(*step 1*)
    let make_even v =
      let tmp = nat_of_int 1 in
      lxor_nat v 0 tmp 0 1 in
    let rec skip_digits v ofs =
      if is_digit_zero v ofs then skip_digits v (ofs + 1)
      else ofs in
    let rec make_odd v ofs len shifts =
      if is_digit_odd v ofs then (shifts, cleanup v ofs len)
      else begin
	shift_right_nat v ofs len waste_digit 0 1;
	make_odd v ofs len (shifts + 1)
      end in
    let tmp = copy_nat n 0 (length_nat n) in
    make_even tmp;
    let zero_digits = skip_digits tmp 0 in
    let (zero_bits, r) =
      make_odd tmp zero_digits ((length_nat n) - zero_digits) 0 in
    (zero_bits * length_of_digit + zero_bits, r) in
(*  let start = Unix.gettimeofday () in
  let stop_and_report t =
    let stop = Unix.gettimeofday () in
    let reps_completed = reps - t + 1 in
    let tim = stop -. start in
    let perrep = (float reps_completed) /. tim in
    Printf.printf "mr: %d reps in %f secs (%f reps/sec)\n"
      reps_completed tim perrep;
    flush stdout;
    false in
*)
  let is_prime init_y s t =		(*for step 2.3*)
    let rec aux j y =
      if (j <= s - 1)
	  & (not (eq_nat y 0 (length_nat y) pred_n 0 num_digits_pred_n))
      then
	let new_y = sq_mod_n y n n_digits in
	(not (eq_nat new_y 0 (length_nat new_y) (nat_of_int 1) 0 1))
	  & (aux (j + 1) new_y)
      else eq_nat y 0 (length_nat y) n 0 n_digits
    in
    let len_init_y = length_nat init_y in
    (eq_nat init_y 0 len_init_y (nat_of_int 1) 0 1)
      or (eq_nat init_y 0 len_init_y  pred_n 0 num_digits_pred_n)
      or (aux 1 init_y) (* or stop_and_report t*) in
  let (s, r) = find_twos n in
  let rec do_reps t =
    if t = 0 then begin (*stop_and_report t; *)true end
    else begin
      let a = rand_range (nat_of_int 2) pred_n n_digits in
      let y = sq_and_mult a r n n_digits in(* step 2.2*)
      (is_prime y s t) & (Log.log_msg "+"; true)
	& (do_reps (t - 1))
    end in
  do_reps reps

let rec prime_gen l =
  if l < 0 or l > 8 then invalid_arg "prime_gen";
  let big_l = 512 + 64 * l in		(*step 1*)
  let big_l_digits = big_l / length_of_digit in
  let n = (big_l - 1) / 160 in
  let b = (big_l - 1) mod 160 in
  let set_hi_and_lo u =
    if length_nat u <> digits_from_bytes 20 then (failwith "internal error");
    let v = nat_of_int 1 in
    let tmp = create_nat 1 in
    lor_digit_nat u 0 v 0;
    let dig = (g - 1) / length_of_digit in
    let ofs = (g - 1) mod length_of_digit in
    shift_left_nat v 0 1 tmp 0 ofs;
    lor_digit_nat u dig v 0 in
  let rec find_q () =			(*step 2*)
    let s_nat = Rand.rand_bytes (g/8) in(*step 2.1*)
    let s = nat_to_bytestring s_nat in
    let ctx = Sha1.sha1init () in	(*step 2.2*)
    Sha1.sha1update ctx s;
(*    let digest_s = String.make 20 ' ' in
    Sha1.sha1final digest_s ctx;*)
    let digest_s = Sha1.sha1final ctx in
    let s_1_nat = copy_nat s_nat 0 sha1_digits in
    incr_nat s_1_nat 0 sha1_digits 1;
    let s_1 = nat_to_bytestring s_1_nat in
    let ctx = Sha1.sha1init () in
    Sha1.sha1update ctx s_1;
(*    let digest_s_1 = String.make 20 ' ' in
    Sha1.sha1final digest_s_1 ctx;*)
    let digest_s_1 = Sha1.sha1final ctx in
    let q = bytestring_to_nat digest_s in
    lxor_nat q 0 (bytestring_to_nat digest_s_1) 0 sha1_digits;
    set_hi_and_lo q;			(*step 2.3*)
    if check_low_primes q & miller_rabin q 18 then begin(*step 2.4*)
	Log.log_msg "Q\n";
      (q, s_nat)
    end
    else find_q () in
  let len_double_q = (digits_from_bytes (g/8 + 1)) in
  let c = make_nat (big_l_digits + len_double_q) in
  let x = create_nat big_l_digits in
  let rec find_p q s i j double_q =(*step 4*)
    if i = 4096 then prime_gen l
    else begin
      let rec aux k =			(*step 4.1 & 4.2*)
	let g_digits = digits_from_bytes (g/8) in
	let j_plus_k_nat = (nat_of_int (j+k)) in
	let tmp = create_nat (g_digits + 1) in
	blit_nat tmp 0 s 0 g_digits;
	set_to_zero_nat tmp g_digits 1;
	let _ = add_nat tmp 0 g_digits j_plus_k_nat 0 1 0 in
	set_to_zero_nat tmp g_digits 1;	(*mod 2^g*)
	let tmp_str = nat_to_bytestring tmp in
	let ctx = Sha1.sha1init () in
	Sha1.sha1update ctx tmp_str;
(*	let v_k_str = String.make 20 ' ' in
	Sha1.sha1final v_k_str ctx;*)
	let v_k_str = Sha1.sha1final ctx in
	let v_k = bytestring_to_nat v_k_str in
	let offs = (160 * k) mod length_of_digit in
	shift_left_nat v_k 0 sha1_digits waste_digit 0 offs;
	let dig = (160 * k) / length_of_digit in
	if k < n then begin
	  if offs <> 0
	  then begin
	    lor_digit_nat x dig v_k 0;
	    blit_nat x (dig + 1) v_k 1 (digits_from_bytes ((160 - offs) / 8));
	  end else
	    blit_nat x dig v_k 0 sha1_digits;
	  aux (k + 1)
	end else begin
	  if b < offs then failwith "internal error: b < offs";
	  if offs <> 0
	  then begin
	    lor_digit_nat x dig v_k 0;
	    blit_nat x (dig + 1) v_k 1 (digits_from_bytes ((b - offs) / 8));
	  end else
	    blit_nat x dig v_k 0 (digits_from_bytes (b/8));
	  let l_1_pos = (big_l - 1) mod length_of_digit in
	  let tmp = make_nat 1 in
	  set_digit_nat tmp 0 1;
	  shift_left_nat tmp 0 1 waste_digit 0 l_1_pos;
	  lor_digit_nat x (big_l_digits - 1) tmp 0
	end in
      set_to_zero_nat x 0 big_l_digits;
      aux 0;
      blit_nat c 0 x 0 big_l_digits;
      div_nat c 0 (big_l_digits + len_double_q) double_q 0 len_double_q;
      set_to_zero_nat c len_double_q big_l_digits;
      let _ = decr_nat c 0 len_double_q 0 in
      let p = copy_nat x 0 big_l_digits in
      let _ = sub_nat p 0 big_l_digits c 0 len_double_q 1 in
      if is_digit_normalized p (big_l_digits - 1)
	  & check_low_primes p
	  & miller_rabin p 5
      then begin
	Log.log_msg "P\n";
	(q, p)
      end
      else begin
	find_p q s (i + 1) (j + n + 1) double_q
      end
    end in
  Log.log_msg "about to find_q\n";
  let (q, s) = find_q () in
  let double_q = make_nat len_double_q in
  blit_nat double_q 0 q 0 (num_digits_nat q 0 (length_nat q));
  shift_left_nat double_q 0 len_double_q waste_digit 0 1;
  Log.log_msg "about to find_p\n";
  find_p q s 0 2 double_q

let fixed_key_gen p q alpha =
  let p_len = length_nat p in
  let q_len = length_nat q in
  let a = rand_range (nat_of_int 1) q q_len in
  let y = sq_and_mult alpha a p p_len in
  ((p, q, alpha, y), a)
  
let key_gen l =
  if l < 0 or l > 8 then invalid_arg "key_gen";
  let big_l = 512 + 64 * l in
  let big_l_digits = big_l / length_of_digit in
  let (q, p) = prime_gen l in
  let p_len = length_nat p in
  let q_len = length_nat q in
  (*Select a generator, alpha, of the unique cyclic group of order q in Z*[p].*)
  let texp_len = 2 * big_l_digits in
  let texp = create_nat texp_len in
  blit_nat texp 0 p 0 p_len;
  set_to_zero_nat texp p_len (texp_len - p_len);
  let _ = decr_nat texp 0 texp_len 0 in
  div_nat texp 0 texp_len q 0 q_len;
  (*exponent for alpha is in texp[q_len..end]*)
  let exp = create_nat (texp_len - q_len) in
  blit_nat exp 0 texp q_len (texp_len - q_len);
  let rec aux () =
    let g = rand_range (nat_of_int 1) p p_len in
    let alpha = sq_and_mult g exp p p_len in
    if eq_nat alpha 0 (length_nat alpha) (nat_of_int 1) 0 1 then aux ()
    else alpha in
  let alpha = aux () in
  fixed_key_gen p q alpha

type nat_sign = { sign : int; value : nat }

let div_2_nat_sign a ofs len =
  shift_right_nat a.value ofs len waste_digit 0 1;
  if is_zero_nat a.value ofs len then {sign=0; value=a.value}
  else a

let sub_nat_sign_2 a aofs alen b boff blen =
(*if (length_nat a.value) < aofs + alen then failwith "sub_nat_sign_2 overrun a";
if (length_nat b.value) < boff + blen then failwith "sub_nat_sign_2 overrun b";*)
  if alen < blen then invalid_arg "sub_nat_sign_2";
  if b.sign = 0 then a
  else if a.sign = 0
  then begin
    blit_nat a.value aofs b.value boff blen;
    set_to_zero_nat a.value (aofs + blen) (alen - blen);
    { sign= -b.sign; value=a.value }
  end
  else if a.sign = b.sign
  then begin
    match compare_nat a.value aofs (num_digits_nat a.value aofs alen)
	b.value boff (num_digits_nat b.value boff blen) with
      0 -> set_to_zero_nat a.value aofs alen;
	{sign=0; value=a.value}
    | 1 -> 
	sub_nat a.value aofs alen b.value boff blen 1;
	a
    | -1 ->
	let b_digits = num_digits_nat b.value boff blen in
	let tmp = create_nat alen in
	blit_nat tmp 0 b.value boff b_digits;
	set_to_zero_nat tmp b_digits (alen - b_digits);
	sub_nat tmp 0 alen a.value aofs alen 1;
	blit_nat a.value aofs tmp 0 alen;
	{sign= -a.sign; value=a.value}
  end else begin
    add_nat a.value aofs alen b.value boff blen 0;
    a
  end

let sub_nat_sign a aofs alen n noff nlen =
  if alen < nlen then invalid_arg "sub_nat_sign";
  if is_zero_nat n noff nlen then a
  else sub_nat_sign_2 a aofs alen {sign=1; value=n} noff nlen

let add_nat_sign_2 a aofs alen b boff blen =
(*if (length_nat a.value) < aofs + alen then failwith "add_nat_sign_2 overrun a";
if (length_nat b.value) < boff + blen then failwith "add_nat_sign_2 overrun b";*)
  if alen < blen then invalid_arg "add_nat_sign_2";
  if b.sign = 0 then a
  else if a.sign = 0
  then begin
    blit_nat a.value aofs b.value boff blen;
    set_to_zero_nat a.value (aofs + blen) (alen - blen);
    {sign=b.sign; value=a.value}
  end
  else if a.sign = b.sign
  then begin
    add_nat a.value aofs alen b.value boff blen 0;
    a
  end else begin
    match compare_nat a.value aofs (num_digits_nat a.value aofs alen)
	b.value boff (num_digits_nat b.value boff blen) with
      0 -> set_to_zero_nat a.value aofs alen;
	{sign=0; value=a.value}
    | 1 ->
	sub_nat a.value aofs alen b.value boff blen 1;
	a
    | -1 ->
	let tmp = create_nat alen in
	blit_nat tmp 0 b.value boff blen;
	set_to_zero_nat tmp blen (alen - blen);
	sub_nat tmp 0 alen a.value aofs alen 1;
	blit_nat a.value aofs tmp 0 alen;
	{sign= -a.sign; value=a.value}
  end

let add_nat_sign a aofs alen n noff nlen =
if (length_nat a.value) < aofs + alen then failwith "add_nat_sign overrun a";
if (length_nat n) < noff + nlen then failwith "add_nat_sign overrun b";
  if is_zero_nat n noff nlen then a
  else add_nat_sign_2 a aofs alen {sign=1; value=n} noff nlen

let lt_nat_sign_2 a aofs alen b boff blen =
  if a.sign < b.sign then true
  else if a.sign > b.sign then false
  else begin (* a.sign == b.sign *)
    match a.sign with
      0 -> false
    | 1 -> lt_nat a.value aofs alen b.value boff blen
    | -1 -> gt_nat a.value aofs alen b.value boff blen
  end

let nat_sign_to_string a =
  Printf.sprintf "%s%s"
    (match a.sign with -1 -> "-" | 0 -> "Z" | 1->"")
    (string_of_nat a.value)

(* Note:  n *must* be a prime > 2 and n > a *)
let a_inv_mod_n a aoff alen n noff nlen =
  if (length_nat a) < aoff + alen then failwith "a_inv_mod_n overrun a";
  if (length_nat n) < noff + nlen then failwith "a_inv_mod_n overrun b";
  if not (is_digit_odd n noff) then invalid_arg "a_inv_mod_n";
  if is_digit_zero a aoff then invalid_arg "a_inv_mod_n";
  if gt_nat a aoff alen n noff nlen then invalid_arg "a_inv_mod_n";
  let u1_len = 2 * nlen in
  let tmp = make_nat u1_len in
  set_digit_nat tmp 0 1;
  let u1 = ref {sign=1; value=tmp} in
  let u2_len = 2 * nlen in
  let tmp = make_nat u2_len in
  let u2 = ref {sign=1; value=tmp} in
  let u3_len = 2 * nlen in
  let tmp = create_nat u3_len in
  blit_nat tmp 0 n noff nlen;
  set_to_zero_nat tmp nlen nlen;
  let u3 = ref {sign=1; value=tmp} in
  let t1_len = 2 * nlen in
  let tmp = create_nat t1_len in
  blit_nat tmp 0 a aoff alen;
  set_to_zero_nat tmp alen (t1_len - alen);
  let t1 = ref {sign=1; value=tmp} in
  let t2_len = 2 * nlen in
  let tmp = create_nat t2_len in
  blit_nat tmp 0 n noff nlen;
  set_to_zero_nat tmp alen (t2_len - alen);
  sub_nat tmp 0 t2_len (nat_of_int 1) 0 1 1;
  let t2 = ref {sign=1; value=tmp} in
  let t3_len = 2 * nlen in
  let tmp = create_nat t3_len in
  blit_nat tmp 0 a aoff alen;
  set_to_zero_nat tmp alen (t3_len - alen);
  let t3 = ref {sign=1; value=tmp} in
  while (!t3).sign > 0 do
    let rec aux () =
      if not (is_digit_odd (!u3).value 0)
      then begin
	if (is_digit_odd (!u1).value 0) or (is_digit_odd (!u2).value 0)
	then begin
	  u1 := add_nat_sign (!u1) 0 u1_len a aoff alen;
	  u2 := add_nat_sign (!u2) 0 u2_len n noff nlen;
	end;
	u1 := div_2_nat_sign (!u1) 0 u1_len;
	u2 := div_2_nat_sign (!u2) 0 u2_len;
	u3 := div_2_nat_sign (!u3) 0 u3_len
      end;
      if (not (is_digit_odd (!t3).value 0))
	  or (lt_nat_sign_2 (!u3) 0 u3_len (!t3) 0 t3_len)
      then begin
	let swap a b =
	  let tmp = (!a) in
	  a := (!b);
	  b := tmp in
	swap u1 t1;
	swap u2 t2;
	swap u3 t3
      end;
      if not (is_digit_odd (!u3).value 0) then aux() else ()
    in
    aux ();
    while (lt_nat_sign_2 (!u1) 0 u1_len (!t1) 0 t1_len)
	or (lt_nat_sign_2 (!u2) 0 u2_len (!t2) 0 t2_len) do
      u1 := add_nat_sign (!u1) 0 u1_len a aoff alen;
      u2 := add_nat_sign (!u2) 0 u1_len n noff nlen
    done;
    u1 := sub_nat_sign_2 (!u1) 0 u1_len (!t1) 0 t1_len;
    u2 := sub_nat_sign_2 (!u2) 0 u2_len (!t2) 0 t2_len;
    u3 := sub_nat_sign_2 (!u3) 0 u3_len (!t3) 0 t3_len
  done;
  while ((!u1).sign >= 0) & (ge_nat (!u1).value 0 u1_len a aoff alen)
    & ((!u2).sign >= 0) & (ge_nat (!u2).value 0 u2_len n noff nlen) do
    u1 := sub_nat_sign (!u1) 0 u1_len a aoff alen;
    u2 := sub_nat_sign (!u2) 0 u2_len n noff nlen
  done;
  if ((!u3).sign <> 1)
      or (not (eq_nat (!u3).value 0 u3_len (nat_of_int 1) 0 1))
  then failwith "found a factor"
  else
    let {sign=s; value=v} = sub_nat_sign (!u2) 0 u2_len n noff nlen in
    if s > 0 then failwith "internal error"
    else v

let sign m (p, q, alpha, y) a k_in =
  let a_digits = num_digits_nat a 0 (length_nat a) in
  let q_len = length_nat q in
  let p_len = length_nat p in
  (* Select a random secret integer k, 0 < k < q *)
  let k = if is_zero_nat k_in 0 (length_nat k_in)
  then rand_range (nat_of_int 1) q q_len
  else copy_nat k_in 0 (length_nat k_in)
  in
  (* Compute r = (alpha^k mod p)... *)
  let r = create_nat (2 * p_len) in
  let tmp = sq_and_mult alpha k p p_len in
  let tmp_len = length_nat tmp in
  blit_nat r 0 tmp 0 tmp_len;
  set_to_zero_nat r tmp_len ((2*p_len) - tmp_len);
  (* ... mod q *)
  div_nat r 0 (2 * p_len) q 0 q_len;
  set_to_zero_nat r q_len ((2*p_len) - q_len);
  (* Compute k^-1 mod q *)
  let k_inv = a_inv_mod_n k 0 q_len q 0 q_len in
  (* Compute k^-1{h(m) + ar} mod q *)
  let ctx = Sha1.sha1init () in
  Sha1.sha1update ctx m;
(*  let digest = String.make 20 ' ' in
  Sha1.sha1final digest ctx;*)
  let digest = Sha1.sha1final ctx in
  (* h(m) *)
  let hash_nat = bytestring_to_nat digest in
  let len_hash_nat = length_nat hash_nat in
  let alloc_ar = 3 * q_len in
  let ar = create_nat alloc_ar in
  blit_nat ar 0 hash_nat 0 len_hash_nat;
  set_to_zero_nat ar len_hash_nat (alloc_ar - len_hash_nat);
  (* h(m) + ar; mult_nat does a mult and add *)
  mult_nat ar 0 alloc_ar a 0 a_digits r 0 (length_nat r);
  let alloc_s = 4 * q_len in
  let s = make_nat alloc_s in
  (* k^-1{h(m) + ar} ... *)
  mult_nat s 0 alloc_s ar 0 (num_digits_nat ar 0 alloc_ar)
    k_inv 0 (num_digits_nat k_inv 0 (length_nat k_inv));
  (* mod q *)
  div_nat s 0 alloc_s q 0 q_len;
  (copy_nat r 0 (length_nat r), copy_nat s 0 q_len)

let verify m (r, s) (p, q, alpha, y) =
  let len_r = num_digits_nat r 0 (length_nat r) in
  let len_s = num_digits_nat s 0 (length_nat s) in
  let len_p = num_digits_nat p 0 (length_nat p) in
  let len_q = num_digits_nat q 0 (length_nat q) in
  let len_alpha = num_digits_nat alpha 0 (length_nat alpha) in
  let len_y = num_digits_nat y 0 (length_nat y) in
  (* Verify that 0 < r < q and 0 < s < q *)
  if is_zero_nat r 0 len_r
     or is_zero_nat s 0 len_s
     or ge_nat r 0 len_r q 0 len_q
     or ge_nat s 0 len_s q 0 len_q then false
  else begin
    (* Compute w = s^-1 mod q... *)
    let w = a_inv_mod_n s 0 len_s q 0 len_q in
    let len_w = num_digits_nat w 0 len_q in
    (* ... and h(m) *)
    let ctx = Sha1.sha1init () in
    Sha1.sha1update ctx m;
(*    let h_m_str = String.make 20 ' ' in
    Sha1.sha1final h_m_str ctx;*)
    let h_m_str = Sha1.sha1final ctx in
    let h_m = bytestring_to_nat h_m_str in
    (* Compute u_1 = w*h(m) mod q... *)
    let u_1_alloc = 3 * len_q in
    let u_1 = make_nat u_1_alloc in
    mult_nat u_1 0 u_1_alloc w 0 len_w
      h_m 0 (num_digits_nat h_m 0 (length_nat h_m));
    div_nat u_1 0 u_1_alloc q 0 len_q;
    set_to_zero_nat u_1 len_q (u_1_alloc - len_q);
    (* ... and u_2 = rw mod q *)
    let u_2_alloc = 3 * len_q in
    let u_2 = make_nat u_2_alloc in
    mult_nat u_2 0 u_2_alloc r 0 len_r w 0 len_w;
    div_nat u_2 0 u_2_alloc q 0 len_q;
    set_to_zero_nat u_2 len_q (u_2_alloc - len_q);
    (* Compute v = (alpha^u_1 * y^u_2 mod p)... *)
    let t1 = sq_and_mult alpha u_1 p len_p in
    let t2 = sq_and_mult y u_2 p len_p in
    let v_alloc = 3 * len_p in
    let v = make_nat v_alloc in
    mult_nat v 0 v_alloc t1 0 len_p t2 0 len_p;
    div_nat v 0 v_alloc p 0 len_p;
    set_to_zero_nat v len_p (v_alloc - len_p);
    (* ...mod q *)
    div_nat v 0 v_alloc q 0 len_q;
    set_to_zero_nat v len_q (v_alloc - len_q);
    eq_nat v 0 v_alloc r 0 len_r
  end
