open Safestd
open Safeunix
open Unixnet
open Safethread
open Log

let ether_type_DOD_INTERNET = 0x800
let ether_type_ADDRESS_RESOLUTION = 0x806

let ares_op_REQUEST = 1
let ares_op_REPLY = 2

let ares_hrd_Ethernet = 1

let entry_timeout = 60			(*1 minute*)

let die = ref false
let die_mutex = Mutex.create ()
let die_condition = Condition.create ()

type eaddr = int * int
type ip_addr = int * int * int * int
type arp_t = { dst : eaddr; src : eaddr; typ : int; ar_hrd : int;
	       ar_pro : int; ar_hln : int; ar_pln : int; ar_op : int;
	       ar_sha : eaddr; ar_spa : ip_addr; ar_tha : eaddr;
	       ar_tpa : ip_addr }

let value = ref None

let hashval() =
  match !value with
    None -> failwith "hash value not set!  (This can't happen.)"
  | Some(v) -> v

let arp_hashtable = Hashtbl.create 257
let arp_hash_m = Mutex.create ()

let ether_broadcast = (0xffffff,0xffffff)
let ether_hln = 6

let ip_pln = 6

(* Caution: this routine returns the address in a byte order that allows
   intercomparison but which would be different from machine to machine *)
let get_eaddr s ofs =
  let a = Char.code (String.get s ofs) in
  let b = Char.code (String.get s (ofs + 1)) in
  let c = Char.code (String.get s (ofs + 2)) in
  let d = Char.code (String.get s (ofs + 3)) in
  let e = Char.code (String.get s (ofs + 4)) in
  let f = Char.code (String.get s (ofs + 5)) in
  ((a lsl 16) lor (b lsl 8) lor c, (d lsl 16) lor (e lsl 8) lor f)

let get_int8 s ofs = Char.code (String.get s ofs)
let get_int16 s ofs = (Char.code (String.get s ofs) lsl 8)
  lor (Char.code (String.get s (ofs + 1)))
let get_ip_addr s ofs = (Char.code (String.get s ofs),
			 Char.code (String.get s (ofs + 1)),
			 Char.code (String.get s (ofs + 2)),
			 Char.code (String.get s (ofs + 3)))

let get_src_eaddr s = get_eaddr s 6
let get_dst_eaddr s = get_eaddr s 0
let get_typ s = get_int16 s 12
let get_ar_hrd s = get_int16 s 14
let get_ar_pro s = get_int16 s 16
let get_ar_hln s = get_int8 s 18
let get_ar_pln s = get_int8 s 19
let get_ar_op s = get_int16 s 20
let get_ar_sha s = get_eaddr s 22
let get_ar_spa s = get_ip_addr s 28
let get_ar_tha s = get_eaddr s 32
let get_ar_tpa s = get_ip_addr s 38

let get_arp_pkt s =
  { dst = get_dst_eaddr s;
    src = get_src_eaddr s;
    typ = get_typ s;
    ar_hrd = get_ar_hrd s;
    ar_pro = get_ar_pro s;
    ar_hln = get_ar_hln s;
    ar_pln = get_ar_pln s;
    ar_op = get_ar_op s;
    ar_sha = get_ar_sha s;
    ar_spa = get_ar_spa s;
    ar_tha = get_ar_tha s;
    ar_tpa = get_ar_tpa s }

let make_eaddr addr =
  let (upper, lower) = addr in
  Printf.sprintf "%c%c%c%c%c%c"
  (Char.chr (upper lsr 16)) (Char.chr ((upper lsr 8) land 0xff))
  (Char.chr (upper land 0xff)) (Char.chr (lower lsr 16))
  (Char.chr ((lower lsr 8) land 0xff)) (Char.chr (lower land 0xff))

let make_ip_addr addr =
  let (a, b, c, d) = addr in
  Printf.sprintf "%c%c%c%c" (Char.chr a) (Char.chr b) (Char.chr c) (Char.chr d)

let make_int8 v = String.make 1 (Char.chr v)
let make_int16 v =
  Printf.sprintf "%c%c" (Char.chr (v lsr 8)) (Char.chr (v land 0xff))
let make_int32 v =
  Printf.sprintf "%c%c%c%c"
  (Char.chr (v lsr 24)) (Char.chr ((v lsr 16) land 0xff))
  (Char.chr ((v lsr 8) land 0xff)) (Char.chr (v land 0xff))

let make_typ s = make_int16 s
let make_ar_hrd s = make_int16 s
let make_ar_pro s = make_int16 s
let make_ar_hln s = make_int8 s
let make_ar_pln s = make_int8 s
let make_ar_op s = make_int16 s
let make_ar_sha s = make_eaddr s
let make_ar_spa s = make_ip_addr s
let make_ar_tha s = make_eaddr s
let make_ar_tpa s = make_ip_addr s

let make_arp_str pkt =
  let str = String.create 42 in
  String.blit (make_eaddr pkt.dst) 0 str 0 6;
  String.blit (make_eaddr pkt.src) 0 str 6 6;
  String.blit (make_typ pkt.typ) 0 str 12 2;
  String.blit (make_ar_hrd pkt.ar_hrd) 0 str 14 2;
  String.blit (make_ar_pro pkt.ar_pro) 0 str 16 2;
  String.blit (make_ar_hln pkt.ar_hln) 0 str 18 1;
  String.blit (make_ar_pln pkt.ar_pln) 0 str 19 1;
  String.blit (make_ar_op pkt.ar_op) 0 str 20 2;
  String.blit (make_ar_sha pkt.ar_sha) 0 str 22 6;
  String.blit (make_ar_spa pkt.ar_spa) 0 str 28 4;
  String.blit (make_ar_tha pkt.ar_tha) 0 str 32 6;
  String.blit (make_ar_tpa pkt.ar_tpa) 0 str 38 4;
  str

let rec conv_to_ipaddrs lst =
  match lst with
    [] -> []
  | (d,a)::t -> (d, get_ip_addr a 0)::(conv_to_ipaddrs t)

let rec conv_to_eaddrs lst =
  match lst with
    [] -> []
  | (d,a)::t ->
      (d, get_eaddr a 0)::(conv_to_eaddrs t)

let get_arps (iport, hashtable, hash_m, start_m) =
  let rec match_addr addr lst =
    match lst with
      [] -> false
    | (device,thisaddr)::t ->
	if thisaddr = addr
    	then true
    	else match_addr addr t in
  let ip_to_eaddr ip iplst elst =
    let rec ip_to_dev ip iplst =
      match iplst with
	[] -> log_msg "fatal error: cannot find device for ip addr\n"; raise Exit
      |	(dev,addr)::t ->
	  if addr = ip then dev else ip_to_dev ip t in
    let rec dev_to_eaddr dev elst =
      match elst with
	[] -> log_msg "fatal error: cannot find eaddr for device\n"; raise Exit
      |	(device,addr)::t ->
	  let (a,b) = addr in
	  if device = dev then addr else dev_to_eaddr dev t in
    dev_to_eaddr (ip_to_dev ip iplst) elst in
  Mutex.lock start_m;
  Mutex.unlock start_m;
  let myeaddrs = conv_to_eaddrs (status_in iport GETHWADDR) in
  let myipaddrs = conv_to_ipaddrs (status_in iport GETIPADDR) in
  let oport = iport_to_oport iport in
  while not !die do
    let pkt = get_next_pkt_in iport in
    let arp = get_arp_pkt pkt.pkt in
    let curtime = time () in
    if arp.typ = ether_type_ADDRESS_RESOLUTION then begin
      Mutex.lock hash_m;
      Hashtbl.remove hashtable (arp.ar_spa, arp.ar_pro);
      Hashtbl.add hashtable (arp.ar_spa, arp.ar_pro) (arp.ar_sha, curtime);
      Mutex.unlock hash_m;
      if arp.ar_op = ares_op_REQUEST then begin
	if not (match_addr arp.ar_tpa myipaddrs)
	then (log_msg "not for me\n"; call_next_function pkt)
	else begin
	  if arp.ar_pro = ether_type_DOD_INTERNET then begin
	    match pkt.addr with
	      ADDR_PKT(dev_type, dev, proto) ->
	      	let addr_out = ADDR_PKT(dev_type, dev, proto) in
		let thiseaddr = ip_to_eaddr arp.ar_tpa myipaddrs myeaddrs in
		let pkt_out =
		  { dst=arp.src;
		    src=thiseaddr;
		    typ=arp.typ;
		    ar_hrd=arp.ar_hrd;
		    ar_pro=arp.ar_pro;
		    ar_hln=arp.ar_hln;
		    ar_pln=arp.ar_pln;
		    ar_op=ares_op_REPLY;
		    ar_sha=thiseaddr;
		    ar_spa=arp.ar_tpa;
		    ar_tha=arp.ar_sha;
		    ar_tpa=arp.ar_spa
		  } in
	      	let outpkt = make_arp_str pkt_out in
	      	send_pkt_out oport outpkt 0 (String.length outpkt) addr_out;
	    	let (dst1, dst2) = pkt_out.dst in
	    	let (src1, src2) = pkt_out.src in
	    	let (sha1, sha2) = pkt_out.ar_sha in
	    	let (spa1, spa2, spa3, spa4) = pkt_out.ar_spa in
	    	let (tha1, tha2) = pkt_out.ar_tha in
	    	let (tpa1, tpa2, tpa3, tpa4) = pkt_out.ar_tpa in
	    	let s = Printf.sprintf
 		  "SEND: %06x.%06x <- %06x.%06x typ: %04x space %d pro: %04x\n\thlen %d plen %d %06x.%06x (%d.%d.%d.%d) -> %06x.%06x (%d.%d.%d.%d)\n"
		  dst1 dst2 src1 src2 arp.typ arp.ar_hrd arp.ar_pro arp.ar_hln
		  arp.ar_pln sha1 sha2 spa1 spa2 spa3 spa4
		  tha1 tha2 tpa1 tpa2 tpa3 tpa4 in
	    	log_msg s; ()
	    | _ -> log_msg "fatal error; pkt.addr != ADDR_PKT\n"; raise Exit
	  end else (log_msg "not my proto\n"; ())
	end;
	let (dst1, dst2) = arp.dst in
	let (src1, src2) = arp.src in
	let (sha1, sha2) = arp.ar_sha in
	let (spa1, spa2, spa3, spa4) = arp.ar_spa in
	let (tha1, tha2) = arp.ar_tha in
	let (tpa1, tpa2, tpa3, tpa4) = arp.ar_tpa in
	let s = Printf.sprintf
 	  "REQUEST: %06x.%06x <- %06x.%06x typ: %04x space %d pro: %04x\n\thlen %d plen %d %06x.%06x (%d.%d.%d.%d) -> %06x.%06x (%d.%d.%d.%d)\n"
	  dst1 dst2 src1 src2 arp.typ arp.ar_hrd arp.ar_pro arp.ar_hln
	  arp.ar_pln sha1 sha2 spa1 spa2 spa3 spa4
	  tha1 tha2 tpa1 tpa2 tpa3 tpa4 in
	log_msg s; () end
      else begin
	if match_addr arp.dst myeaddrs
 	then ()
 	else call_next_function pkt;
	let (dst1, dst2) = arp.dst in
	let (src1, src2) = arp.src in
	let (sha1, sha2) = arp.ar_sha in
	let (spa1, spa2, spa3, spa4) = arp.ar_spa in
	let (tha1, tha2) = arp.ar_tha in
	let (tpa1, tpa2, tpa3, tpa4) = arp.ar_tpa in
	let s = Printf.sprintf
 	  "REPLY: %06x.%06x <- %06x.%06x typ: %04x space %d pro: %04x\n\thlen %d plen %d %06x.%06x (%d.%d.%d.%d) -> %06x.%06x (%d.%d.%d.%d)\n"
	  dst1 dst2 src1 src2 arp.typ arp.ar_hrd arp.ar_pro arp.ar_hln
	  arp.ar_pln sha1 sha2 spa1 spa2 spa3 spa4
	  tha1 tha2 tpa1 tpa2 tpa3 tpa4 in
	log_msg s; () end
    end else call_next_function pkt
  done

let timeout_thread (hashtable, hash_m, start_m) =
  Mutex.lock start_m;
  Mutex.unlock start_m;
  while not !die do
    let curtime = time () in
    let check_time key value =
      let (_, entry_time) = value in
      if entry_time - curtime > entry_timeout then (*timeout*)
      	Hashtbl.remove hashtable key in
    Mutex.lock hash_m;
    Hashtbl.iter check_time hashtable;
    Mutex.unlock hash_m;
    Safethread.delay 60.0
  done

let debug_thread (hashtable, hash_m, start_m) =
  Mutex.lock start_m;
  Mutex.unlock start_m;
  let dmp_tbl key value =
    let (spa, pro) = key in
    let (sha, tim) = value in
    let (spa1, spa2, spa3, spa4) = spa in
    let (sha1, sha2) = sha in
    let s = Printf.sprintf "%d.%d.%d.%d (%04x) -> %06x.%06x (time: %d)\n"
    spa1 spa2 spa3 spa4 pro sha1 sha2 tim in
    log_msg s in
  while not !die do
    Mutex.lock hash_m;
    Hashtbl.iter dmp_tbl hashtable;
    Mutex.unlock hash_m;
    Safethread.delay 23.0
  done

let top_thread () =
  let start_m = Mutex.create () in
  let rec start_handlers () =
    let iport = bind_type ether_type_ADDRESS_RESOLUTION in
    [(Safethread.create (hashval())
      get_arps (iport, arp_hashtable, arp_hash_m, start_m));
      (Safethread.create (hashval ())
       timeout_thread (arp_hashtable, arp_hash_m, start_m));
      (Safethread.create (hashval ())
       debug_thread (arp_hashtable, arp_hash_m, start_m))]
  in
  let rec kill_all_threads pid_list =
    match pid_list with
      [] -> ()
    | h :: t -> Safethread.kill h; kill_all_threads t in
  let thread_list = Mutex.lock start_m; start_handlers () in
  Mutex.unlock start_m;
  Mutex.lock die_mutex;
  while not !die do
    Condition.wait die_condition die_mutex
  done;
  kill_all_threads thread_list

let arp ignored =
  let t = Safethread.create (hashval()) top_thread () in
  Safethread.join t;
  ""

let _ = value := Some(Func.register "arp" arp)

(*
This code currently has the following problems:
 1) it has no access to a iport to do the status_in calls
 2) it will (slowly) infinite loop if the remote host does
    not exist

let send_arp myip hisip s ofs =
  let rec ip_to_dev addr lst =
    match lst with
      [] -> raise Exit
    | (d,a)::t -> if a = addr then d else ip_to_dev addr t in
  let rec dev_to_eaddr dev lst =
    match lst with
      [] -> raise Exit
    | (d,a)::t -> if d = dev then a else dev_to_eaddr dev t in
  let myeaddrs = conv_to_eaddrs (status_in iport GETHWADDR) in
  let myipaddrs = conv_to_ipaddrs (status_in iport GETIPADDR) in
  let dev = ip_to_dev myip myipaddrs in
  let eaddr = dev_to_eaddr dev myeaddrs in
  let addr_out =
    ADDR_PKT(ares_hrd_Ethernet, dev, ether_type_ADDRESS_RESOLUTION) in
  let pkt_out =
    { dst=ether_broadcast;
      src=eaddr;
      typ=ether_type_ADDRESS_RESOLUTION;
      ar_hrd=ares_hrd_Ethernet;
      ar_pro=ether_type_DOD_INTERNET;
      ar_hln=ether_hln;
      ar_pln=ip_pln;
      ar_op=ares_op_REQUEST;
      ar_sha=eaddr;
      ar_spa=myip;
      ar_tha=ether_broadcast;
      ar_tpa=hisip
    } in
  let outpkt = mark_arp_str pkt_out in
  let oport = bind_out dev in
  send_pkt_out oport outpkt 0 (String.length output) addr_out;
  unbind_out oport;
  Thread.delay 5.0;
  ip_to_ether myip hisip s ofs

and ip_to_ether myip hisip s ofs =
  Mutex.lock arp_hash_m;
  try
    let (ether, _) = Hashtbl.find arp_hashtable hisip in
    let addr = make_eaddr ether in
    String.blit addr 0 s ofs 6
  with Not_found -> send_arp myip hisip s ofs

*)
