(* Helper functions to be used with the spanning tree. Included are functions
   to print bpdus, turn packets into bpdus, *)
open Safeloader

type protocol_version = int
type bpdu_type = int
type flag = bool
type timer_func = unit -> unit

 type entry = { dev : string; mutable state : int} 

type identifier = { mutable priority1 : int;
		    mutable priority2 : int;
		    mutable addr1: int;
		    mutable addr2: int;
		    mutable addr3: int;
		    mutable addr4: int; 
		    mutable addr5: int;
		    mutable addr6: int 
		  }

type cost = int
type port_id = { mutable priority: int;
		 mutable port_no: int
	       } 
type time = int

(* Configuration BPDU Parameters *)
type config_bpdu = { mutable ctype : bpdu_type;
		     mutable croot_id : identifier;
		     mutable croot_path_cost : cost;
		     mutable cbridge_id : identifier;
		     mutable cport_id : port_id;
		     mutable cmessage_age : time;
		     mutable cmax_age : time;
		     mutable chello_time : time;
		     mutable cforward_delay : time;
		     mutable ctopology_change_acknowledgement : flag;
		     mutable ctopology_change : flag
		   } 

(* Topology Change Notification BPDU Parameters *)
type tcn_bpdu = { mutable tcn_type : bpdu_type }

(* Bridge Parameters *)
type bridge_data = { mutable btype : bpdu_type;
                     mutable bdesignated_root : identifier;
		     mutable broot_path_cost : cost;
		     mutable broot_port : int;
		     mutable bmax_age : time;
		     mutable bhello_time : time;
		     mutable bforward_delay : time;
		     mutable bbridge_type : bpdu_type;
		     mutable bbridge_id : identifier;
		     mutable bbridge_max_age : time;
		     mutable bbridge_hello_time : time;
		     mutable bbridge_forward_delay : time;
		     mutable btopology_change_detected : bool;
		     mutable btopology_change : bool;
		     mutable btopology_change_time : time;
		     mutable bhold_time : time
		   } 

(* Port Parameters *)
type port_data = { mutable pport_id : port_id;
		   mutable pstate : int;
		   mutable ppath_cost : int;
		   mutable pdesignated_root : identifier;
		   mutable pdesignated_cost : int;
		   mutable pdesignated_bridge : identifier;
		   mutable pdesignated_port : port_id;
		   mutable ptopology_change_acknowledge : bool;
		   mutable pconfig_pending : bool
		 } 

(* types to support timers *)
type timer = { mutable active : bool;
	       mutable valu : float;
	       mutable started : time;
	       mutable func : timer_func
	     } 


(* *********************************************************************** *)

(* *********************************************************************** *)

let form_it s ofs n =
  if(n==1) then 
    Char.code (String.get s ofs)
  else if(n==2) then begin
    let t1 = Char.code (String.get s ofs) in
    let t2 = Char.code (String.get s (ofs+1)) in
    (t1 * 0x100) + t2
  end
  else if(n==4) then begin
    let t1 = Char.code (String.get s ofs) in
    let t2 = Char.code (String.get s (ofs+1)) in
    let t3 = Char.code (String.get s (ofs+2)) in
    let t4 = Char.code (String.get s (ofs+3)) in
    (t1 * 0x1000000) + (t2 * 0x10000) + (t3 * 0x100) + t4
  end
  else 0


let format_identifier s ofs =
  { priority1 = Char.code (String.get s ofs);
    priority2 = Char.code (String.get s (ofs+1));
    addr1 = Char.code (String.get s (ofs+2));
    addr2 = Char.code (String.get s (ofs+3));
    addr3 = Char.code (String.get s (ofs+4));
    addr4 = Char.code (String.get s (ofs+5));
    addr5 = Char.code (String.get s (ofs+6));
    addr6 = Char.code (String.get s (ofs+7)) }


let format_port_id s ofs =
  { priority = Char.code (String.get s ofs);
    port_no = Char.code (String.get s (ofs+1)) } 

let format_dec_port_id s ofs =
  { priority=0;
    port_no = Char.code (String.get s ofs) }


let format_flags s ofs =
  let tmp = Char.code (String.get s ofs) in
  ((tmp land 0x80) != 0, (tmp land 1) != 0)

let format_ieee_config_bpdu s =
  let (tca, tc) = format_flags s 21 in
  { ctype = form_it s 20 1;
    ctopology_change_acknowledgement = tca;
    ctopology_change = tc;
    croot_id = format_identifier s 22;
    croot_path_cost = form_it s 30 4;
    cbridge_id = format_identifier s 34;
    cport_id = format_port_id s 42;
    cmessage_age = form_it s 44 2;
    cmax_age = form_it s 46 2;
    chello_time = form_it s 48 2;
    cforward_delay = form_it s 50 2 }

let form_dec_it s ofs =
  let t1 = Char.code (String.get s ofs) in
  (t1*0x100)

let format_dec_flags s ofs =
  let tmp = Char.code (String.get s ofs) in
  ((tmp land 0x02) != 0, (tmp land 0x01) != 0)

let format_dec_config_bpdu s =
  let (tca, tc) = format_dec_flags s 17 in
  { ctype = form_it s 15 1;
    ctopology_change_acknowledgement = tca;
    ctopology_change = tc;
    croot_id = format_identifier s 18;
    croot_path_cost = form_it s 26 2;
    cbridge_id = format_identifier s 28;
    cport_id = format_dec_port_id s 36;
    cmessage_age = form_dec_it s 37;
    cmax_age = form_dec_it s 38;
    chello_time = form_dec_it s 39;
    cforward_delay = form_dec_it s 40 }

let format_config_bpdu s =
  let ityp = form_it s 20 1 in
  let dtyp = form_it s 15 1 in
  if((ityp == 0)&&(dtyp!=0x19)) then
    format_ieee_config_bpdu s
  else
    format_dec_config_bpdu s
    
    
let format_tcn_bpdu s =
  let ityp = form_it s 20 1 in
  let dtyp = form_it s 15 1 in
  if((ityp==128)&&(dtyp!=0x02)) then
    { tcn_type = ityp }
  else
     { tcn_type = dtyp }

(* *********************************************************************** *)
let print_identifier b =
  Printf.sprintf "0x %02x:%02x-%02x:%02x:%02x:%02x:%02x:%02x"
  b.priority1 b.priority2 b.addr1 b.addr2 b.addr3 b.addr4 b.addr5 b.addr6 

let print_address b =
  Printf.sprintf "%02x:%02x:%02x:%02x:%02x:%02x " 
  b.addr1 b.addr2 b.addr3 b.addr4 b.addr5 b.addr6

let print_port_id b =
  Printf.sprintf "{priority: 0x%2x port_no: 0x%2x}" b.priority b.port_no

let print_config b =
  let tca =
    if b.ctopology_change_acknowledgement then "true" else "false" in
  let tc =
    if b.ctopology_change then "true" else "false" in
  if (b.ctype == 0) then
    let l1 =
      Printf.sprintf
      "CONFIG BPDU: Type: 0x%02x\n\tRoot ID: %s Root Path Cost: 0x%02x\n"
      b.ctype (print_identifier b.croot_id) b.croot_path_cost in
    let l2 = Printf.sprintf "\tBridge ID: %s  Port ID: %s\n"
      (print_identifier b.cbridge_id) (print_port_id b.cport_id) in
    let l3 = Printf.sprintf
      "\tMsg Age: 0x%02x Max Age: 0x%02x Hello Time: 0x%02x Forward Delay: 0x%02x\n"
      b.cmessage_age b.cmax_age b.chello_time b.cforward_delay in
    let l4 = Printf.sprintf
      "\tTopology Change Acknowledgement: %s  Topology Change %s\n\n"
      tca tc in
    l1 ^ l2 ^ l3 ^ l4
  else 
    let l1 = Printf.sprintf
      "CONFIG BPDU: Type: 0x%02x\n\tRoot ID: %s Root Path Cost: 0x%02x\n"
(*      b.ctype (print_identifier b.croot_id) (b.croot_path_cost / 0x100) in*)
      b.ctype (print_identifier b.croot_id) b.croot_path_cost in
    let l2 = Printf.sprintf "\tBridge ID: %s  Port ID: 0x%02x\n"
      (print_identifier b.cbridge_id) (b.cport_id.port_no) in
    let l3 = Printf.sprintf
      "\tMsg Age: 0x%02x Max Age: 0x%02x Hello Time: 0x%02x Forward Delay: 0x%02x\n"
      (b.cmessage_age/0x100) (b.cmax_age/0x100) (b.chello_time/0x100)
      (b.cforward_delay/0x100) in
    let l4 = Printf.sprintf
      "\tTopology Change Acknowledgement: %s  Topology Change %s\n\n"
      tca tc in
    l1 ^ l2 ^ l3 ^ l4

let print_tcn b =
  Printf.sprintf "TCN BPDU: Type: 0x%2x\n\n" b.tcn_type
  
(* *********************************************************************** *)

let identifier_to_string b =
  let d0 = Char.chr (b.priority1) in
  let d1 = Char.chr (b.priority2) in
  let d2 = Char.chr (b.addr1) in
  let d3 = Char.chr (b.addr2) in
  let d4 = Char.chr (b.addr3) in
  let d5 = Char.chr (b.addr4) in
  let d6 = Char.chr (b.addr5) in
  let d7 = Char.chr (b.addr6) in
  Printf.sprintf "%c%c%c%c%c%c%c%c" d0 d1 d2 d3 d4 d5 d6 d7

let address_to_string b =
  let d2 = Char.chr (b.addr1) in
  let d3 = Char.chr (b.addr2) in
  let d4 = Char.chr (b.addr3) in
  let d5 = Char.chr (b.addr4) in
  let d6 = Char.chr (b.addr5) in
  let d7 = Char.chr (b.addr6) in
  Printf.sprintf "%c%c%c%c%c%c" d2 d3 d4 d5 d6 d7

let port_id_to_string b =
  let p = Char.chr b.priority in
  let pn = Char.chr b.port_no in
  Printf.sprintf "%c%c" p pn


let to_string b =
  Printf.sprintf "%c" (Char.chr b)

let int_to_string b i =
  if (i==0) then begin
    let s0 = Char.chr (b / 256) in
    let s1 = Char.chr (b mod 256) in
    Printf.sprintf "%c%c" s0 s1
  end
  else begin
    let s0 = Char.chr (b / 0x1000000) in
    let a0 = b mod 0x1000000 in
    let s1 = Char.chr (a0 / 0x10000) in
    let a1 = a0 mod 0x10000 in
    let s2 = Char.chr (a1 / 0x100) in
    let a2 = a1 mod 0x100 in
    let s3 = Char.chr a2 in
    Printf.sprintf "%c%c%c%c" s0 s1 s2 s3
  end

let rec create_pad n =
  if (n!=0) then 
    Printf.sprintf "%c%s" (Char.chr 0) (create_pad (n-1))
  else ""
