(*
 * control program:
 *  binds to the IEEE all bridges address and waits for a BDPU
 *  pause the DEC STA and grab its iport
 *  start the IEEE STA and give it the ieee_iport
 *  throw away DEC STA packets until ?
 *  If DEC STA packet arrives, revert to DEC
 *     and if IEEE has an error, revert to DEC
 *)

open Safestd
open Safeunix
open Unixnet
open Safethread
open Log
open Bob

(* This is the magic cookie that will let us get more threads *)
let value = ref None

(* Wrap the cookie in a function that does the correct checking *)
let hashval() =
  match !value with
    None -> failwith "hash value not set!  (This can't happen.)"
  | Some(v) -> v

(* The 802.1D address for reaching all bridges *)
let ieee_all_bridges = "\001\128\194\000\000\000"

let top_thread () =
  log_msg "control:top_thread started\n";
  (* Grab an 802.1D packets *)
  let ieee_iport = bind_addr ieee_all_bridges in

  (* This function checks to see the the SAP values, protocol ID,
     version, and message type correspond to an 802.1D Hello BPDU *)
  let validate_ieee_pkt pkt =
    (Char.code (String.get pkt.pkt 14) = 0x42)
    & (Char.code (String.get pkt.pkt 15) = 0x42)
    & (Char.code (String.get pkt.pkt 17) = 0)
    & (Char.code (String.get pkt.pkt 18) = 0)
    & (Char.code (String.get pkt.pkt 19) = 0)
    & (Char.code (String.get pkt.pkt 20) = 0) in

  (* This function tail recurses until it finds a valid 802.1D Hello BPDU *)
  let rec find_ieee_pkt iport =
    let pkt = get_next_pkt_in iport in
    if validate_ieee_pkt pkt then begin
      log_msg "control:find_ieee_pkt done\n"; () end
    else begin log_msg "control:find_ieee_pkt recursing\n";
      find_ieee_pkt iport end in

  (* This function throws out any DEC packets that we receive *)
  let rec supress_dec_pkts iport tim =
    try
      let delta = tim -. (gettimeofday()) in
      if delta > 0.0 then begin
      	timed_get_next_pkt_in iport delta;
	log_msg "control:top_thread suppress\n";
      	supress_dec_pkts iport tim
      end else
	let s = Printf.sprintf
	  "control: suppression ends at %f\n" (gettimeofday())
    	in log_msg s; ()
    with Unix_error(ETIMEDOUT,_,_) ->
      let s = Printf.sprintf
	"control: suppression ends at %f\n" (gettimeofday())
      in log_msg s; () in

  (* This function turns a Bob.identifier into a printable string *)
  let identifer_to_string id =
    Printf.sprintf "%02x%02x:%02x.%02x.%02x.%02x.%02x.%02x"
    id.priority1 id.priority2 id.addr1 id.addr2
    id.addr3 id.addr4
    id.addr5 id.addr6 in

  (* And similarly for a Bob.port_id *)
  let port_id_to_string pid =
    Printf.sprintf "%04x:%04x" pid.priority pid.port_no in

  (* And a Bob.time *)
  let time_to_string tim =
    Printf.sprintf "%d %d/256" (tim lsr 8) (tim land 0xff) in

  (* This function dumps out a Bob.bridge_data to the log file *)
  let dump_bridge_data bd =
    let s = Printf.sprintf "type: %x root: %s cost: %d\n" bd.btype
      (identifer_to_string bd.bdesignated_root) bd.broot_path_cost in
    let t = Printf.sprintf "port id: %d max age: %s hello time: %s fwd delay: %s\n"
      bd.broot_port (time_to_string bd.bmax_age)
      (time_to_string bd.bhello_time)
      (time_to_string bd.bforward_delay) in
    let u = Printf.sprintf "br type: %x br id: %s br max age: %s br hello: %s\n"
      bd.bbridge_type (identifer_to_string bd.bbridge_id)
      (time_to_string bd.bbridge_max_age)
      (time_to_string bd.bbridge_hello_time) in
    let v = Printf.sprintf "br fwd dly: %s tcd: %s tc: %s tct: %s hold time: %s\n"
      (time_to_string bd.bbridge_forward_delay)
      (if bd.btopology_change_detected then "true" else "false")
      (if bd.btopology_change then "true" else "false")
      (time_to_string bd.btopology_change_time)
      (time_to_string bd.bhold_time) in
    log_msg s; log_msg t; log_msg u; log_msg v in

  (* And this one dumps a Bob.port_data *)
  let dump_port_data pd =
    let s = Printf.sprintf "id: %s state: %d path cost: %d des root: %s\n"
      (port_id_to_string pd.pport_id) pd.pstate pd.ppath_cost
      (identifer_to_string pd.pdesignated_root) in
    let t = Printf.sprintf
      "des cost: %d des br: %s des pt: %s tca: %s pend: %s\n"
      pd.pdesignated_cost (identifer_to_string pd.pdesignated_bridge)
      (port_id_to_string pd.pdesignated_port)
      (if pd.ptopology_change_acknowledge then "true" else "false")
      (if pd.pconfig_pending then "true" else "false") in
    log_msg (s ^ t) in

  (* This routine compares the final state of the DEC protocol with
   * the current state of the IEEE protocol to make sure that the
   * IEEE protocol is functioning properly.
   *)
  let check_state dbi dpi ibi ipi =
    (* compare two ports;  disabled or blocking always match *)
    let compare_pids dpid ipid =
      if (ipid.pstate = 0 or ipid.pstate = 4) then
	dpid.pstate = ipid.pstate
& log_msg "simple OK\n" > 0
      else
      	dpid.pstate = ipid.pstate
& log_msg "state OK\n" >= 0
      	& dpid.ppath_cost = ipid.ppath_cost
& log_msg "path cost OK\n" >= 0
      	& dpid.pdesignated_root = ipid.pdesignated_root
& log_msg "des root OK\n" >= 0
      	& dpid.pdesignated_cost = ipid.pdesignated_cost
& log_msg "des cost OK\n" >= 0
      	& dpid.pdesignated_bridge = ipid.pdesignated_bridge
& log_msg "des br OK\n" >= 0
      	& dpid.pdesignated_port = ipid.pdesignated_port 
& log_msg "des port OK\n" >= 0 in

    (* compare port arrays *)
    let check_all_ports dports iports =
      let cmp a dp ip =	a & (compare_pids dp ip) in
      Array.length dports = Array.length iports
      & List.fold_left2
      cmp true (Array.to_list dports) (Array.to_list iports) in

    (* selected bridge info and port info must match *)
    dbi.bdesignated_root = ibi.bdesignated_root
& log_msg "roots ok\n" >= 0
    & dbi.broot_path_cost = ibi.broot_path_cost
& log_msg "root paths ok\n" >= 0
    & dbi.broot_port = ibi.broot_port
& log_msg "root ports ok\n" >= 0
    & check_all_ports dpi ipi in

  let revert iport =
    Cieee.do_suspend (); Timer.destroy (); Cieee.do_kill ();
    unbind_in ieee_iport;
    Cdec.dec_do_start iport;
    let s = Printf.sprintf "control: reverted and finished at %f\n"
      (gettimeofday ()) in
    log_msg s;
    () in

  let revert_if_dec_pkt iport =
    let _ = get_next_pkt_in iport in
    log_msg "control:top_thread revert\n";
    revert iport in

  log_msg "control:top_thread got ieee_iport\n";
  find_ieee_pkt ieee_iport;
  let s = Printf.sprintf
    "control:top_thread got ieee pkt at %f\n" (gettimeofday ()) in
  log_msg s;
  Cdec.do_suspend ();
  log_msg "control:top_thread dec suspended\n";
  Timer.destroy ();

  (* Grab the DEC information so that we can check on the
     802.1D implemenation later. *)
  let dec_port_info = Cdspan.port_info in
  let dec_bridge_info = Cdspan.bridge_info in
  dump_bridge_data dec_bridge_info;
  Array.iter dump_port_data dec_port_info;
  Cieee.hostname := !Cdec.hostname;
  match !Cdec.dec_iport with
    None -> failwith "no DEC iport"
  | Some(dec_iport) ->
      let s = Printf.sprintf "control:top_thread about to start ieee at %f\n"
 	(gettimeofday ()) in
      log_msg s;
      Cieee.ieee_do_start ieee_iport;
      log_msg "control:top_thread ieee started; about to kill dec\n";
      Cdec.do_kill ();
      log_msg "control:top_thread ding dong, the dec is dead\n";
log_msg (Printf.sprintf "supression delay: %f\n"
((float dec_bridge_info.bforward_delay) /. 64.0));
      supress_dec_pkts dec_iport
      (* forward_delay is in units of 1/256 s.  IEEE starts forwarding
       * packets after 2*forward_delay.  To be safe, we double that
       * before checking for a reasonable spanning tree.  Thus,
       * we want 4*(forward_delay/.256.) + current_time.  We do
       * the cancellation to get the expression below.
       *)
      ((float dec_bridge_info.bforward_delay) /. 64.0 +. (gettimeofday()));
      let ieee_port_info = Cispan.port_info in
      let ieee_bridge_info = Cispan.bridge_info in
      dump_bridge_data ieee_bridge_info;
      Array.iter dump_port_data ieee_port_info;
      if check_state dec_bridge_info dec_port_info
      	ieee_bridge_info ieee_port_info then
      	revert_if_dec_pkt dec_iport
      else begin
	let s = Printf.sprintf "control: check_state failed at %f\n"
	  (gettimeofday())
	in log_msg s;
	revert dec_iport
      end
	

let control ignored =
  let s = Printf.sprintf "control: started at %f\n" (gettimeofday ()) in
  log_msg s;
  let t = Safethread.create (hashval()) top_thread () in
  Safethread.join t;
  ""

let _ = value := Some(Func.register "control" control)
