open Safestd
open Safeunix
open Unixnet
open Safethread
open Log
open Local

type ip_port = IP_PORT of int * local_pkt Queue.t * Mutex.t * Condition.t

type ip_addr = int * int * int * int

type ip_hdr =
    { version : int;			(*4 bits*)
      ihl : int;			(*4 bits*)
      tos : int;			(*8 bits*)
      length : int;			(*16 bits*)
      id : int;				(*8 bits*)
      flags : int;			(*3 bits*)
      frag_offset : int;		(*13 bits*)
      ttl : int;			(*8 bits*)
      protocol : int;			(*8 bits*)
      hdr_chksum : int;			(*16 bits*)
      source_addr : ip_addr;		(*32 bits*)
      dest_addr : ip_addr		(*32 bits*)
    }

type ip_pkt = { ip_hdr : ip_hdr; ip_data : string; ip_ehdr : Local.ehdr;
		ip_uaddr : sockaddr }

exception Already_bound

let value = ref None
let hashval() =
  match !value with
    None -> failwith "hash value not set!  (This can't happen.)"
  | Some(v) -> v

let die = ref false
let die_mutex = Mutex.create ()
let die_condition = Condition.create ()

let ip_hashtable = Hashtbl.create 257
let ip_hash_m = Mutex.create ()

let lport = ref None

let default_ttl = ref 255

let prev_ip_ident = ref 1
let ip_ident_mutex = Mutex.create ()

let get_next_ident () =
  let ident =
    Mutex.lock ip_ident_mutex;
    incr prev_ip_ident;
    !prev_ip_ident in
  Mutex.unlock ip_ident_mutex;
  ident
  
let get_int8 s ofs = Char.code (String.get s ofs)
let get_int16 s ofs = (Char.code (String.get s ofs) lsl 8)
  lor (Char.code (String.get s (ofs + 1)))
let get_ip_addr s ofs = (Char.code (String.get s ofs),
			 Char.code (String.get s (ofs + 1)),
			 Char.code (String.get s (ofs + 2)),
			 Char.code (String.get s (ofs + 3)))

  (*Ethernet header takes bytes 0-13*)
let get_version s = (get_int8 s 0) lsr 4
let get_ihl s = (get_int8 s 0) land 0xf
let get_tos s = get_int8 s 1
let get_length s = get_int16 s 2
let get_id s = get_int8 s 4
let get_flags s = (get_int8 s 6) lsr 5
let get_frag_offset s = (get_int16 s 6) land 0x1fff
let get_ttl s = get_int8 s 8
let get_protocol s = get_int8 s 9
let get_hdr_chksum s = get_int16 s 10
let get_source_addr s = get_ip_addr s 12
let get_dest_addr s = get_ip_addr s 16

let to_ip_hdr s =
    { version = get_version s;
      ihl = get_ihl s;
      tos = get_tos s;
      length = get_length s;
      id = get_id s;
      flags = get_flags s;
      frag_offset = get_frag_offset s;
      ttl = get_ttl s;
      protocol = get_protocol s;
      hdr_chksum = get_hdr_chksum s;
      source_addr = get_source_addr s;
      dest_addr = get_dest_addr s
    }

let from_ip_hdr hdr =
  let (srca, srcb, srcc, srcd) = hdr.source_addr in
  let (dsta, dstb, dstc, dstd) = hdr.dest_addr in
  Printf.sprintf "%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c"
    (Char.chr (((hdr.version land 0xf) lsl 4) lor (hdr.ihl land 0xf)))
    (Char.chr hdr.tos)
    (Char.chr (hdr.length lsr 8))
    (Char.chr (hdr.length land 0xff))
    (Char.chr (hdr.id lsr 8))
    (Char.chr (hdr.id land 0xff))
    (Char.chr ((hdr.flags lsl 4) land ((hdr.frag_offset lsr 8) land 0xf)))
    (Char.chr (hdr.frag_offset land 0xff))
    (Char.chr hdr.ttl)
    (Char.chr hdr.protocol)
    (Char.chr (hdr.hdr_chksum lsr 8))
    (Char.chr (hdr.hdr_chksum land 0xff))
    (Char.chr srca)
    (Char.chr srcb)
    (Char.chr srcc)
    (Char.chr srcd)
    (Char.chr dsta)
    (Char.chr dstb)
    (Char.chr dstc)
    (Char.chr dstd)

let from_ip_hdr_and_csum hdr =
  let (srca, srcb, srcc, srcd) = hdr.source_addr in
  let (dsta, dstb, dstc, dstd) = hdr.dest_addr in
  let s = Printf.sprintf "%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c"
      (Char.chr (((hdr.version land 0xf) lsl 4) lor (hdr.ihl land 0xf)))
      (Char.chr hdr.tos)
      (Char.chr (hdr.length lsr 8))
      (Char.chr (hdr.length land 0xff))
      (Char.chr (hdr.id lsr 8))
      (Char.chr (hdr.id land 0xff))
      (Char.chr ((hdr.flags lsl 4) land (hdr.frag_offset land 0xf)))
      (Char.chr (hdr.frag_offset land 0xff))
      (Char.chr hdr.ttl)
      (Char.chr hdr.protocol)
      '\000'
      '\000'
      (Char.chr srca)
      (Char.chr srcb)
      (Char.chr srcc)
      (Char.chr srcd)
      (Char.chr dsta)
      (Char.chr dstb)
      (Char.chr dstc)
      (Char.chr dstd) in
  let csum = Ipcsum.ipcsum s 0 (String.length s) in
  String.set s 10 (Char.chr (csum lsr 8));
  String.set s 11 (Char.chr (csum land 0xff));
  s

let build_reply_ip_hdr hdr len =
  { version = 4;
    ihl = 5;
    tos = hdr.tos;
    length = len;
    id = get_next_ident ();
    flags = 0;
    frag_offset = 0;
    ttl  =  !default_ttl;
    protocol = hdr.protocol;
    hdr_chksum = 0;			(*XXX*)
    source_addr = hdr.dest_addr;
    dest_addr = hdr.source_addr
  }

let local_to_ip lpkt =
  let ip_hdr = to_ip_hdr lpkt.data in
  let hdr_bytes = ip_hdr.ihl * 4 in
  { ip_hdr = ip_hdr;
    ip_data =
    String.sub lpkt.data hdr_bytes (String.length lpkt.data - hdr_bytes);
    ip_ehdr = lpkt.ehdr;
    ip_uaddr = lpkt.uaddr
  }

let ip_to_local ip_pkt =
  { ehdr=ip_pkt.ip_ehdr;
    data=
    Printf.sprintf "%s%s" (from_ip_hdr_and_csum ip_pkt.ip_hdr) ip_pkt.ip_data;
    uaddr=ip_pkt.ip_uaddr
  }

let ip_to_local_reply ip_pkt =
  let tot_len = 20 + (String.length ip_pkt.ip_data) in
  { ehdr=ip_pkt.ip_ehdr;
    data =
    Printf.sprintf "%s%s"
    (from_ip_hdr_and_csum
       (build_reply_ip_hdr ip_pkt.ip_hdr tot_len)) ip_pkt.ip_data;
    uaddr = ip_pkt.ip_uaddr
  } 

let handle_ip (lport, start_m) =
  let checksum hdr =
    Ipcsum.ipcsum hdr 0 20 = 0 in
  try
    Mutex.lock start_m;
    Mutex.unlock start_m;

    while not !die do
      let lpkt = get_packet lport in
      let hdr = to_ip_hdr lpkt.data in
      let elen = String.length lpkt.data in
      if hdr.version = 4
	  & hdr.ihl = 5
	  & ((elen = hdr.length) or (elen = 46))
	  & hdr.frag_offset = 0
	  & checksum lpkt.data
      then begin
      	Mutex.lock ip_hash_m;
      	try
      	  let (q, m, c) = Hashtbl.find ip_hashtable hdr.protocol in
      	  Mutex.unlock ip_hash_m;
      	  Mutex.lock m;
      	  Queue.add lpkt q;
      	  Condition.signal c;
      	  Mutex.unlock m
      	with Not_found -> Mutex.unlock ip_hash_m  (*discard packet*)
      end else
      	let s = Printf.sprintf
	    "Discarded IP packet: ver %d ihl %d elen %d iplen %d offset %d csum %04x (%04x)\n"
	    hdr.version hdr.ihl (String.length lpkt.data)
 	    hdr.length hdr.frag_offset hdr.hdr_chksum (Ipcsum.ipcsum lpkt.data 0 20)
	in
      	log_msg s; ()
    done
  with Unix_error(e, s1, s2) ->
    Printf.printf "ip: handle_ip: unix error: %s '%s'; '%s'\n" (error_message e) s1 s2
  | e -> Printf.printf "ip: handle_ip: unknown exception\n"

let bind_ip_proto proto =
  try
    Mutex.lock ip_hash_m;
    Hashtbl.find ip_hashtable proto;
    Mutex.unlock ip_hash_m;
    raise Already_bound
  with Not_found -> begin
    let q = Queue.create () in
    let m = Mutex.create () in
    let c = Condition.create () in
    Hashtbl.add ip_hashtable proto (q, m, c);
    Mutex.unlock ip_hash_m;
    IP_PORT(proto, q, m, c)
  end
  | e -> begin
      Mutex.unlock ip_hash_m;
      raise e
  end

let unbind_ip_proto ipport =
  let IP_PORT(proto, _, _, _) = ipport in
  Mutex.lock ip_hash_m;
  Hashtbl.remove ip_hashtable proto;
  Mutex.unlock ip_hash_m

let get_packet ipport =
  let IP_PORT(_, q, m, c) = ipport in
  Mutex.lock m;
  while (Queue.length q = 0) do
    Condition.wait c m
  done;
  let pkt = Queue.take q in
  Mutex.unlock m;
  local_to_ip pkt

let send_packet ipport pkt ofs len addr =
  match !lport with
    None -> raise Exit
  | Some(prt) -> Local.send_packet prt pkt ofs len addr

let send_ip ipport ip_pkt =
  match !lport with
    None -> raise Exit
  | Some(prt) ->
      let lpkt = ip_to_local ip_pkt in
      Local.send_local_pkt prt lpkt

let reply_ip ipport ip_pkt =
  match !lport with
    None -> raise Exit
  | Some(prt) ->
      let lpkt = ip_to_local_reply ip_pkt in
      Local.reply_local_pkt prt lpkt

let top_thread () =
  try
    let int16_to_string x = try
      Printf.sprintf "%c%c" (Char.chr (x lsr 8)) (Char.chr (x land 0xff))
    	with Invalid_argument(_) ->
	  begin Printf.printf "int16_to_string screwup: ia\n"; raise Exit end
    in
    let start_m = Mutex.create () in
    Mutex.lock start_m;
    let loc_port = bind_local_type (int16_to_string 0x800) in
    lport := Some(loc_port);
    let tid = Safethread.create (hashval()) handle_ip (loc_port, start_m) in
    Mutex.unlock start_m;
    Mutex.lock die_mutex;
    while not !die do
      Condition.wait die_condition die_mutex
    done;
    Safethread.kill tid;
    unbind_local_type loc_port
  with Unix_error(e, s1, s2) ->
    Printf.printf "ip: unix error: %s '%s'; '%s'\n" (error_message e) s1 s2
  | e -> Printf.printf "ip: unknown exception\n"


let ip ignored =
  let t = Safethread.create (hashval()) top_thread () in
  Safethread.join t;
  ""

let _ = value := Some(Func.register "ip" ip)
