open Safestd
open Safeunix
open Unixnet
open Safethread
open Log

type local_port = LOCAL_PORT of string * packet Queue.t * Mutex.t * Condition.t

type eaddr = int * int
type ehdr = { dst : eaddr; src : eaddr; proto : int }
type local_pkt = { ehdr : ehdr; data : string; uaddr : sockaddr }

exception Already_bound

let value = ref None
let hashval() =
  match !value with
    None -> failwith "hash value not set!  (This can't happen.)"
  | Some(v) -> v

let die = ref false
let die_mutex = Mutex.create ()
let die_condition = Condition.create ()

let oport = ref None

let local_hashtable = Hashtbl.create 257
let local_hash_m = Mutex.create ()

let get_int16 s ofs = (Char.code (String.get s ofs) lsl 8)
  lor (Char.code (String.get s (ofs + 1)))
let get_eaddr s ofs = ((Char.code (String.get s ofs) lsl 16)
		       lor (Char.code (String.get s (ofs + 1)) lsl 8)
		       lor (Char.code (String.get s (ofs + 2))),
		       (Char.code (String.get s (ofs + 3)) lsl 16)
		       lor (Char.code (String.get s (ofs + 4)) lsl 8)
		       lor (Char.code (String.get s (ofs + 5))))

let pkt_to_local (pkt : Unixnet.packet) =
  { ehdr={
      dst=get_eaddr pkt.pkt 0;
      src=get_eaddr pkt.pkt 6;
      proto=get_int16 pkt.pkt 12
    };
    data=String.sub pkt.pkt 14 (pkt.len - 14);
    uaddr=pkt.addr
  }

let local_to_pkt local_pkt =
  try
    let (hi_dst, low_dst) = local_pkt.ehdr.dst in
    let (hi_src, low_src) = local_pkt.ehdr.src in
    let data = Printf.sprintf "%c%c%c%c%c%c%c%c%c%c%c%c%c%c%s"
    	(Char.chr (hi_dst lsr 16))
    	(Char.chr ((hi_dst lsr 8) land 0xff))
    	(Char.chr (hi_dst land 0xff))
    	(Char.chr (low_dst lsr 16))
    	(Char.chr ((low_dst lsr 8) land 0xff))
    	(Char.chr (low_dst land 0xff))
    	(Char.chr (hi_src lsr 16))
    	(Char.chr ((hi_src lsr 8) land 0xff))
    	(Char.chr (hi_src land 0xff))
    	(Char.chr (low_src lsr 16))
    	(Char.chr ((low_src lsr 8) land 0xff))
    	(Char.chr (low_src land 0xff))
    	(Char.chr (local_pkt.ehdr.proto lsr 8))
    	(Char.chr (local_pkt.ehdr.proto land 0xff))
    	local_pkt.data in
    { len=String.length data;
      addr=local_pkt.uaddr;
      pkt=data;
      next_fn=no_next
    }
  with Invalid_argument(_) ->
    begin Printf.printf "local_to_pkt screwup: ia\n"; raise Exit end

let handle_if (iport, start_m) =
  let get_type s = String.sub s 12 2 in
  Mutex.lock start_m;
  Mutex.unlock start_m;

  while not !die do
    let pkt = get_next_pkt_in iport in
    let typ = get_type pkt.pkt in
    Mutex.lock local_hash_m;
    try
      let (q, m, c) = Hashtbl.find local_hashtable typ in
      Mutex.unlock local_hash_m;
      Mutex.lock m;
      Queue.add pkt q;
      Condition.signal c;
      Mutex.unlock m
    with Not_found -> begin
      Mutex.unlock local_hash_m;
      call_next_function pkt
    end
  done

let bind_local_type typ =
  try
    Mutex.lock local_hash_m;
    Hashtbl.find local_hashtable typ;
    Mutex.unlock local_hash_m;
    raise Already_bound
  with Not_found -> begin
    let q = Queue.create () in
    let m = Mutex.create () in
    let c = Condition.create () in
    Hashtbl.add local_hashtable typ (q, m, c);
    Mutex.unlock local_hash_m;
    LOCAL_PORT(typ, q, m, c)
  end
  | e -> begin
      Mutex.unlock local_hash_m;
      raise e
  end

let unbind_local_type lport =
  let LOCAL_PORT(typ, _, _, _) = lport in
  Mutex.lock local_hash_m;
  Hashtbl.remove local_hashtable typ;
  Mutex.unlock local_hash_m

let get_packet lport =
  let LOCAL_PORT(_, q, m, c) = lport in
  Mutex.lock m;
  while (Queue.length q = 0) do
    Condition.wait c m
  done;
  let pkt = Queue.take q in
  Mutex.unlock m;
  pkt_to_local pkt

let send_packet lport pkt ofs len addr =
  match !oport with
    None -> log_msg "send_packet: !oport = None\n"; raise Exit
  | Some(prt) -> send_pkt_out prt pkt ofs len addr

let send_local_pkt lport lpkt =
  match !oport with
    None -> log_msg "send_local_pkt: !oport = None\n"; raise Exit
  | Some(prt) ->
      let pkt = local_to_pkt lpkt in
      send_pkt_out prt pkt.pkt 0 (String.length pkt.pkt) pkt.addr

let reply_local_pkt lport lpkt =
  match !oport with
    None -> log_msg "reply_local_pkt: !oport = None\n"; raise Exit
  | Some(prt) ->
      let lpkt_out = {
	ehdr={dst=lpkt.ehdr.src; src=lpkt.ehdr.dst; proto=lpkt.ehdr.proto};
	data=lpkt.data;
	uaddr=lpkt.uaddr } in
      let pkt = local_to_pkt lpkt_out in
      send_pkt_out prt pkt.pkt 0 (String.length pkt.pkt) pkt.addr

let top_thread () =
  let rec get_all_ifs () =
    try
      let iport = bind_any_local_addr () in
      iport :: get_all_ifs ()
    with No_interface -> [] in
  let start_m = Mutex.create () in
  let rec start_handlers iport_list =
    match iport_list with
      [] -> []
    | h :: t ->
	(Safethread.create (hashval()) handle_if (h, start_m))
	:: start_handlers t in
  let rec kill_all_threads pid_list =
    match pid_list with
      [] -> ()
    | h :: t -> Safethread.kill h; kill_all_threads t in
  let rec clean_up iport_list =
    match iport_list with
      [] -> []
    | h :: t ->
	let oport = iport_to_oport h in
	unbind_out oport; unbind_in h; clean_up t in
  let port_list = get_all_ifs () in
  oport := Some(iport_to_oport (List.hd port_list));
  let thread_list = Mutex.lock start_m; start_handlers port_list in
  Mutex.unlock start_m;
  Mutex.lock die_mutex;
  while not !die do
    Condition.wait die_condition die_mutex
  done;
  kill_all_threads thread_list;
  clean_up port_list

let local ignored =
  let t = Safethread.create (hashval()) top_thread () in
  Safethread.join t;
  ""

let _ = value := Some(Func.register "local" local)
