open Safestd
open Safeunix
open Unixnet
open Safethread
open Log
open Loudbob

(* States *)
let disabled = 0
let listening = 1
let learning = 2
let forwarding = 3
let blocking = 4

type eaddr = int * int
(* type entry = { dev : string; mutable state : int} *)

let value = ref None

let hashval() =
  match !value with
    None -> failwith "hash value not set!  (This can't happen.)"
  | Some(v) -> v

let delimiter =
  "\239\190\173\222\206\250\173\222\239\190\173\222\206\250\173\222"
let multicast_bit = 0x010000		(*bit from high word for multicast*)
let die = ref false
let die_mutex = Mutex.create ()
let die_condition = Condition.create ()
let all_outs = ref []
let all_devs = ref []
let hash_timeout = 300			(* 5 minutes *)

(* Should search through the list l and return the element corresponding to s*)
let finddev s l =
  let tmp=ref {dev=""; state=0} in
  let rec inner l =
    (match l with 
      [] -> ()
    | h::t -> 
      	if (h.dev == s) then tmp :=h
      	else
	  inner t) in
  inner l;
  !tmp

(* Caution: this routine returns the address in a byte order that allows
   intercomparison but which would be different from machine to machine *)
let get_eaddr s ofs =
  let a = Char.code (String.get s ofs) in
  let b = Char.code (String.get s (ofs + 1)) in
  let c = Char.code (String.get s (ofs + 2)) in
  let d = Char.code (String.get s (ofs + 3)) in
  let e = Char.code (String.get s (ofs + 4)) in
  let f = Char.code (String.get s (ofs + 5)) in
  ((a lsl 16) lor (b lsl 8) lor c, (d lsl 16) lor (e lsl 8) lor f)

let get_src_eaddr s = get_eaddr s 6
let get_dst_eaddr s = get_eaddr s 0

let format_saddr_pkt saddr =
  match saddr with
    ADDR_UNIX s -> "family AF_UNIX path '" ^ s ^ "'"
  | ADDR_INET (a, port) -> "family AF_INET addr " ^ string_of_inet_addr a ^ " port "
      ^ string_of_int port
  | ADDR_PKT (fam, dev, prot) -> "family PACKET dev type " ^ string_of_int fam ^
      " device " ^ dev ^ " proto " ^ string_of_int prot

let handle_if (iport, hashtable, h_mutex, start_m, out_triple) =
  let map_aux (q, m, c) pkt =
    Mutex.lock m;
    Queue.add pkt q;
    Condition.signal c;
    Mutex.unlock m in
  let rec no_map_aux input_port pkt port_list dev_list =
    match port_list with
      [] -> ()
    | h :: t ->
	if h <> input_port then begin
	  let tmp = List.hd dev_list in
	  if(tmp.state == forwarding) then
	    map_aux h pkt;
	end;
	no_map_aux input_port pkt t (List.tl dev_list) in
  let s =
    Printf.sprintf "handle_if started (%s)\n" (debug_iport_to_string iport) in
  log_msg s;
  Mutex.lock start_m;
  Mutex.unlock start_m;

  let this_port = finddev (debug_iport_to_string iport) !all_devs in

  while not !die do
    let pkt = get_next_pkt_in iport in
    let dst = get_dst_eaddr pkt.pkt in
    let src = get_src_eaddr pkt.pkt in
    let curtime = time () in
    let (dst1, dst2) = dst in
    let (src1, src2) = src in
    let s =
(*  Printf.sprintf "handle_if: %06x.%06x -> %06x.%06x\n" src1 src2 dst1 dst2*)
     Printf.sprintf "handle_if %s(%d): %06x.%06x -> %06x.%06x\n" (debug_iport_to_string iport) this_port.state src1 src2 dst1 dst2
    in
    log_msg s;

    if (this_port.state == forwarding) then begin

      Mutex.lock h_mutex;
      if (src1 land multicast_bit) = 0 then begin
      	Hashtbl.remove hashtable src;
      	Hashtbl.add hashtable src (out_triple, curtime)
      end;
      if (dst1 land multicast_bit) <> 0 then begin
      	Mutex.unlock h_mutex;
      	log_msg "\tmulticast\n";
      	no_map_aux out_triple pkt !all_outs !all_devs
      end
      else begin
      	try
	  let (out, tbltime) = Hashtbl.find hashtable dst in
	  let diff_time = curtime - tbltime in
	  if diff_time > hash_timeout then begin
	    Hashtbl.remove hashtable dst;
	    Mutex.unlock h_mutex;
	    log_msg "\ttimeout\n";
	    no_map_aux out_triple pkt !all_outs !all_devs
	  end
	  else begin
	    Mutex.unlock h_mutex;
	    if out <> out_triple then begin
	      log_msg "\tmapped\n";
	      map_aux out pkt
	    end
	    else begin
	      log_msg "\tmapped; no send\n";
	      ()
	    end
	  end
      	with Not_found ->
	  (Mutex.unlock h_mutex;
	   log_msg "\tno map\n";
	   no_map_aux out_triple pkt !all_outs !all_devs)
      end
    end;
  done;
  let s =
    Printf.sprintf "handle_if stopped (%s)\n" (debug_iport_to_string iport) in
  log_msg s

let handle_out (oport, q, m, c, start_m) =
  Mutex.lock start_m;
  Mutex.unlock start_m;
  let s =
    Printf.sprintf "handle_out started (%s)\n" (debug_oport_to_string oport) in
  log_msg s;
  let oport_dev = debug_oport_to_string oport in
  while not !die do
    Mutex.lock m;
    while Queue.length q = 0 do
      Condition.wait c m
    done;
    let pkt = Queue.take q in
    let (dst1, dst2) = get_dst_eaddr pkt.pkt in
    let (src1, src2) = get_src_eaddr pkt.pkt in
    let s = Printf.sprintf
      "handle_out: oport %s len %d (%s: %06x.%06x -> %06x.%06x)\n"
      oport_dev pkt.len (format_saddr_pkt pkt.addr)
      src1 src2 dst1 dst2 in
    log_msg s;
    (match pkt.addr with
      ADDR_PKT(dev_type, _, proto) ->
	let addr_out = ADDR_PKT(dev_type, oport_dev, proto) in
(* If CRC is added, then need to send "pkt.len - 4" instead of pkt.len *)
    	send_pkt_out oport pkt.pkt 0 pkt.len addr_out
    | _ -> raise Exit);
    Mutex.unlock m
  done;
  let s =
    Printf.sprintf "handle_out stopped (%s)\n" (debug_oport_to_string oport) in
  log_msg s

let top_thread () =
  let rec get_all_ifs () =
    try
      let iport = get_iport () in
      iport :: get_all_ifs ()
    with No_interface -> []
  in let hashtable = Hashtbl.create 257 in
  let hash_m = Mutex.create () in
  let start_m = Mutex.create () in
  let rec start_handlers iport_list =
    match iport_list with
      [] -> []
    | h :: t ->
	let out_q = Queue.create () in
	let out_m = Mutex.create () in
	let out_c = Condition.create () in
	let oport = iport_to_oport h in
	all_outs := (out_q, out_m, out_c) :: !all_outs;
	all_devs := {dev=(debug_iport_to_string h);state=forwarding} :: !all_devs;
	(Safethread.create (hashval())
	 handle_if (h, hashtable, hash_m, start_m, (out_q, out_m, out_c)))
	:: (Safethread.create (hashval()) handle_out
	    (oport, out_q, out_m, out_c, start_m))
	:: start_handlers t
  in let rec kill_all_threads pid_list =
    match pid_list with
      [] -> ()
    | h :: t -> Safethread.kill h; kill_all_threads t in
  let rec clean_up iport_list =
    match iport_list with
      [] -> []
    | h :: t ->
	let oport = iport_to_oport h in
	unbind_out oport; unbind_in h; clean_up t
  in let port_list = get_all_ifs () in
  let thread_list = Mutex.lock start_m; start_handlers port_list in
  Mutex.unlock start_m;
  Mutex.lock die_mutex;
  while not !die do
    Condition.wait die_condition die_mutex
  done;
  kill_all_threads thread_list;
  clean_up port_list

let bridge ignored =
  let t = Safethread.create (hashval()) top_thread () in
  Safethread.join t;
  ""

let _ = value := Some(Func.register "bridge" bridge)

