(* routtab.mli : implementation to manage the routing table
 *
 * This is pretty simple for now.  It just lets you read in
 * a static table from disk and access routes.  No runtime
 * modification allowed.
 *
 * $Revision: 1.1 $
 *)

let max_route_entries = 257

(*********)
(* Types *)
(*********)

type rout_table = (Activehost.activehost,Activehost.activehost) Hashtbl.t *
      Mutex.t

(********************)
(* Helper Functions *)
(********************)

(* Token extracting routines *)

let rec skip_space fin =
  let c = input_char fin in
    if c = ' '  or
       c = '\n' or
       c = '\t' then
      skip_space fin
    else
      seek_in fin ((pos_in fin) - 1)

let rec get_token fin buf ofs =
  try
    let c = input_char fin in
      if not (c = ' ' or c = '\n' or c = '\t') then (
        String.set buf ofs c;
        get_token fin buf (ofs + 1)
      )
      else
        ofs
  with End_of_file ->
    if ofs = 0 then -1 else ofs

let next_token fin =
  let buf = String.create 65535 in
    (skip_space fin;
     let len = get_token fin buf 0 in
       if len = -1 then raise End_of_file
       else String.sub buf 0 len)

(***********************)
(* Interface Functions *)
(***********************)

(* Creates a new routing table *)

exception BadRouteTableFile of string
let init_table rout_file defPort =
  let routes = (Hashtbl.create max_route_entries) in
  let hashlock = Mutex.create () in
  let tab = (routes,hashlock) in
  let fin = 
    try
     open_in rout_file
    with Sys_error _ -> raise (BadRouteTableFile rout_file)
  in
  try
    while true do (
      let dest = next_token fin in
      let hop = next_token fin in
      let hop_addr = Activehost.stringToActiveHost hop defPort in
      if (dest = "default") then
        Hashtbl.add routes Activehost.Unknown hop_addr
      else
        Hashtbl.add routes (Activehost.stringToActiveHost hop defPort) hop_addr
    ) done;
    tab
  with 
    End_of_file -> (close_in fin; tab)
  | e -> raise (BadRouteTableFile rout_file)

(* Returns the next hop to eventually reach the destination; if no
   such hop exists, raise the exception Not_found *)

exception No_route
let next_hop (dest,tab) =
  let (routes,m) = tab in
  Mutex.lock m;
  try
    let hop = Hashtbl.find routes dest in
    Mutex.unlock m;
    hop
  with Not_found ->
    (try
      let hop = Hashtbl.find routes Activehost.Unknown in
      Mutex.unlock m;
      hop
    with Not_found ->
      (Mutex.unlock m;
       let ebuf = Printf.sprintf "No route for address %s, no default route\n"
           (Activehost.activeHostToString dest) in
       Log.log_msg ebuf;
       raise No_route))
  | e -> Mutex.unlock m; raise e

