
(*
  * These functions illustrate how to play with very basic file I/O 
  * and string matching in ocaml.  Their function is to provide 
  * a mechanism for doing route lookups in our active node.
  *
  * Sensei      : Scott
  * Grasshopper : Bill 
  *)

open Printf
open Str
open Unix

module Get_bytecode = struct
exception No_bytecode

let hashtablesize = 257

let hashtable = Hashtbl.create hashtablesize
let hashlock = Mutex.create ()

type bytecode_item = {name : string; filename : string}

(* returns the contents of a channel descriptor as a string*)
let rec make_string inch =
  try
    let foo = input_line inch in
    sprintf "%s\n%s" foo (make_string inch)
  with End_of_file -> ""

(* returns a type route_item from file descriptor "inch" *)
let make_ri inch =
  let str = input_line inch in
  let strlist = split (regexp "[ \t]+") str in
  match strlist with
    name  :: filename :: [] -> {name=name; filename=filename}
  | _ -> raise (Failure "Bad Bytecode Mapping Table")

(* given a filename create a list of route_items *)
let make_rte_list filename = 
  let inch = open_in filename in
  let rec aux inch =
    try 
      let ri = make_ri inch in
      ri :: (aux inch)
    with
      End_of_file -> [] in
  aux inch

(* given a list of route_items create a hash table *)
let rec make_rte_hash rte_list =
  match rte_list with
    [] -> () (* this returns a unit *)
  | head :: tail -> Hashtbl.add hashtable head.name head.filename;
      make_rte_hash tail

(*
  * Read the contents of a file into a string 
  *)
let read_file filename = 
  let inch = open_in filename in
  let len  = in_channel_length inch in
  let str = String.create (len + 1) in
  input inch str 0 len;
  close_in inch;
  str
    ;;


(* Given a destination return a next hop *)
let get_bytecode dest = 
  try   
    Mutex.lock hashlock;
    let filename = Hashtbl.find hashtable dest in
    Mutex.unlock hashlock;
    read_file filename
  with Not_found   ->	Mutex.unlock hashlock; 
    eprintf "No Bytecode found in Hash";
    raise No_bytecode
  | e -> 	Mutex.unlock hashlock; 
      raise e

(* Single Line Invocation of all the useful stuff so far *)
let bytecode_init filename =
  try
    Mutex.lock hashlock;
    make_rte_hash (make_rte_list filename);
    Mutex.unlock hashlock
  with e -> 	Mutex.unlock hashlock; 
    raise e

let _ = bytecode_init "bytecode_table"
end
