(* Additional PLAN service for dynamic loading of PLAN services *)
module Install_svc_impl = struct

open Basis
open Eval
open PlanExn
open Services

(* Interface functions *)

(* given a list of strings, which should be Caml bytecodes, installs 
   each of them in order, left to right *)
let install (p,l) =
  match l with
    [VList bcs] -> 
      List.iter 
	(function v ->
	  match v with
	    (VTuple [String name; Blob b]) ->
	      (try
                (* this will load the file and register its services *)
		Func.load_string b;
                if name <> "" then
		  Func.call_func (name,"")
	      with Dynlink.Error e -> 
        	Log.log_msg 
		  ("Service loading failed: "^(Dynlink.error_message e)^"\n");
        	raise (PLANException("BadService")))
	  | v ->
	      raise (ExecException(TypeError((Variable "blob"),v,BlobType))))
	bcs; Unit
  | [v] -> raise (ExecException
		    (TypeError((Variable "list"),v,
			       (ListType (TupleType [StringType;BlobType])))))
  | _ -> raise (ExecException(ArgMismatch("install",
					  List.map
					    (function v -> Val v)
					    l,
					  1)))

let register_svcs () =
  register_svc("installServices",install);
  register_exn("BadService")

end

(* This defines the bridge-visible interface to the PLAN packet handler *)
module Pland = struct
open Basis

(* Standard error processing function *)
let send_error_packet s pkt =
  if (pkt.handler = "") then
    ()
  else
    match pkt.interface with
      None -> Log.log_msg "Got bad interface in send_error_packet\n"; ()
    | Some(ifc) ->
	let c = ([],
		 pkt.handler,
		 (* arguments to handler: name of exception
		    raised, host on which it was raised,
		    top-level (truncated) chunk that was executing
		    when it happened *)
		 [String s; 
		   Host (Net.meDev ifc);
		   Chunk([],
			 pkt.fn_to_exec,
			 List.map
			   (function v -> truncate_value v)
			   pkt.bindings)])
	in
	Svc_impl.abort c pkt.session

(* standard packet processing function *)
let handler pkt iss_opt dev =
  pkt.rb <- pkt.rb - 1;               (* charge 1 rb for arriving *)
  pkt.interface <- Some(dev);
  if dev = Net.local_if then
    (let ebuf = Printf.sprintf "[ LOC ] : %s\n" pkt.fn_to_exec in
    Log.log_msg ebuf)
  else
    (let ebuf = Printf.sprintf "[ IN ] : %s\n" pkt.fn_to_exec in
    Log.log_msg ebuf);
  (try
    (* If this is our evaluation destination, we must unmarshall
       the packet the rest of the way, and then evaluate *)
    if ((Net.isMe pkt.evalDest) || 
    (pkt.evalDest = Activehost.Broadcast_addr)) then
      ((match iss_opt with
	Some(iss) ->
	  Marshalling.unmarshall_packet_rest pkt iss
      | _ -> ());
       Eval.interpret (pkt.code, pkt, Environ.top_level); 
       let ebuf = Printf.sprintf "Eval thread %d completed OK\n" 
	   (Thread.id(Thread.self()))
       in Log.log_msg ebuf; ())
    else
      try
        (* send it on towards its evalDest *)
      	(match (Environ.lookup(pkt.routFun,Environ.top_level)) with
	  Some (Environ.Svc sfun) ->
	    (match (sfun(pkt, [ Host pkt.evalDest ])) with
	      VTuple [Host hop; Dev (h,dev)] ->
	      	if Net.isMeDev h dev then
	      	  if pkt.rb >= 1 then
		    (* Send the packet *)
		    (match iss_opt with
		      Some(iss) ->
			let buf = Marshall_prim.fromInStringStream iss in
	      		Marshalling.set_packet_rb buf pkt.rb;
			Net.send_active_packet (Net.Buffer buf) dev hop
		    | None ->
			Net.send_active_packet (Net.Packet pkt) dev hop)
		  else
		    raise (PlanExn.PLANException "OutOfRB")
	      	else
		  raise (PlanExn.PLANException "InvalidDevice")
	    | _ -> raise (Eval.ExecException(Eval.BadRoutFun(pkt.routFun))))
      	| _ -> raise (Eval.ExecException(Eval.BadRoutFun(pkt.routFun))))
      with e ->
	(* Unmarshall the bindings before calling the error handler *)
	(match iss_opt with
	  Some(iss) ->
	    Marshalling.unmarshall_packet_rest_no_code pkt iss;
	| _ -> ());
	raise e
  with 
  | e ->
      ((try
	send_error_packet
	  (match e with
	    PlanExn.PLANException s -> s
	  | ParseExn.UndefinedType(_,_,_) -> "ParseExn.UndefinedType"
	  | ParseExn.SyntaxError(_,_) -> "ParseExn.SyntaxError"
	  | Eval.ExecException ee ->
	      (match ee with
		Eval.TypeError(_,_,_) -> "ExecExn.TypeError"
	      | Eval.ArgMismatch(_,_,_) -> "ExecExn.ArgMismatch"
	      | Eval.AppError(_,_) -> "ExecExn.AppError"
	      | Eval.UnboundVariable _ -> "ExecExn.UnboundVariable"
	      | Eval.BadRoutFun _ -> "ExecExn.BadRoutFun"
	      | Eval.EmptySequence -> "ExecExn.EmptySequence"
	      | Eval.Generic s -> "ExecExn.Generic")
	  | Unix.Unix_error(_,_,_) -> "Unix.Unix_error"
	  | Marshall_prim.Marshalling_exception m -> "Net.MarshallingException"
	  | _ -> "Unknown")
	  pkt
      (* don't worry if we can't send error packet *)
      with _ -> ());
       (* also log it to console *)
       let ebuf = Printf.sprintf "Eval thread %d exiting with failure:\n  "
	   (Thread.id(Thread.self())) in
       Log.log_msg (ebuf ^ 
         (match e 
	 with 
           PlanExn.PLANException s -> "Uncaught packet exception "^s^"\n"
         | ParseExn.UndefinedType (i, l, c) -> "Undefined type " ^
             (Basis.id2str i) ^ " at " ^
             (string_of_int l) ^ "." ^ (string_of_int c) ^ "\n"
         | ParseExn.SyntaxError (l, c) -> "Syntax error at " ^
             (string_of_int l) ^ "." ^ (string_of_int c) ^
             ": code =\n" ^ (Basis.program2str pkt.code)
         | Eval.ExecException ee -> (Eval.exec_exn_to_string ee) ^ "\n"
         | Unix.Unix_error (e,s1,s2) -> "Unix error : " ^
             (Unix.error_message e) ^ " on call to " ^ s1 ^ "\n"
         |  _ ->
             (try
               Printexc.print (function x -> raise e) ()
             with _ -> "Error: uncaught exception\n")));
       ()))

let proc_packets (p,dev) =
  let go () =
    match p with
      Net.Packet ap -> handler ap None dev
    | Net.Buffer buf ->
	(try
	  let (pkt,iss) = Marshalling.unmarshall_packet_header buf in
	  handler pkt (Some(iss)) dev
        (* XXX this won't make it back to the source *)
    	with Marshall_prim.Marshalling_exception m ->
	  let ebuf = Printf.sprintf "Marshalling exception %s from net.\n" m in
	  Log.log_msg ebuf; ())
  in
  (let t = Thread.create go () in
  let ebuf = Printf.sprintf "Spawned thread %d to evaluate\n" (Thread.id t) in
  Log.log_msg ebuf; ())

let start_plan (verbose,portnum,ifc_file) =
  Eval.register_default_exns ();
  Svc_impl.register_default_svcs ();
  Port_svc_impl.register_svcs ();
  Arp_svc_impl.register_svcs ();
  Resident_svc_impl.register_svcs ();
  Csum_svc_impl.register_svcs ();
  Frag_svc_impl.register_svcs ();
  Reliable_svc_impl.register_svcs ();
  Install_svc_impl.register_svcs ();
  if verbose then
    (Log.log_msg "Beginning to receive packets...\n"; ());
  let net_thread =
    Net.start_network (verbose,portnum,ifc_file,
		       proc_packets,send_error_packet,Parser.parse) in
  Rip.rip_start(Net.me(),Net.local_if);
  Rip_svc_impl.register_svcs();
  Thread.join net_thread;
  ()
end
