(*
 * These functions illustrate how to play with very basic file I/O 
 * and string matching in ocaml.  Their function is to provide 
 * a mechanism for doing route lookups in our active node.
 *
 * Sensei      : Scott
 * Grasshopper : Bill 
 *)

open Printf
open Str
open Safeunix

module Route = struct
exception No_route

let hashtablesize = 257

let hashtable = Hashtbl.create hashtablesize
let hashlock = Mutex.create ()

type route_item = {addr1: string; addr2: string}

(* returns the contents of a channel descriptor as a string*)
let rec make_string inch =
  try
    let foo = input_line inch in
    sprintf "%s\n%s" foo (make_string inch)
  with End_of_file -> ""


(* returns a type route_item from file descriptor "inch" *)
let make_ri inch =
  let str = input_line inch in
  let strlist = split (regexp "[ \t]+") str in
  match strlist with
    dest  :: addr :: [] -> {addr1=dest; addr2=addr}
  | _ -> raise (Failure "Bad Routing Table")

(* given a filename create a list of route_items *)
let make_rte_list filename = 
  let inch = open_in filename in
  let rec aux inch =
    try 
      let ri = make_ri inch in
      ri :: (aux inch)
    with End_of_file -> []
  in
  aux inch

(* given a list of route_items create a hash table *)
let rec make_rte_hash rte_list =
  match rte_list with
    [] -> () (* this returns a unit *)
  | head :: tail ->
      Hashtbl.add hashtable head.addr1 (inet_addr_of_string head.addr2);
      make_rte_hash tail

(* Given a destination return a next hop *)
let get_route dest = 
  try   
    Mutex.lock hashlock;
    let addr = Hashtbl.find hashtable dest in
    Mutex.unlock hashlock;
    addr
  with Not_found -> Mutex.unlock hashlock; 
    eprintf "No Route to '%s' found in Hash\n" dest;
    raise No_route
  | Failure msg -> Mutex.unlock hashlock; 	
      eprintf "Improper Address Format\n";
      raise No_route
  | e -> Mutex.unlock hashlock; 
      raise e

(* Single Line Invocation of all the useful stuff so far *)
let route_init filename =
  try
    Mutex.lock hashlock;
    make_rte_hash (make_rte_list filename);
    Mutex.unlock hashlock
  with e -> Mutex.unlock hashlock; 
    raise e

(* Tests some things out for now *)
(*route_init "rt.table";;
get_route "echidna";;
get_route "oilhead";;
get_route "toscott";;
get_route "tobill";;*)
end
