(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: safethread.ml,v 1.4 1998/02/08 22:44:55 salex Exp $ *)

(* User-level threads *)

module Safethread = struct
type t = Thread.t

type magic_cookie = Magic of int

type resources = { mutable max_threads : int; mutable cur_threads : int }

let resource_table = Hashtbl.create 257

(* It is mucho important that the primitives that reschedule are called 
   through an ML function call, not directly. That's because when such a
   primitive returns, the bytecode interpreter is only semi-obedient:
   it takes sp from the new thread, but keeps pc from the old thread.
   But that's OK if all calls to rescheduling primitives are immediately
   followed by a RETURN operation, which will restore the correct pc
   from the stack. *)

let id = Thread.id

let wait_read = Thread.wait_read
let wait_write = Thread.wait_write
let delay = Thread.delay
let join = Thread.join
let self = Thread.self
let kill = Thread.kill

let wait_timed_read = Thread.wait_timed_read
let wait_timed_write = Thread.wait_timed_write
let wait_pid = Thread.wait_pid

(* For new, make sure the function passed to thread_new always terminates
   by calling exit. *)

let next_magic = ref 0
let rec magic_create () =
  let hashval =
(*    try
      let rand_chan = open_in_bin "/dev/urandom" in
      let result = input_binary_int rand_chan in
      close_in rand_chan;
      Magic(result)
    with Sys_error(_) -> prerr_endline "/dev/random not found"; Magic(0)
    | End_of_file -> prerr_endline "EOF on /dev/random; this can't happen";
      	Magic(0)*)
    let res = !next_magic in
    incr next_magic;
    if res > !next_magic then failwith("magic_create: too many requestors")
    else Magic(res)
  in
  try
    Hashtbl.find resource_table hashval;
    magic_create ()
  with Not_found ->
    Hashtbl.add resource_table hashval
    (ref { max_threads = 1000; cur_threads = 0});
    hashval

let exit hashval =
  begin try
    let ent = Hashtbl.find resource_table hashval in
    if (!ent).cur_threads > 0 then
      (!ent).cur_threads <- (!ent).cur_threads - 1
  with Not_found -> () end;
  Thread.exit ()

let create hashval fn arg =
  try
    let ent = Hashtbl.find resource_table hashval in
    (*if (!ent).cur_threads + 1 < (!ent).max_threads then begin*)
      (!ent).cur_threads <- (!ent).cur_threads + 1;
      Thread.create (fun x -> fn x; exit hashval) arg
    (*end
    else failwith "Too many threads"*)
  with Not_found -> raise (Invalid_argument("Bad magic cookie"))
end

module Mutex = Mutex
module Condition = Condition
