(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: safethread.mli,v 1.1 1998/01/06 03:07:32 salex Exp $ *)

(* Module [Thread]: user-level lightweight threads *)

module Safethread : sig
type t
        (* The type of thread handles. *)

type magic_cookie

(** Thread creation and termination *)

val magic_create : unit -> magic_cookie
(* Actually, this should be thinned from the interface seen by switchlets.
   As it stands, a switchlet could create an arbitrary number of magic
   cookies and thus overallocated resources. XXX *)

val create : magic_cookie -> ('a -> 'b) -> 'a -> t
        (* [Thread.create funct arg] creates a new thread of control,
           in which the function application [funct arg]
           is executed concurrently with the other threads of the program.
           The application of [Thread.create]
           returns the handle of the newly created thread.
           The new thread terminates when the application [funct arg]
           returns, either normally or by raising an uncaught exception.
           In the latter case, the exception is printed on standard error,
           but not propagated back to the parent thread. Similarly, the
           result of the application [funct arg] is discarded and not
           directly accessible to the parent thread. *)
val self : unit -> t
        (* Return the thread currently executing. *)
val id : t -> int
        (* Return the identifier of the given thread. A thread identifier
           is an integer that identifies uniquely the thread.
           It can be used to build data structures indexed by threads. *)
val exit : magic_cookie -> unit
        (* Terminate prematurely the currently executing thread. *)
val kill : t -> unit
        (* Terminate prematurely the thread whose handle is given. *)

(** Suspending threads *)

val delay: float -> unit
        (* [delay d] suspends the execution of the calling thread for
           [d] seconds. The other program threads continue to run during
           this time. *)
val join : t -> unit
        (* [join th] suspends the execution of the calling thread
           until the thread [th] has terminated. *)
(*val wait_read : Unix.file_descr -> unit
val wait_write : Unix.file_descr -> unit
        (* Suspend the execution of the calling thread until at least
           one character is available for reading ([wait_read]) or
           one character can be written without blocking ([wait_write])
           on the given Unix file descriptor. *)
val wait_timed_read : Unix.file_descr -> float -> bool
val wait_timed_write : Unix.file_descr -> float -> bool
        (* Same as [wait_read] and [wait_write], but wait for at most
           the amount of time given as second argument (in seconds).
           Return [true] if the file descriptor is ready for input/output
           and [false] if the timeout expired. *)
val wait_pid : int -> int * Unix.process_status*)
        (* [wait_pid p] suspends the execution of the calling thread
           until the Unix process specified by the process identifier [p]
           terminates. A pid [p] of [-1] means wait for any child.
           A pid of [0] means wait for any child in the same process group
           as the current process. Negative pid arguments represent
           process groups. Returns the pid of the child caught and
           its termination status, as per [Unix.wait]. *)
end

module Mutex : sig
type t
        (* The type of mutexes. *)
val create: unit -> t
        (* Return a new mutex. *)
val lock: t -> unit
        (* Lock the given mutex. Only one thread can have the mutex locked
           at any time. A thread that attempts to lock a mutex already locked
           by another thread will suspend until the other thread unlocks
           the mutex. *)
val try_lock: t -> bool
        (* Same as [try_lock], but does not suspend the calling thread if
           the mutex is already locked: just return [false] immediately
           in that case. If the mutex is unlocked, lock it and
           return [true]. *)
val unlock: t -> unit
        (* Unlock the given mutex. Other threads suspended trying to lock
           the mutex will restart. *)
end

module Condition : sig
type t
        (* The type of condition variables. *)
val create: unit -> t
        (* Return a new condition variable. *)
val wait: t -> Mutex.t -> unit
        (* [wait c m] atomically unlocks the mutex [m] and suspends the
           calling process on the condition variable [c]. The process will
           restart after the condition variable [c] has been signalled.
           The mutex [m] is locked again before [wait] returns. *)
val signal: t -> unit
        (* [signal c] restarts one of the processes waiting on the 
           condition variable [c]. *)
val broadcast: t -> unit
        (* [broadcast c] restarts all processes waiting on the 
           condition variable [c]. *)
end
