open Unix

module Safeunix = struct
type error =
    E2BIG
  | EACCES
  | EAGAIN
  | EBADF
  | EBUSY
  | ECHILD
  | EDEADLK
  | EDOM
  | EEXIST
  | EFAULT
  | EFBIG
  | EINTR
  | EINVAL
  | EIO
  | EISDIR
  | EMFILE
  | EMLINK
  | ENAMETOOLONG
  | ENFILE
  | ENODEV
  | ENOENT
  | ENOEXEC
  | ENOLCK
  | ENOMEM
  | ENOSPC
  | ENOSYS
  | ENOTDIR
  | ENOTEMPTY
  | ENOTTY
  | ENXIO
  | EPERM
  | EPIPE
  | ERANGE
  | EROFS
  | ESPIPE
  | ESRCH
  | EXDEV
  | EWOULDBLOCK
  | EINPROGRESS
  | EALREADY
  | ENOTSOCK
  | EDESTADDRREQ
  | EMSGSIZE
  | EPROTOTYPE
  | ENOPROTOOPT
  | EPROTONOSUPPORT
  | ESOCKTNOSUPPORT
  | EOPNOTSUPP
  | EPFNOSUPPORT
  | EAFNOSUPPORT
  | EADDRINUSE
  | EADDRNOTAVAIL
  | ENETDOWN
  | ENETUNREACH
  | ENETRESET
  | ECONNABORTED
  | ECONNRESET
  | ENOBUFS
  | EISCONN
  | ENOTCONN
  | ESHUTDOWN
  | ETOOMANYREFS
  | ETIMEDOUT
  | ECONNREFUSED
  | EHOSTDOWN
  | EHOSTUNREACH
  | ELOOP
  | EUNKNOWNERR

exception Unix_error of error * string * string
external error_message : error -> string = "unix_error_message"

(*** Time functions *)

type process_times = Unix.process_times =
  { tms_utime : float;          (* User time for the process *)
    tms_stime : float;          (* System time for the process *)
    tms_cutime : float;         (* User time for the children processes *)
    tms_cstime : float }        (* System time for the children processes *)

        (* The execution times (CPU times) of a process. *)

type tm = Unix.tm =
  { tm_sec : int;                       (* Seconds 0..59 *)
    tm_min : int;                       (* Minutes 0..59 *)
    tm_hour : int;                      (* Hours 0..23 *)
    tm_mday : int;                      (* Day of month 1..31 *)
    tm_mon : int;                       (* Month of year 0..11 *)
    tm_year : int;                      (* Year - 1900 *)
    tm_wday : int;                      (* Day of week (Sunday is 0) *)
    tm_yday : int;                      (* Day of year 0..365 *)
    tm_isdst : bool }                   (* Daylight time savings in effect *)

        (* The type representing wallclock time and calendar date. *)

(* external time : unit -> int = "unix_time" *)

let time = time
and gettimeofday = gettimeofday
and gmtime = gmtime
and localtime = localtime
and mktime = mktime
and times = times

type inet_addr = Unix.inet_addr
        (* The abstract type of Internet addresses. *)

type sockaddr = Unix.sockaddr =
    ADDR_UNIX of string
  | ADDR_INET of inet_addr * int
  | ADDR_PKT of int * string * int

external inet_addr_of_string : string -> inet_addr = "unix_inet_addr_of_string"

let string_of_inet_addr = Unix.string_of_inet_addr
(*"unix_string_of_inet_addr"*)

let inet_addr_any = Unix.inet_addr_any
    (*= inet_addr_of_string "0.0.0.0"*)

let gethostname = Unix.gethostname

type hwaddr = Unix.hwaddr = ETHER of int * int

(*
(* required for timer *)
(*external pause : unit -> unit = "unix_pause"*)
let pause = pause
type interval_timer =
    ITIMER_REAL

type interval_timer_status =
  { it_interval: float;                 (* Period *)
    it_value: float }                   (* Current value of the timer *)

external getitimer: interval_timer -> interval_timer_status
   = "unix_getitimer" "unix_getitimer_native"
external setitimer:
  interval_timer -> interval_timer_status -> interval_timer_status
  = "unix_setitimer" "unix_setitimer_native"
*)
end

module Unixnet = struct

(* packet is the type of data returned to the calling function.  The
 * len field is the total length of the data returned in the pkt field.
 * The addr contains information about the origin of the packet as
 * produced by the OS.  The pkt field contains the actual data in the
 * packet.
 *)
type next_fn = (packet -> unit) option

and packet = { len : int; addr : Safeunix.sockaddr; pkt : string;
		next_fn : next_fn }

(* iport and oport are the types for input and output ports, respectively.
 * Each port type is simplex and is opaque to its user.
 *)
type iport = IPORT of string * packet Queue.t * Mutex.t * Condition.t
type oport = OPORT of string

(* type status_request = GETHWADDR | GETIPADDR *)

exception Already_bound
exception No_interface

let sfd = ref None
let demux_thread = ref None
let demux_thread_mutex = Mutex.create () (* also covers initial socket() call*)

let queue_out = Queue.create ()
let mutex_out = Mutex.create ()
let condition_out = Condition.create ()

(* Some support functions *)
let rec strncpy str ofs len =
    let c = String.get str ofs in
    if len = 0 or Char.code c = 0 then ""
    else (Char.escaped c) ^ (strncpy str (ofs + 1) (len - 1))

let get_ifreq sfd dev = Unix.siocgifflags sfd (Unix.FLAGS(dev, []))

let rec find_flag flag flag_list =
  match flag_list with
    [] -> false
  | h :: t -> if h = flag then true else find_flag flag t

let set_promisc sfd ifr =
  match ifr with
    (Unix.FLAGS(dev, flags)) ->
      let s = Printf.sprintf "set_promisc %s\n" dev in
      ThreadUnix.write Unix.stderr s 0 (String.length s);
      Unix.siocsifflags sfd (Unix.FLAGS(dev, Unix.IFF_PROMISC :: flags))
  | _ -> Log.log_msg "set_promisc: ifr != Unix.FLAGS\n"; raise Exit

let clear_promisc sfd ifr =
  let rec clear_flag flag =
    match flag with
      [] -> []
    | Unix.IFF_PROMISC :: t -> clear_flag t
    | h :: t -> h :: (clear_flag t)
  in
  match ifr with
    (Unix.FLAGS(dev, flags)) ->
      let s = Printf.sprintf "clear_promisc %s\n" dev in
      ThreadUnix.write Unix.stderr s 0 (String.length s);
      Unix.siocsifflags sfd (Unix.FLAGS(dev, clear_flag flags))
  | _ -> Log.log_msg "clear_promisc: ifr != Unix.FLAGS\n"; raise Exit

let rec to_promisc sfd ifreq_list =
  match ifreq_list with
    [] -> []
  | (Unix.ADDR(dev, _)) :: t->
      let oifr = get_ifreq sfd dev in
      (match oifr with
	(Unix.FLAGS(_, flags)) ->
	  if find_flag Unix.IFF_LOOPBACK flags
	  then to_promisc sfd t
	  else begin
      	    set_promisc sfd oifr;
      	    oifr :: to_promisc sfd t
	  end
      |	_ -> Log.log_msg "to_promisc: oifr != Unix.FLAGS\n"; raise Exit)
  | _ -> Log.log_msg "to_promisc: ifreq_list != Unix.ADDR...\n"; raise Exit

let rec from_promisc sfd ifreq_list =
  match ifreq_list with
    [] -> ()
  | (Unix.ADDR(dev, _)) :: t ->
      let oifr = get_ifreq sfd dev in
      clear_promisc sfd oifr;
      from_promisc sfd t
  | _ -> Log.log_msg "from_promisc: ifreq_list != Unix.ADDR...\n"; raise Exit

let to_promisc_all sfd =
  let ifreq_list = Unix.siocgifconf sfd in
  to_promisc sfd ifreq_list

let from_promisc_all sfd =
  let ifreq_list = Unix.siocgifconf sfd in
  from_promisc sfd ifreq_list

let promisc_on_all () =
  match !sfd with
    None ->
      failwith "promisc_on_all: Unixnet not initialized"
  | Some(s) ->
      to_promisc_all s; ()

let promisc_off_all () =
  match !sfd with
    None ->
      failwith "promisc_off_all: Unixnet not initialized"
  | Some(s) ->
      from_promisc_all s; ()

(* The sizes of these tables don't have to be the same.  I'm guessing that
 * 10 is a reasonable max number of devices or addresses on which we might
 * switch.  Thus, 23 is a good size for
 * each of these tables if these guesses are correct.
 *)
let demux_dev_tbl = Hashtbl.create 23
let demux_addrs_tbl = Hashtbl.create 23
let demux_type_tbl = Hashtbl.create 23
let demux_num_devs = ref 0
let demux_dev_tbl_mutex = Mutex.create ()
let demux_str_len = 2048
let nxt_str = ref 0
let str_arr_sz = 3
let demux_strs = Array.create str_arr_sz ""

let call_next_function pkt =
  match pkt.next_fn with
    None -> ()
  | Some(func) -> func pkt

let no_next = None

let check_for_def pkt =
  let new_pkt =
    { len=pkt.len; addr=pkt.addr; pkt=pkt.pkt; next_fn=None }
  in
  match pkt.addr with
    ADDR_PKT(_, dev, _) ->
      Mutex.lock demux_dev_tbl_mutex;
      (try
    	(Hashtbl.find demux_dev_tbl dev) new_pkt
      with Not_found ->
    	Mutex.unlock demux_dev_tbl_mutex; call_next_function new_pkt);
      Mutex.unlock demux_dev_tbl_mutex
  | _ -> ()

let check_for_type pkt =
  let to_int16 str ofs =
    let high = Char.code (String.get str ofs) in
    let low = Char.code (String.get str (ofs+1)) in
    (high lsl 8) lor low in
  let new_pkt =
    { len=pkt.len; addr=pkt.addr; pkt=pkt.pkt; next_fn=Some(check_for_def) }
  in
  let typ = to_int16 pkt.pkt 12 in
  Mutex.lock demux_dev_tbl_mutex;
  (try (Hashtbl.find demux_type_tbl typ) new_pkt
  with Not_found ->
    Mutex.unlock demux_dev_tbl_mutex; call_next_function new_pkt);
  Mutex.unlock demux_dev_tbl_mutex

let check_for_addr pkt =
  let new_pkt =
    { len=pkt.len; addr=pkt.addr; pkt=pkt.pkt; next_fn=Some(check_for_type) }
  in
  if !demux_num_devs = 0 then call_next_function new_pkt
  else begin
    let dest = String.sub pkt.pkt 0 6 in
    Mutex.lock demux_dev_tbl_mutex;
    (try
      (Hashtbl.find demux_addrs_tbl dest) new_pkt
    with Not_found ->
      Mutex.unlock demux_dev_tbl_mutex; call_next_function new_pkt);
    Mutex.unlock demux_dev_tbl_mutex
  end

let demux_next_pkt sfd =
  let str = Array.get demux_strs !nxt_str in
  if !nxt_str = (str_arr_sz - 1) then
    nxt_str := 0
  else incr nxt_str;
  let (len, saddr) = ThreadUnix.recvfrom sfd str 0 demux_str_len [] in
  call_next_function
  { len=len; addr=saddr; pkt=str; next_fn=Some(check_for_addr) }

let demuxer () =
  match !sfd with
    None ->
      Log.log_msg "demuxer: !sfd = None\n";
      raise Exit
  | Some(s) ->
      while true do
	try
	  demux_next_pkt s
	with e -> (Log.log_msg (Printexc.to_string e); ())
      done

let demux_register dev f =
  Hashtbl.add demux_dev_tbl dev f;
  incr demux_num_devs;
  if !demux_num_devs = 1
  then match !sfd with
    Some(s) -> to_promisc_all s; ()
  | None -> Log.log_msg "demux_register: !sfd = None\n"; raise Exit

let demux_register_addr addr f =
  Hashtbl.add demux_addrs_tbl addr f;
  incr demux_num_devs;
  if !demux_num_devs = 1
  then match !sfd with
    Some(s) -> to_promisc_all s; ()
  | None -> Log.log_msg "demux_register_addr: !sfd = None\n"; raise Exit

let demux_register_type typ f =
  if !demux_num_devs = 0
  then raise No_interface
  else Hashtbl.add demux_type_tbl typ f

let demux_deregister dev =
  decr demux_num_devs;
  Hashtbl.remove demux_dev_tbl dev;
  if !demux_num_devs = 0
  then match !sfd with
    Some(s) -> from_promisc_all s; ()
  | None -> Log.log_msg "demux_deregister: !sfd = None\n"; raise Exit

let demux_deregister_addr addr =
  decr demux_num_devs;
  Hashtbl.remove demux_addrs_tbl addr;
  if !demux_num_devs = 0
  then match !sfd with
    Some(s) -> from_promisc_all s; ()
  | None -> Log.log_msg "demux_deregister_addr: !sfd = None\n"; raise Exit

let demux_deregister_type typ =  Hashtbl.remove demux_type_tbl typ

let demux_query tbl value =
  try
    let _ = Hashtbl.find tbl value in
    true
  with Not_found -> false

let demux_query_dev dev = demux_query demux_dev_tbl dev
let demux_query_addr addr = demux_query demux_addrs_tbl addr
let demux_query_type typ = demux_query demux_type_tbl typ

(* Input ports *)
let compare_iport (IPORT(dev1, q1, m1, c1)) (IPORT(dev2, q2, m2, c2)) =
  Log.log_msg "compare_iport\n";
  if dev1 = dev2 then begin
    Log.log_msg "dev1 = dev2\n";
    if q1 = q2 then begin
      Log.log_msg "q1 = q2\n";
      if m1 = m2 then begin
	Log.log_msg "m1 = m2\n";
	if c1 = c2 then begin
	  Log.log_msg "c1 = c2\n";
	  true
	end else begin
	  Log.log_msg "c1 <> c2\n";
	  false
	end
      end else begin
	Log.log_msg "m1 <> m2\n";
	false
      end
    end else begin
      Log.log_msg "q1 <> q2\n";
      false
    end
  end else begin
    Log.log_msg "dev1 <> dev2\n";
    false
  end

let start_socket () =
  Mutex.lock demux_thread_mutex;
  if !sfd = None then begin
    sfd := Some(Unix.socket Unix.PF_INET Unix.SOCK_PACKET 0x300);
    for i = 0 to (str_arr_sz - 1) do
      Array.set demux_strs i (String.make demux_str_len ' ')
    done;
    demux_thread := Some(Thread.create demuxer ()) end
  else ();
  Mutex.unlock demux_thread_mutex
    
let queue_in q m c pkt =
  Mutex.lock m;
  Queue.add pkt q;
  Condition.signal c;
  Mutex.unlock m

let bind_in dev =
  Mutex.lock demux_dev_tbl_mutex;
  if demux_query_dev dev
  then begin Mutex.unlock demux_dev_tbl_mutex; raise Already_bound end
  else begin
    let q = Queue.create () in
    let m = Mutex.create () in
    let c = Condition.create () in
    start_socket();
    demux_register dev (queue_in q m c);
    Mutex.unlock demux_dev_tbl_mutex;
    IPORT(dev, q, m, c) end

(* Must do at least one bind_in before calling *)
let bind_addr addr =
  Mutex.lock demux_dev_tbl_mutex;
  if demux_query_addr addr
  then begin Mutex.unlock demux_dev_tbl_mutex; raise Already_bound end
  else begin
    let q = Queue.create () in
    let m = Mutex.create () in
    let c = Condition.create () in
    start_socket();
    demux_register_addr addr (queue_in q m c);
    Mutex.unlock demux_dev_tbl_mutex;
    IPORT(addr, q, m, c) end

(* Must do at least one bind_in before calling *)
let bind_type typ =
  Mutex.lock demux_dev_tbl_mutex;
  if demux_query_type typ
  then begin Mutex.unlock demux_dev_tbl_mutex; raise Already_bound end
  else begin
    let q = Queue.create () in
    let m = Mutex.create () in
    let c = Condition.create () in
    start_socket();
    demux_register_type typ (queue_in q m c);
    Mutex.unlock demux_dev_tbl_mutex;
    IPORT(string_of_int typ, q, m, c) end

let unbind_in port =
  match port with
    IPORT(dev, _, _, _) ->
      Mutex.lock demux_dev_tbl_mutex;
      demux_deregister dev;
      Mutex.unlock demux_dev_tbl_mutex

let unbind_addr port =
  match port with
    IPORT(addr, _, _, _) ->
      Mutex.lock demux_dev_tbl_mutex;
      demux_deregister_addr addr;
      Mutex.unlock demux_dev_tbl_mutex

let unbind_type port =
  match port with
    IPORT(typ, _, _, _) ->
      Mutex.lock demux_dev_tbl_mutex;
      demux_deregister_type (int_of_string typ);
      Mutex.unlock demux_dev_tbl_mutex

let pkts_waiting_p_in port =
  match port with
    IPORT(_, q, m, _) ->
      Mutex.lock m;
      let ret = (Queue.length q) <> 0 in
      Mutex.unlock m; ret

let get_next_pkt_in port =
  match port with
    IPORT(_, q, m, c) ->
      Mutex.lock m;
      while Queue.length q = 0 do
    	Condition.wait c m;
      done;
      let pkt = Queue.take q in
      Mutex.unlock m; pkt

let timed_get_next_pkt_in port timeout =
  match port with
    IPORT(_, q, m, c) ->
      let finished = ref false in
      let bing timeout =
	try
	  Thread.delay timeout;
          finished := true;
          Condition.broadcast c
      	with e -> (Log.log_msg (Printexc.to_string e); raise e) in
      let t = Thread.create bing timeout in
      Mutex.lock m;
      while (not !finished) & Queue.length q = 0 do
        Condition.wait c m
      done;
      if not !finished then begin
      	Thread.kill t;
      	let pkt = Queue.take q in
      	Mutex.unlock m; pkt
      end else begin
	Mutex.unlock m;
      	raise (Safeunix.Unix_error(Safeunix.ETIMEDOUT, "timed_get_next_pkt_in", ""))
      end

let rec to_flags s ifreq_list =
  match ifreq_list with
    [] -> []
  | (Unix.ADDR(dev, _)) :: t -> get_ifreq s dev :: (to_flags s t)
  | _ -> Log.log_msg "to_flags: ifreq_list != Unix.ADDR...\n"; raise Exit

let get_iport () =
  let _ = start_socket () in
  match !sfd with
    None -> Log.log_msg "get_iport: !sfd = None\n"; raise Exit
  | Some(s) ->
      let ifreq_list = Unix.siocgifconf s in
      let ifflags_list = to_flags s ifreq_list in
      let rec find_if flags_list =
      	match flags_list with
      	  [] -> raise No_interface
	| (Unix.FLAGS(dev, f)) :: t ->
	    if find_flag Unix.IFF_LOOPBACK f then find_if t
	    else
	      (try bind_in dev
	      with e -> find_if t)
	| _ -> Log.log_msg "get_iport: flags_list != Unix.FLAGS...\n";
	    raise Exit
      in
      find_if ifflags_list

let get_ether_list () =
(*  match !sfd with
    None -> Log.log_msg "get_ether_list: !sfd = None\n"; raise Exit
  | Some(s) ->
      let rec get_addr ifreq_list addr_list =
	(match ifreq_list with
	  [] -> addr_list
	| (Unix.ADDR(dev, _)) :: t ->
	    let ifr = get_ifreq s dev in
	    begin match ifr with
	      Unix.FLAGS(_, flags) ->
		if find_flag Unix.IFF_LOOPBACK flags
		then get_addr t addr_list
		else begin
		  let ifreq = String.make 32 '\000' in
	  	  String.blit dev 0 ifreq 0 (String.length dev);
		  let retval = Unix.ioctl_ptr s 0x8927 ifreq in
		  if retval = -1 then get_addr t addr_list
		  else get_addr t ((dev, String.sub ifreq 18 6) :: addr_list)
		end
	    | _ -> Log.log_msg "get_ether_list: ifr != Unix.FLAGS\n";
		raise Exit
	    end
	| _ -> Log.log_msg "get_ether_list: ifreq_list != Unix.ADDR...\n";
	    raise Exit) in
      let ifreq_list = Unix.siocgifconf s in
      get_addr ifreq_list [] *) []

(* must do at least one bind_in before calling *)
let bind_any_local_addr () =
  let _ = start_socket () in
  let rec try_to_bind lst =
    match lst with
      [] -> raise No_interface
    | (d,a)::t ->
	(try bind_addr a
	with e -> try_to_bind t) in
  let ether_list = get_ether_list () in
  try_to_bind ether_list

let debug_dmp str =
  let rec dump_char n =
    if n = String.length str then Printf.printf "\n"
    else begin Printf.printf "%02x " (Char.code (String.get str n));
      dump_char (n + 1) end
  in
  dump_char 0

let get_hw_addr port =
  let IPORT(dev, _, _, _) = port in
  begin match !sfd with
    None -> Log.log_msg "status_in: !sfd = None\n"; raise Exit
  | Some(s) ->
      Unix.siocgifhwaddr s dev
  end

(*let status_in port req =
  match req with
    GETHWADDR ->
      let IPORT(dev, _, _, _) = port in
      begin match !sfd with
	None -> Log.log_msg "status_in: !sfd = None\n"; raise Exit
      |	Some(s) ->
	  let c = String.get dev 0 in
	  if c >= 'a' & c <= 'z' then begin (* assume devs start with
					       lowercase alpha *)
	    let ifreq = String.make 32 '\000' in
	    String.blit dev 0 ifreq 0 (String.length dev);
	    if Unix.ioctl_ptr s 0x8927 ifreq = -1 then []
	    else [(dev, String.sub ifreq 18 6)] end
	  else get_ether_list ()
      end

  | GETIPADDR ->
      let hostname = Unix.gethostname () in
      Hashtbl.find Ipconfig.ipconfig_hash hostname*)

(* Output ports *)
let out_dev_table = Hashtbl.create 5
let out_dev_table_mutex = Mutex.create ()

let bind_out dev =
  Mutex.lock demux_dev_tbl_mutex;
  start_socket ();
  Mutex.unlock demux_dev_tbl_mutex;
  Mutex.lock out_dev_table_mutex;
(*  (try Hashtbl.find out_dev_table dev
    with Not_found ->
      (Mutex.unlock out_dev_table_mutex; raise Already_bound));*)
  Hashtbl.add out_dev_table dev ();
  Mutex.unlock out_dev_table_mutex;
  OPORT(dev)
    
let send_pkt_out oport pkt ofs len addr =
  let OPORT(dev) = oport in
  match !sfd with
    None -> Log.log_msg "send_pkt_out: !sfd = None\n"; raise Exit
  | Some(s) -> Unix.sendto s pkt ofs len [] addr

let unbind_out oport =
  let OPORT(dev) = oport in
  Mutex.lock out_dev_table_mutex;
  Hashtbl.remove out_dev_table dev;
  Mutex.unlock out_dev_table_mutex

let get_oport () =
  let _ = start_socket () in
  match !sfd with
    None -> Log.log_msg "get_oport: !sfd = None\n"; raise Exit
  | Some(s) ->
      let ifreq_list = Unix.siocgifconf s in
      let rec find_if ifreq_list =
      	match ifreq_list with
      	  [] -> raise No_interface
	| (Unix.ADDR(dev, _)) :: t ->
	    (try bind_out dev
	     with e -> find_if t)
	| _ -> Log.log_msg "get_oport: ifreq_list != Unix.ADDR...\n";
	    raise Exit
  in
  find_if ifreq_list

(* should do something with select(), but for now, this should work *)
let ready_to_send_p_out oport = true

(*let status_out port req =
  match req with
    GETHWADDR ->
      let ifreq = String.make 32 '\000' in
      let OPORT(dev) = port in
      begin match !sfd with
	None -> Log.log_msg "status_out: !sfd = None\n"; raise Exit
      |	Some(s) ->
	  let rec get_addr ifreq_list addr_list =
	    (match ifreq_list with
	      [] -> addr_list
	    | (Unix.ADDR(dev, _)) :: t ->
		let ifr = get_ifreq s dev in
		begin match ifr with
		  Unix.FLAGS(_, flags) ->
		    if find_flag Unix.IFF_LOOPBACK flags
		    then get_addr t addr_list
		    else begin
	  	      String.blit dev 0 ifreq 0 (String.length dev);
		      let retval = Unix.ioctl_ptr s 0x8927 ifreq in
		      if retval = -1 then get_addr t addr_list
		      else get_addr t ((dev, String.sub ifreq 18 6) :: addr_list)
		    end
		| _ -> Log.log_msg "status_out: ifr != Unix.FLAGS\n";
		    raise Exit
		end
	    | _ -> Log.log_msg "status_out: ifreq_list != Unix.ADDR...\n";
		raise Exit) in
	  let c = String.get dev 0 in
	  if c >= 'a' & c <= 'z' then begin (* assume devs start with
					       lowercase alpha *)
	    String.blit dev 0 ifreq 0 (String.length dev);
	    if Unix.ioctl_ptr s 0x8927 ifreq = -1 then []
	    else [(dev, String.sub ifreq 18 6)] end
	  else begin
	    let ifreq_list = Unix.siocgifconf s in
	    get_addr ifreq_list []
	  end end

  | GETIPADDR ->
      let hostname = Unix.gethostname () in
      Hashtbl.find Ipconfig.ipconfig_hash hostname*)

(* Generic functions *)
let iport_to_oport iport =
  let IPORT(dev, _, _, _) = iport in
  OPORT(dev)

(* Debugging aids *)
let debug_iport_to_string iport =
  let IPORT(dev, _, _, _) = iport in
  dev

let debug_oport_to_string oport =
  let OPORT(dev) = oport in
  dev

let debug_demux_num_devs () = !demux_num_devs

end

