open Safestd
open Safeunix
open Unixnet
open Safethread
open Log
open Ip
open Udp

let tftp_port = 69

let opcode_rrq = 1
let opcode_wrq = 2
let opcode_data = 3
let opcode_ack = 4
let opcode_error = 5

type tftp_pkt = RRQ of udp_pkt * string * string
  | WRQ of udp_pkt * string * string
  | DATA of udp_pkt * int * string
  | ACK of udp_pkt * int
  | ERROR of udp_pkt * int * string

let cur_tid = ref 2048
let rec next_tid src_ip dst_ip src_tid =
  try
    incr cur_tid;
    (bind_udp_conn (src_ip, dst_ip, src_tid, !cur_tid), !cur_tid)
  with Already_bound -> next_tid src_ip dst_ip src_tid

let err_no_error = 0
let err_file_not_found = 1
let err_access_violation = 2
let err_disk_full = 3
let err_illegal_tftp_op = 4
let err_unknown_tid = 5
let err_file_exists = 6
let err_unknown_user = 7

let err_msgs = [|
  "Not defined, see error message (if any).";
  "File not found.";
  "Access violation.";
  "Disk full or allocation exceeded.";
  "Illegal TFTP operation.";
  "Unknown transfer ID.";
  "File already exists.";
  "No such user." |]

let value = ref None
let hashval() =
  match !value with
    None -> failwith "hash value not set!  (This can't happen.)"
  | Some(v) -> v

let get_int16 s ofs = (Char.code (String.get s ofs) lsl 8)
  lor (Char.code (String.get s (ofs + 1)))
let from_int16 i =
  try
    Printf.sprintf "%c%c"
      (Char.chr (i lsr 8)) (Char.chr (i land 0xff))
  with Invalid_argument(_) ->
    begin
      Printf.printf "from_int16 screwup: i == %x\n" i;
      raise Exit
    end

let rec get_c_str_len s ofs =
  if String.get s ofs = '\000' then 0
  else 1 + get_c_str_len s (ofs + 1)

let udp_to_tftp upkt =
  match get_int16 upkt.udp_data 0 with
    1 when 1 = opcode_rrq ->
      let filename_len = get_c_str_len upkt.udp_data 2 in
      let mode_len = get_c_str_len upkt.udp_data (2 + filename_len + 1) in
      RRQ (upkt,
	   String.sub upkt.udp_data 2 filename_len,
	   String.sub upkt.udp_data (2 + filename_len + 1) mode_len)
  | 2 when 2 = opcode_wrq ->
      let filename_len = get_c_str_len upkt.udp_data 2 in
      let mode_len = get_c_str_len upkt.udp_data (2 + filename_len + 1) in
      WRQ (upkt,
	   String.sub upkt.udp_data 2 filename_len,
	   String.sub upkt.udp_data (2 + filename_len + 1) mode_len)
  | 3 when 3 = opcode_data ->
      DATA (upkt, get_int16 upkt.udp_data 2,
	    String.sub upkt.udp_data 4 (String.length upkt.udp_data - 4))
  | 4 when 4 = opcode_ack -> ACK (upkt, get_int16 upkt.udp_data 2)
  | 5 when 5 = opcode_error ->
      let msg_len = get_c_str_len upkt.udp_data 4 in
      ERROR (upkt, get_int16 upkt.udp_data 2,
	     String.sub upkt.udp_data 4 msg_len)
  | _ -> raise Exit

let send_error err port pkt =
  reply_udp port
    { udp_hdr = {
      src_port = pkt.udp_hdr.src_port;
      dst_port =
      if pkt.udp_hdr.dst_port = 69 then 1091 else pkt.udp_hdr.dst_port;
      length = 8 + 5 + (String.length err_msgs.(err));(*UDP hdr + opcode/errcode + err msg*)
      checksum = 0 };
      udp_data = Printf.sprintf "%s%s%s\000"
	(from_int16 opcode_error) (from_int16 err) err_msgs.(err);
      udp_iphdr = pkt.udp_iphdr;
      udp_ehdr = pkt.udp_ehdr;
      udp_uaddr = pkt.udp_uaddr
    }; ()

let send_ack blkno port tid pkt =
  reply_udp port
    { udp_hdr = {
      src_port = pkt.udp_hdr.src_port;
      dst_port = tid;
      length = 8 + 4;			(*UDP hdr + opcode/blk no*)
      checksum = 0 };
      udp_data = Printf.sprintf "%s%s" (from_int16 opcode_ack) (from_int16 blkno);
      udp_iphdr = pkt.udp_iphdr;
      udp_ehdr = pkt.udp_ehdr;
      udp_uaddr = pkt.udp_uaddr
    }; ()

let complete_connection conn tid maxblk=
  let timeout tim =
    Safethread.delay 60.0 in
  let rec handle_stragglers () =
    let pkt = get_packet conn in
    let tpkt = udp_to_tftp pkt in
    match tpkt with
      WRQ(_) | RRQ(_) -> send_error err_illegal_tftp_op conn pkt
    | ACK(_) | ERROR(_)-> ()
    | DATA(_, blkno, _) ->
	if blkno > maxblk then send_error err_illegal_tftp_op conn pkt
	else send_ack blkno conn tid pkt in
  let stragglers = Safethread.create (hashval()) handle_stragglers () in
  let timer = Safethread.create (hashval()) handle_stragglers () in
  Safethread.join timer;
  Safethread.kill stragglers

let handle_data conn tid =
  let rec aux old_data old_blk =
    let pkt = get_packet conn in
    let tpkt = udp_to_tftp pkt in
    match tpkt with
      WRQ(_) | RRQ(_) | ACK(_) -> send_error err_illegal_tftp_op conn pkt; 0
    | ERROR(_) -> 0
    | DATA(_, blkno, dat) ->
	if blkno <= old_blk then begin
	  send_ack blkno conn tid pkt;
	  aux old_data old_blk
	end else if blkno = succ old_blk then begin
	  send_ack blkno conn tid pkt;
	  if pkt.udp_hdr.length = 524 then(*512 bytes of data?*)
	    aux (old_data ^ dat) blkno
	  else begin
	    Func.do_op (Func.LOADS(old_data ^ dat));
	    let s = Printf.sprintf "got file: len %d\n"
		(String.length (old_data ^ dat)) in
	    Log.log_msg s;
	    blkno
	  end
	end
	else begin send_error err_illegal_tftp_op conn pkt; 0 end in
  aux "" 0

let handle_wrq port upkt fname mode =
  let (conn, tid) =
    next_tid upkt.udp_iphdr.source_addr upkt.udp_iphdr.dest_addr upkt.udp_hdr.src_port in
  if (String.lowercase mode) <> "octet" then begin
    send_error err_access_violation conn upkt;
    complete_connection conn tid 0
  end else begin
    send_ack 0 conn tid upkt;
    let max_blk = handle_data conn tid in
    complete_connection conn tid max_blk
  end

let tftp (port, pkt) =
  let tpkt = udp_to_tftp pkt in
  match tpkt with
    WRQ(upkt, filename, mode) -> handle_wrq port upkt filename mode
  | RRQ(upkt, filename, mode) -> send_error err_file_not_found port pkt
  | ACK(_) | ERROR(_) -> ()
  | DATA(_) -> send_error err_unknown_tid port pkt

let tftpd ignored =
  begin
    try
      let udp_port = listen_udp_port tftp_port in
      try
      	while true do
	  let pkt = get_packet udp_port in
	  Safethread.create (hashval()) tftp (udp_port, pkt)
    	done
      with e -> (unbind_udp udp_port; raise e)
    with Already_bound -> Printf.printf "UDP port already in use\n"
  end;
  ""

let _ = value := Some(Func.register "tftpd" tftpd)
