open Safestd
open Safeunix
open Safethread
open Log
open Lower
open Udpnet
open Udpsock

module Udp = UdpnetFUNCTOR(Udpsock)

let tftp_port = 69

let opcode_rrq = 1
let opcode_wrq = 2
let opcode_data = 3
let opcode_ack = 4
let opcode_error = 5

type tftp_pkt = RRQ of string * string
  | WRQ of string * string
  | DATA of int * string
  | ACK of int
  | ERROR of int * string

let cur_tid = ref 2048
let rec next_tid his_ip my_ip his_tid =
(*Printf.printf "next_tid\n"; Pervasives.flush Pervasives.stdout;*)
  try
    incr cur_tid;
(*Printf.printf "\t!cur_tid -> %d\n" (!cur_tid); Pervasives.flush Pervasives.stdout;*)
    let conn = Udp.bind_udp_conn (my_ip, his_ip, !cur_tid, his_tid) in
(*Printf.printf "\tsuccess!\n"; Pervasives.flush Pervasives.stdout;*)
    (conn, !cur_tid)
  with Udp.Already_bound -> begin
(*    Printf.printf "try again\n"; Pervasives.flush Pervasives.stdout;*)
    next_tid his_ip my_ip his_tid end

let err_no_error = 0
let err_file_not_found = 1
let err_access_violation = 2
let err_disk_full = 3
let err_illegal_tftp_op = 4
let err_unknown_tid = 5
let err_file_exists = 6
let err_unknown_user = 7

let err_msgs = [|
  "Not defined, see error message (if any).";
  "File not found.";
  "Access violation.";
  "Disk full or allocation exceeded.";
  "Illegal TFTP operation.";
  "Unknown transfer ID.";
  "File already exists.";
  "No such user." |]

let value = ref None
let hashval() =
  match !value with
    None -> failwith "hash value not set!  (This can't happen.)"
  | Some(v) -> v

let get_int16 s ofs = (Char.code (String.get s ofs) lsl 8)
  lor (Char.code (String.get s (ofs + 1)))
let from_int16 i =
  try
    Printf.sprintf "%c%c"
      (Char.chr (i lsr 8)) (Char.chr (i land 0xff))
  with Invalid_argument(_) ->
    begin
      Printf.printf "from_int16 screwup: i == %x\n" i;
      raise Exit
    end

let rec get_c_str_len s ofs =
  if String.get s ofs = '\000' then 0
  else 1 + get_c_str_len s (ofs + 1)

let udp_to_tftp len data =
  match get_int16 data 0 with
    1 when 1 = opcode_rrq ->
      let filename_len = get_c_str_len data 2 in
      let mode_len = get_c_str_len data (2 + filename_len + 1) in
      RRQ (String.sub data 2 filename_len,
	   String.sub data (2 + filename_len + 1) mode_len)
  | 2 when 2 = opcode_wrq ->
      let filename_len = get_c_str_len data 2 in
      let mode_len = get_c_str_len data (2 + filename_len + 1) in
      WRQ (String.sub data 2 filename_len,
	   String.sub data (2 + filename_len + 1) mode_len)
  | 3 when 3 = opcode_data ->
      DATA (get_int16 data 2, String.sub data 4 (len - 4))
  | 4 when 4 = opcode_ack -> ACK (get_int16 data 2)
  | 5 when 5 = opcode_error ->
      let msg_len = get_c_str_len data 4 in
      ERROR (get_int16 data 2, String.sub data 4 msg_len)
  | _ -> raise Exit

let send_error err port hdr =
  let outpkt = Printf.sprintf "%s%s%s\000"
      (from_int16 opcode_error) (from_int16 err) err_msgs.(err) in
  Udp.sendto_udp port outpkt (Udp.get_src_addr hdr) (Udp.get_src_port hdr);
  ()

let send_ack blkno port =
  let outpkt = Printf.sprintf "%s%s"
      (from_int16 opcode_ack) (from_int16 blkno) in
  Udp.send_udp port outpkt;
  ()

let complete_connection conn maxblk=
  let timeout tim = Safethread.delay tim in
  let rec handle_stragglers () =
    let (hdr, len, pkt) = Udp.get_packet conn in
    let tpkt = udp_to_tftp len pkt in
    match tpkt with
      WRQ(_) | RRQ(_) -> send_error err_illegal_tftp_op conn hdr
    | ACK(_) | ERROR(_)-> ()
    | DATA(blkno, _) ->
	if blkno > maxblk then send_error err_illegal_tftp_op conn hdr
	else send_ack blkno conn in
  let stragglers = Safethread.create (hashval()) handle_stragglers () in
  let timer = Safethread.create (hashval()) timeout 60.0 in
  Safethread.join timer;
  Safethread.kill stragglers

let handle_data conn =
  let rec aux old_data old_blk =
    let (hdr, len, pkt) = Udp.get_packet conn in
    let tpkt = udp_to_tftp len pkt in
    match tpkt with
      WRQ(_) | RRQ(_) | ACK(_) -> send_error err_illegal_tftp_op conn hdr; 0
    | ERROR(_) -> 0
    | DATA(blkno, dat) ->
	if blkno <= old_blk then begin
	  send_ack blkno conn;
	  aux old_data old_blk
	end else if blkno = succ old_blk then begin
	  send_ack blkno conn;
	  if String.length dat = 512 then(*512 bytes of data?*)
	    aux (old_data ^ dat) blkno
	  else begin
	    Func.do_op (Func.LOADS(old_data ^ dat));
	    let s = Printf.sprintf "got file: len %d\n"
		(String.length (old_data ^ dat)) in
	    Log.log_msg s;
	    blkno
	  end
	end
	else begin send_error err_illegal_tftp_op conn hdr; 0 end in
  aux "" 0

let handle_wrq port hdr fname mode =
(*Printf.printf "in handle_wrq\n";
Pervasives.flush Pervasives.stdout;*)
  let src_addr = Udp.get_src_addr hdr in
  let dst_addr = Udp.get_dst_addr hdr in
  let src_port = Udp.get_src_port hdr in
  let (conn, _) = next_tid src_addr dst_addr src_port in
  if (String.lowercase mode) <> "octet" then begin
(*Printf.printf "about to send_error\n";
Pervasives.flush Pervasives.stdout;*)
    send_error err_access_violation conn hdr;
    complete_connection conn 0
  end else begin
(*Printf.printf "about to send_ack\n";
Pervasives.flush Pervasives.stdout;*)
    send_ack 0 conn;
(*Printf.printf "about to handle_data\n";
Pervasives.flush Pervasives.stdout;*)
    let max_blk = handle_data conn in
(*Printf.printf "about to complete_connection\n";
Pervasives.flush Pervasives.stdout;*)
    complete_connection conn max_blk
  end

let tftp (port, hdr, len, pkt) =
  let tpkt = udp_to_tftp len pkt in
  match tpkt with
    WRQ(filename, mode) -> handle_wrq port hdr filename mode
  | RRQ(filename, mode) -> send_error err_file_not_found port hdr
  | ACK(_) | ERROR(_) -> ()
  | DATA(_) -> send_error err_unknown_tid port hdr

let tftpd ignored =
  begin
    try
      let udp_port = Udp.listen_udp_port tftp_port in
      try
      	while true do
	  let (hdr, len, pkt) = Udp.get_packet udp_port in
	  Safethread.create (hashval()) tftp (udp_port, hdr, len, pkt)
    	done
      with e -> (Udp.unbind_udp udp_port; raise e)
    with Udp.Already_bound -> Printf.printf "UDP port already in use\n"
  end;
  ""

let _ = value := Some(Func.register "tftpd" tftpd)
