open Safestd
open Safeunix
open Unixnet
open Safethread
open Log

module UdpFUNCTOR =
  functor (Lower : IP) ->
    struct
      type udp_port =
	  UDP_LISTEN_PORT of int * ip_pkt Queue.t * Mutex.t * Condition.t
  	| UDP_CONN_PORT of
	    (IPaddr.ip_addr * IPaddr.ip_addr * int * int)
	      * ip_pkt Queue.t * Mutex.t * Condition.t

      type udp_hdr =
	  { src_port : int;
	    dst_port : int;
	    length : int;
	    checksum : int
	  }

      type hdr = { udp_hdr : udp_hdr; lower_hdr : Lower.hdr }

      exception Already_bound

      let udp_proto = 17

      let value = ref None
      let hashval() =
  	match !value with
	  None -> failwith "hash value not set!  (This can't happen.)"
  	| Some(v) -> v

      let die = ref false
      let die_mutex = Mutex.create ()
      let die_condition = Condition.create ()

      let udp_conn_hashtable = Hashtbl.create 257
      let udp_conn_hash_m = Mutex.create ()

      let udp_listen_hashtable = Hashtbl.create 257
      let udp_listen_hash_m = Mutex.create ()

      let iport = ref None

      let get_int16 s ofs = (Char.code (String.get s ofs) lsl 8)
	  lor (Char.code (String.get s (ofs + 1)))
      let from_int16 i =
  	try
	  Printf.sprintf "%c%c"
	    (Char.chr (i lsr 8)) (Char.chr (i land 0xff))
  	with Invalid_argument(_) ->
	  begin
	    Printf.printf "from_int16 screwup: i == %x\n" i;
	    raise Exit
	  end

      let get_src_port s = get_int16 s 0
      let get_dst_port s = get_int16 s 2
      let get_length s = get_int16 s 4
      let get_checksum s = get_int16 s 6

      let to_udp_hdr s =
  	{ src_port = get_src_port s;
	  dst_port = get_dst_port s;
	  length = get_length s;
	  checksum = get_checksum s
  	}

      let inet_addr_to_bin ia =
	let str = Safeunix.string_of_inet_addr ia in
	let l = List.map int_of_string (Str.split (Str.regexp "\.") str) in
	match l with
	  [a; b; c; d] ->
	    Printf.sprintf "%c%c%c%c"
	      (Char.code a) (Char.code b) (Char.code c) (Char.code d)
	| _ -> raise (Failure "inet_addr_to_bin")

      let from_udp_hdr (his_addr, my_addr, his_port, my_port) data =
	let csum_pseudo () =
	  let len_str = from_int16 ((String.length data) + 8) in
	  let ph = Printf.sprintf "%s%s%c%c%s"
	      (inet_addr_to_bin my_addr)
	      (inet_addr_to_bin his_addr)
	      (Char.code 0)
	      (Char.code udp_proto)
	      len_str in
	  Ipcsum.ipcsum_step 0 ph 0 (String.length ph) in
   	let hdr = Printf.sprintf "%s%s%s%s"
	    (from_int16 my_port) (from_int16 his_port)
	    (from_int16 (String.length data)) (from_int16 0) in
	let step1 =
	  Ipcsum.ipcsum_step (csum_pseudo ()) hdr 0 (String.length hdr) in
	let step2 =
	  Ipcsum.ipcsum_step step1 data 0 (String.length data) in
	let csum_str = (from_int16 (Ipcsum.ipcsum_complete step2) in
	Printf.sprintf "%s%s%s%s%s"
	  (from_int16 my_port) (from_int16 his_port)
	  (from_int16 (String.length data)) csum_str data

      let build_reply_udp_hdr hdr =
  	{ src_port = hdr.dst_port;
	  dst_port = hdr.src_port;
	  length = hdr.length;
	  checksum = 0
  	}

      let lower_to_udp (ihdr, len, udpstr) =
  	let udp_hdr = to_udp_hdr udpstr in
  	let hdr_bytes = 8 in
	({ udp_hdr = udp_hdr;
	   lower_hdr = ihdr },
	 len - hdr_bytes,
	 String.sub udpstr hdr_bytes (len - hdr_bytes))

      let udp_to_lower conn data =
	let (his_addr, my_addr, his_port, my_port) = conn in
	let s = Printf.sprintf "%s%s" (from_udp_hdr conn data) data in
	s

      let udp_to_lower_reply upkt =
	(upkt.lower_hdr,
	 Printf.sprintf "%s%s"
	   (from_udp_hdr (build_reply_udp_hdr upkt.udp_hdr)) upkt.udp_data)

      let handle_udp (iport, start_m) =
  	let checksum hdr = true in
  	Mutex.lock start_m;
  	Mutex.unlock start_m;

  	while not !die do
	  let ipkt as (ihdr, len, udp_str) = Lower.get_packet iport in
	  let hdr = to_udp_hdr udp_str in
	  if (*(String.length ipkt.ip_data) = hdr.length
	       & *)checksum udp_str
	  then begin
	    let try_lookup m table key func pkt =
	      Mutex.lock m;
	      try
	  	let conn = Hashtbl.find table key in
	  	Mutex.unlock m;
	  	func conn pkt;
	  	true
	      with Not_found -> (Mutex.unlock m; false) in
	    let requeue (q, m, c) pkt =
	      Mutex.lock m;
	      Queue.add pkt q;
	      Condition.signal c;
	      Mutex.unlock m in
	    if not (try_lookup udp_conn_hash_m udp_conn_hashtable
		      (Lower.get_source ihdr, Lower.get_dest ihdr,
		       hdr.src_port, hdr.dst_port) requeue ipkt) then begin
			 try_lookup udp_listen_hash_m udp_listen_hashtable
			   hdr.dst_port requeue ipkt;
			 ()
		       end
	  end else
	    let s = Printf.sprintf
		"Discarded UDP packet: ilen %d udplen %d\n"
 		(String.length udp_str) hdr.length in
	    log_msg s; ()
  	done

      let listen_udp_port port =
  	try
	  Mutex.lock udp_listen_hash_m;
	  Hashtbl.find udp_listen_hashtable port;
	  Mutex.unlock udp_listen_hash_m;
	  raise Already_bound
  	with Not_found -> begin
	  let q = Queue.create () in
	  let m = Mutex.create () in
	  let c = Condition.create () in
	  Hashtbl.add udp_listen_hashtable port (q, m, c);
	  Mutex.unlock udp_listen_hash_m;
	  UDP_LISTEN_PORT(port, q, m, c)
  	end
  	| e -> begin
	    Mutex.unlock udp_listen_hash_m;
	    raise e
  	end

      let bind_udp_conn conn =
  	try
	  Mutex.lock udp_conn_hash_m;
	  Hashtbl.find udp_conn_hashtable conn;
	  Mutex.unlock udp_conn_hash_m;
	  raise Already_bound
   	with Not_found -> begin
	  let q = Queue.create () in
	  let m = Mutex.create () in
	  let c = Condition.create () in
	  Hashtbl.add udp_conn_hashtable conn (q, m, c);
	  Mutex.unlock udp_conn_hash_m;
	  UDP_CONN_PORT(conn, q, m, c)
   	end
   	| e -> begin
	    Mutex.unlock udp_conn_hash_m;
	    raise e
   	end

      let unbind_udp port =
   	match port with
	  UDP_CONN_PORT(conn, _, _, _) ->
	    begin
 	      Mutex.lock udp_conn_hash_m;
 	      Hashtbl.remove udp_conn_hashtable conn;
 	      Mutex.unlock udp_conn_hash_m
	    end
   	| UDP_LISTEN_PORT(dst, _, _, _) ->
	    begin
       	      Mutex.lock udp_listen_hash_m;
       	      Hashtbl.remove udp_listen_hashtable dst;
       	      Mutex.unlock udp_listen_hash_m
	    end

      let get_packet uport =
   	match uport with
	  UDP_CONN_PORT(_, q, m, c) ->
	    begin
 	      Mutex.lock m;
 	      while (Queue.length q = 0) do
 	  	Condition.wait c m
 	      done;
 	      let pkt = Queue.take q in
 	      Mutex.unlock m;
 	      lower_to_udp pkt
	    end
   	| UDP_LISTEN_PORT(_, q, m, c) ->
	    begin
 	      Mutex.lock m;
 	      while (Queue.length q = 0) do
 	  	Condition.wait c m
 	      done;
 	      let pkt = Queue.take q in
 	      Mutex.unlock m;
 	      lower_to_udp pkt
	    end

(*      let send_packet uport pkt ofs len addr =
   	match !iport with
	  None -> raise Exit
   	| Some(prt) -> Ip.send_packet prt pkt ofs len addr*)

      let send_udp uport udata =
   	match !iport with
	  None -> raise Exit
   	| Some(prt) -> begin
	    match uport with
	      UDP_LISTEN_PORT(_) -> raise (Invalid_argument "send_udp")
	    | UDP_CONN_PORT(conn as (his_addr, my_addr, his_port, my_port),
			    _, _, _) ->
	    	let (lower_hdr, payload) = udp_to_lower conn udata in
	    	Lower.send prt his_addr my_addr payload
	end

      let reply_udp uport upkt =
   	match !iport with
	  None -> raise Exit
   	| Some(prt) ->
	    let (lower_hdr, payload) = udp_to_lower_reply upkt in
	    Lower.reply prt lower_hdr payload

      let top_thread () =
   	let start_m = Mutex.create () in
   	Mutex.lock start_m;
   	let ip_port = Lower.bind_proto udp_proto in
   	iport := Some(ip_port);
   	let tid = Safethread.create (hashval()) handle_udp (ip_port, start_m) in
   	Mutex.unlock start_m;
   	Mutex.lock die_mutex;
   	while not !die do
	  Condition.wait die_condition die_mutex
   	done;
   	Safethread.kill tid;
   	Lower.unbind_proto ip_port

      let udp ignored =
   	let t = Safethread.create (hashval()) top_thread () in
   	Safethread.join t;
   	""

      let _ = value := Some(Func.register "udp" udp)
    end
      
