open Safeloader
open Safeunix
open Udpnet
open Sha1

module Udp = Udpnet

let value = ref None
let hashval() =
  match !value with
    None -> failwith "hash value not set!  (This can't happen.)"
  | Some(v) -> v

let uport_val = ref None
let uport () =
  match !uport_val with
    None -> raise (Failure "uport")
  | Some(p) -> p

let default_anep_port = ref 3322

type wf_hdr = { do_forward : bool;
		type_id : int
	      }

type eaddr = int * int
type addr_opt = IPv4 of Safeunix.inet_addr | IPv6 of string | Ethernet of eaddr

type integrity_opt = unit

type nn_auth_opt = SPKI of string | X509 of string

type auth_opt = HMAC of int

type opt =
    NN_AUTH of nn_auth_opt
  | INTEGRITY of integrity_opt
  | SOURCE of addr_opt
  | DEST of addr_opt
  | AUTH of auth_opt

type option = { priv : bool;
		discard : bool;
		option : opt
	      }

let source_id_opt = 1
let dest_id_opt = 2
let integrity_opt = 3
let nn_auth_opt = 4
let auth_opt = 5

let die = ref false

let conn_hash = Hashtbl.create 255
let conn_hash_m = Mutex.create ()

let spi_send_hash = Hashtbl.create 255
let spi_send_hash_m = Mutex.create ()
let spi_recv_hash = Hashtbl.create 255
let spi_recv_hash_m = Mutex.create ()

exception No_option
exception Unimplemented
exception Auth_failure

let from_int8 i = String.make 1 (Char.chr (i land 0xff))

let from_int16 i =
  Printf.sprintf "%c%c"
    (Char.chr ((i lsr 8) land 0xff)) (Char.chr (i land 0xff))

let from_int32 i =
  Printf.sprintf "%c%c%c%c"
    (Char.chr ((i lsr 24) land 0xff))
    (Char.chr ((i lsr 16) land 0xff))
    (Char.chr ((i lsr 8) land 0xff))
    (Char.chr (i land 0xff))

let to_byte c = Char.code c

let to_int8 s ofs = to_byte s.[ofs]

let to_int16 s ofs = ((to_byte s.[ofs]) lsl 8) lor (to_byte s.[ofs+1])

let to_int24 s ofs = 
  ((to_byte s.[ofs+0]) lsl 16) lor
  ((to_byte s.[ofs+1]) lsl 8) lor
  (to_byte s.[ofs+2])

let to_int32 s ofs =
  ((to_byte s.[ofs+0]) lsl 24) lor
  ((to_byte s.[ofs+1]) lsl 16) lor
  ((to_byte s.[ofs+2]) lsl 8) lor
  (to_byte s.[ofs+3])

let rec find_integrity_option opts =
  match opts with
    [] -> raise No_option
  | head :: tail ->
      if head.option = (INTEGRITY()) then head
      else find_integrity_option tail

let proc_flags opt =
  let priv = if opt.priv then 0x8000 else 0 in
  let discard = if opt.discard then 0x4000 else 0 in
  priv lor discard

let rec gen_pad i = if i = 0 then "" else "\000" ^ (gen_pad (i - 1))

let spki_val = 1
let x509_val = 2

let process_nn_auth opt auth =
  let type_val = (proc_flags opt) lor nn_auth_opt in
  match auth with
    SPKI(str) ->
      let len = 2 + ((String.length str) + 3) / 4 in
      Printf.sprintf "%s%s%s%s"
	(from_int16 type_val)
	(from_int16 len)
	(from_int16 spki_val)
	str
  | X509(str) ->
      let len = 2 + ((String.length str) + 3) / 4 in
      Printf.sprintf "%s%s%s%s"
	(from_int16 type_val)
	(from_int16 len)
	(from_int32 x509_val)
	str

let hmac_val = 1

let process_auth opt auth data =
  let type_val = (proc_flags opt) lor auth_opt in
  match auth with
    HMAC(send_spi) ->
      Mutex.lock spi_send_hash_m;
      try
	let (c1, c2, cntr, ath) = Hashtbl.find spi_send_hash send_spi in
	let cnt = !cntr in
	incr cntr;
	Mutex.unlock spi_send_hash_m;
	let spi_str = from_int32 send_spi in
	let cnt_str = from_int32 cnt in
	(* make a copy of the contexts so that we don't change them. *)
	let ctx1 = ctx_copy c1 in
	let ctx2 = ctx_copy c2 in
	sha1update ctx1 spi_str;
	sha1update ctx1 cnt_str;
	sha1update ctx1 data;
(*	let tmp = String.make 20 ' ' in
	sha1final tmp ctx1;*)
	let tmp = sha1final ctx1 in
	sha1update ctx2 tmp;
(*	let hmac = String.make 20 ' ' in
	sha1final hmac ctx2;*)
	let hmac = sha1final ctx2 in
	Printf.sprintf "%s%s%s%s%s%s"
	  (from_int16 type_val)
	  (from_int16 7)
	  (from_int32 hmac_val)
	  spi_str
	  cnt_str
	  (String.sub hmac 0 12)
      with Not_found -> invalid_arg "bad SPI"

let dot_regexp = Str.regexp "\."

let from_unix addr =
  let str = Safeunix.string_of_inet_addr addr in
  (*let l = List.map int_of_string (Str.split (Str.regexp "\.") str) in*)
  let l = List.map int_of_string (Str.split dot_regexp str) in
  match l with
    [a; b; c; d] ->
      Printf.sprintf "%c%c%c%c"
 	(Char.chr a) (Char.chr b) (Char.chr c) (Char.chr d) 
  | _ -> raise (Failure "from_unix")
let to_unix str ofs =
  Safeunix.inet_addr_of_string
    (Printf.sprintf "%d.%d.%d.%d"
       (Char.code str.[ofs]) (Char.code str.[1+ofs])
       (Char.code str.[2+ofs]) (Char.code str.[3+ofs]))

let ipv4_scheme = from_int32 1
let ipv4_len = from_int16 3
let ipv6_scheme = from_int32 2
let ipv6_len = from_int16 6
let ethernet_scheme = from_int32 3
let ethernet_len = from_int16 4

let process_addr opt_type opt addr =
  let from_ethernet (high, low) =
    Printf.sprintf "%c%c%c%c%c%c"
      (Char.chr (high lsr 16))
      (Char.chr ((high lsr 8) land 0xff))
      (Char.chr (high land 0xff))
      (Char.chr (low lsr 16))
      (Char.chr ((low lsr 8) land 0xff))
      (Char.chr (low land 0xff)) in
  let type_val = from_int16 ((proc_flags opt) lor opt_type) in
  match addr with
    IPv4(addr) ->
      Printf.sprintf "%s%s%s%s"	type_val ipv4_len ipv4_scheme (from_unix addr)
  | IPv6(addr_str) ->
      Printf.sprintf "%s%s%s%s"	type_val ipv6_len ipv6_scheme addr_str
  | Ethernet(eaddr) ->
      Printf.sprintf "%s%s%s%s%s"
	type_val ethernet_len ethernet_scheme (from_ethernet eaddr) (gen_pad 2)

let process_source opt addr = process_addr source_id_opt opt addr
let process_dest opt addr = process_addr dest_id_opt opt addr

let rec process_no_integrity opts data =
  match opts with
    [] -> ("", "")
  | head :: tail ->
      match head.option with
	NN_AUTH(auth) ->
	  let (pre, post) = process_no_integrity tail data in
	  ((process_nn_auth head auth) ^ pre, post)
      |	INTEGRITY() ->
	  let (pre, post) = process_no_integrity tail data in
	  if post <> "" then
	    raise (Failure "process_no_integrity")
	  else ("", pre)
      |	SOURCE(addr) ->
	  let (pre, post) = process_no_integrity tail data in
	  ((process_source head addr) ^ pre, post)
      |	DEST(addr) ->
	  let (pre, post) = process_no_integrity tail data in
	  ((process_dest head addr) ^ pre, post)
      |	AUTH(auth) ->
	  let (pre, post) = process_no_integrity tail data in
	  ((process_auth head auth data) ^ pre, post)

let process_and_integrity hdr_str pre integrity post func_name code data =
  raise (Failure "process_and_integrity")

let from_hdr hdr =
  let proc_hdr_flags hdr =
    if hdr.do_forward then from_int8 0x80 else from_int8 0 in
  let vers = from_int8 1 in
  let flags = proc_hdr_flags hdr in
  let type_id = from_int16 hdr.type_id in
  Printf.sprintf "%s%s%s" vers flags type_id

let process_and_options hdr_str opts func_name code data =
  let func_name_len = from_int32 (String.length func_name) in
  let code_len = from_int32 (String.length code) in
  let data_len = from_int32 (String.length data) in
  let payload = Printf.sprintf "%s%s%s%s%s%s"
      func_name_len code_len data_len func_name code data in
  let (opt_str, _) = process_no_integrity opts payload in
  let hdr_len =
    ((String.length hdr_str) + (String.length opt_str)) / 4 + 1 in
  let pkt_len = (4 * hdr_len) + 12 + (String.length func_name)
      + (String.length code) + (String.length data) in
  let pkt = Printf.sprintf "%s%s%s%s%s"
      hdr_str (from_int16 hdr_len) (from_int16 pkt_len)
      opt_str payload in
  pkt

let register_type_id id =
  Mutex.lock conn_hash_m;
  try
    Hashtbl.find conn_hash id;
    Mutex.unlock conn_hash_m;
    raise (Invalid_argument "register_type_id")
  with Not_found -> begin
    let q = Queue.create () in
    let m = Mutex.create () in
    let c = Condition.create () in
    Hashtbl.add conn_hash id (q, m, c);
    Mutex.unlock conn_hash_m
  end
  | e -> begin
      Mutex.unlock conn_hash_m;
      raise e
  end

let unregister_type_id id =
  Mutex.lock conn_hash_m;
  Hashtbl.remove conn_hash id;
  Mutex.unlock conn_hash_m

let base1 =
  "\036\036\036\036\036\036\036\036\036\036\036\036\036\036\036\036"
  ^ "\036\036\036\036\036\036\036\036\036\036\036\036\036\036\036\036"
  ^ "\036\036\036\036\036\036\036\036\036\036\036\036\036\036\036\036"
  ^ "\036\036\036\036\036\036\036\036\036\036\036\036\036\036\036\036"
let base2 =
  "\092\092\092\092\092\092\092\092\092\092\092\092\092\092\092\092"
  ^ "\092\092\092\092\092\092\092\092\092\092\092\092\092\092\092\092"
  ^ "\092\092\092\092\092\092\092\092\092\092\092\092\092\092\092\092"
  ^ "\092\092\092\092\092\092\092\092\092\092\092\092\092\092\092\092"

let xor_str base key_in =
  let ctx = Sha1.sha1init () in
  Sha1.sha1update ctx key_in;
(*  let key = String.make 20 ' ' in
  Sha1.sha1final key ctx;*)
  let key = Sha1.sha1final ctx in
  let key_len = String.length key in
  let base_len = String.length base in
  if key_len > base_len then (invalid_arg "xor_str: key_len > base_len");
  let res = String.copy base in
  for i = 0 to (base_len / key_len) - 1 do
    for j = 0 to key_len - 1 do
      res.[i * key_len + j] <-
	 Char.chr ((Char.code res.[i * key_len + j]) lxor (Char.code key.[j]))
    done
  done;
  let off = (base_len / key_len) * key_len in
  for i = 0 to (base_len mod key_len) - 1 do
    res.[off + i] <-
       Char.chr ((Char.code res.[off + i]) lxor (Char.code key.[i]))
  done;
  res

let register_spi send_spi send_key recv_spi recv_key authorization =
  let send_ipad = xor_str base1 send_key in
  let send_opad = xor_str base2 send_key in
  let send_ctx1 = Sha1.sha1init () in
  Sha1.sha1update send_ctx1 send_ipad;
  let send_ctx2 = Sha1.sha1init () in
  Sha1.sha1update send_ctx2 send_opad;
  let recv_ipad = xor_str base1 recv_key in
  let recv_opad = xor_str base2 recv_key in
  let recv_ctx1 = Sha1.sha1init () in
  Sha1.sha1update recv_ctx1 recv_ipad;
  let recv_ctx2 = Sha1.sha1init () in
  Sha1.sha1update recv_ctx2 recv_opad;
  Mutex.lock spi_send_hash_m;
  try
    Hashtbl.find spi_send_hash send_spi;
    Mutex.unlock spi_send_hash_m;
    raise (Invalid_argument "register_spi")
  with Not_found -> begin
    Hashtbl.add spi_send_hash send_spi (send_ctx1, send_ctx2, ref 1, "");
    Mutex.unlock spi_send_hash_m;
    Mutex.lock spi_recv_hash_m;
    try
      Hashtbl.find spi_recv_hash recv_spi;
      (* something is badly horked; maybe we can clean up a little *)
      Mutex.unlock spi_recv_hash_m;
      Mutex.lock spi_send_hash_m;
      Hashtbl.remove spi_send_hash send_spi;
      Mutex.unlock spi_send_hash_m;
      raise (Invalid_argument "register_spi");
    with Not_found -> begin
      Hashtbl.add spi_recv_hash recv_spi (recv_ctx1, recv_ctx2, ref 0, authorization);
      Mutex.unlock spi_recv_hash_m
    end
    | e -> begin Mutex.unlock spi_recv_hash_m; raise e end
  end
  | e -> begin
      Mutex.unlock spi_send_hash_m;
      raise e
  end

let unregister_spi send_spi recv_spi =
  Mutex.lock spi_recv_hash_m;
  Hashtbl.remove spi_recv_hash recv_spi;
  Mutex.unlock spi_recv_hash_m;
  Mutex.lock spi_send_hash_m;
  Hashtbl.remove spi_send_hash send_spi;
  Mutex.unlock spi_send_hash_m

let version = from_int8 1

let hdr_to_str hdr =
  let do_flags do_forward =
    if do_forward then from_int8 0x80 else from_int8 0 in
  Printf.sprintf "%s%s%s"
    version (do_flags hdr.do_forward) (from_int16 hdr.type_id)

let send_wf dest hdr opts func_name code data =
  let hdr_str = hdr_to_str hdr in
  begin try
    let integrity_opt = find_integrity_option opts in
    let (pre, post) = process_no_integrity opts data in
    let pkt =
      process_and_integrity hdr_str pre integrity_opt post
 	func_name code data in
    Udp.sendto_udp (uport()) pkt dest !default_anep_port
  with No_option ->
    let pkt = process_and_options hdr_str opts func_name code data in
    Udp.sendto_udp (uport()) pkt dest !default_anep_port
  end;
  ()

let recv_wf typ_id =
  Mutex.lock conn_hash_m;
  try
    let (q, m, c) = Hashtbl.find conn_hash typ_id in
    Mutex.unlock conn_hash_m;
    Mutex.lock m;
    while (Queue.length q = 0) do
      Condition.wait c m
    done;
    let wf = Queue.take q in
    Mutex.unlock m;
    wf
  with Not_found ->
    Mutex.unlock conn_hash_m;
    raise (Invalid_argument "recv_wf")

let get_version pkt = to_int8 pkt 0
let get_flags pkt =
  let flags = to_int8 pkt 1 in
  let do_forward = (flags land 0x80) > 0 in
  let rest = flags land 0x7f in
  (do_forward, rest)
let get_type_id s = to_int16 s 2
let get_hdr_len s = to_int16 s 4
let get_pkt_len s = to_int16 s 6

let get_address s ofs opt_len =
  let scheme = to_int32 s ofs in
  match scheme with
    1 -> IPv4(to_unix s (ofs+4))
  | 2 -> IPv6(String.sub s (ofs+4) 16)
  | 3 -> Ethernet(to_int24 s (ofs+4), to_int24 s (ofs+7))
  | _ -> raise (Invalid_argument "get_address")

let get_nn_auth s ofs opt_len =
  let scheme = to_int32 s ofs in
  match scheme with
    1 -> SPKI(String.sub s ofs ((opt_len - 2) * 4))
  | 2 -> X509(String.sub s ofs ((opt_len - 2) * 4))
  | _ -> raise (Invalid_argument "get_nn_auth")

(* Convert a SHA1 digest to an ASCII string *)
(*let present_digest digest =
  let aa = Printf.sprintf "%02X" (Char.code digest.[0]) in
  let aa = (aa ^ Printf.sprintf "%02X" (Char.code digest.[1])) in
  let aa = (aa ^ (Printf.sprintf "%02X" (Char.code digest.[2]))) in
  let aa = (aa ^ (Printf.sprintf "%02X" (Char.code digest.[3]))) in
  let aa = (aa ^ (Printf.sprintf "%02X" (Char.code digest.[4]))) in
  let aa = (aa ^ (Printf.sprintf "%02X" (Char.code digest.[5]))) in
  let aa = (aa ^ (Printf.sprintf "%02X" (Char.code digest.[6]))) in
  let aa = (aa ^ (Printf.sprintf "%02X" (Char.code digest.[7]))) in
  let aa = (aa ^ (Printf.sprintf "%02X" (Char.code digest.[8]))) in
  let aa = (aa ^ (Printf.sprintf "%02X" (Char.code digest.[9]))) in
  let aa = (aa ^ (Printf.sprintf "%02X" (Char.code digest.[10]))) in
  let aa = (aa ^ (Printf.sprintf "%02X" (Char.code digest.[11]))) in
  let aa = (aa ^ (Printf.sprintf "%02X" (Char.code digest.[12]))) in
  let aa = (aa ^ (Printf.sprintf "%02X" (Char.code digest.[13]))) in
  let aa = (aa ^ (Printf.sprintf "%02X" (Char.code digest.[14]))) in
  let aa = (aa ^ (Printf.sprintf "%02X" (Char.code digest.[15]))) in
  let aa = (aa ^ (Printf.sprintf "%02X" (Char.code digest.[16]))) in
  let aa = (aa ^ (Printf.sprintf "%02X" (Char.code digest.[17]))) in
  let aa = (aa ^ (Printf.sprintf "%02X" (Char.code digest.[18]))) in
  let aa = (aa ^ (Printf.sprintf "%02X" (Char.code digest.[19]))) in
  aa*)
let present_digest digest =
  Printf.sprintf
    "%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X"
    (Char.code digest.[0]) (Char.code digest.[1]) (Char.code digest.[2])
    (Char.code digest.[3]) (Char.code digest.[4]) (Char.code digest.[5])
    (Char.code digest.[6]) (Char.code digest.[7]) (Char.code digest.[8])
    (Char.code digest.[9]) (Char.code digest.[10]) (Char.code digest.[11])
    (Char.code digest.[12]) (Char.code digest.[13]) (Char.code digest.[14])
    (Char.code digest.[15]) (Char.code digest.[16]) (Char.code digest.[17])
    (Char.code digest.[18]) (Char.code digest.[19])

let get_auth s ofs opt_len payload =
  let scheme = to_int32 s ofs in
  match scheme with
    1 -> (* hmac *)
      if opt_len <> 7
      then (Log.log_msg "bad length\n"; raise Auth_failure);
      let recv_spi = to_int32 s (ofs + 4) in
      let cnt = to_int32 s (ofs + 8) in
      let hmac = String.sub s (ofs + 12) 12 in
      Mutex.lock spi_recv_hash_m;
      (try
	let (c1, c2, prev, aut) = Hashtbl.find spi_recv_hash recv_spi in
	if cnt > (!prev) then prev := cnt
	else (Log.log_msg "Replay\n"; raise Auth_failure);
	Mutex.unlock spi_recv_hash_m;
	let spi_str = String.sub s (ofs + 4) 4 in
	let cnt_str = String.sub s (ofs + 8) 4 in
	let ctx1 = ctx_copy c1 in
	let ctx2 = ctx_copy c2 in
	sha1update ctx1 spi_str;
	sha1update ctx1 cnt_str;
	sha1update ctx1 payload;
(*	let tmp = String.make 20 ' ' in
	sha1final tmp ctx1;*)
	let tmp = sha1final ctx1 in
	sha1update ctx2 tmp;
(*	let calc_hmac = String.make 20 ' ' in
	sha1final calc_hmac ctx2;*)
	let calc_hmac = sha1final ctx2 in
	if (String.sub calc_hmac 0 12) <> hmac
	then (Log.log_msg (Printf.sprintf "bad HMAC: %s <> %s\n"
			     calc_hmac hmac); raise Auth_failure)
	else HMAC(recv_spi)
      with Not_found -> Mutex.unlock spi_recv_hash_m;
	Log.log_msg "Unknown SPI\n";
	raise Auth_failure)
  | _ -> (Log.log_msg "Unknown scheme\n"; raise Auth_failure)

let rec get_options opts ofs len payload =
  let get_flags s ofs =
    let x = Char.code s.[ofs] in
    let priv = (x land 0x80) > 0 in
    let action = (x land 0x40) > 0 in
    (priv, action) in
  let get_an_option s ofs =
    let (priv, action) = get_flags s ofs in
    let opt_type = (to_int16 s ofs) land 0x3f in
    let opt_len = to_int16 s (ofs+2) in
    match opt_type with
      1 -> (* source id *)
      	(opt_len,
	 { priv=priv; discard=action;
	   option=SOURCE(get_address s ofs opt_len) })
    | 2 -> (* dest id *)
      	(opt_len,
	 { priv=priv; discard=action; option=DEST(get_address s ofs opt_len) })
    | 3 -> (* integrity *)
	raise Unimplemented
    | 4 -> (* n/n auth *)
	(opt_len,
	 { priv=priv; discard=action;
	   option=NN_AUTH(get_nn_auth s ofs opt_len) })
    | 5 -> (* auth *)
	(opt_len,
	 {priv=priv; discard=action;
	   option=AUTH(get_auth s (ofs + 4) opt_len payload) })
    | _ -> (* should handle properly *)
      	raise (Invalid_argument "get_options") in
  if len = 0 then []
  else
    let (opt_len, opt) = get_an_option opts ofs in
    opt :: (get_options opts (ofs+(4*opt_len)) (len - opt_len) payload)

let get_code_data pkt hdr_len pkt_len =
  let hdr_bytes = hdr_len * 4 in
  let func_name_len = to_int32 pkt hdr_bytes in
  let code_len = to_int32 pkt (hdr_bytes + 4) in
  let data_len = to_int32 pkt (hdr_bytes + 8) in
  (String.sub pkt (hdr_bytes + 12) func_name_len,
   String.sub pkt (hdr_bytes + 12 + func_name_len) code_len,
   String.sub pkt (hdr_bytes + 12 + func_name_len + code_len) data_len)

let get_payload pkt hdr_len pkt_len =
  String.sub pkt (hdr_len * 4) (pkt_len - (hdr_len * 4))

(* XXX Let SANE through. We really want a list of accepted modules here *)
let trusted_module digres =
  if (digres = "09F3F37CDFA15A358E85AE27193921BCE59AACBD")
  then (Log.log_msg "SANE authorized to pass.\n"; true)
  else (Log.log_msg ("Failed to authorize " ^ digres ^ "\n");
	false)

(* XXX It would probably be better to convert the authorized digests to
 * some other data structure (list?) rather than involving regexp every time.
 *)
let angle_regexp = Str.regexp "<\([0-9A-Fa-f]+\)>"

let code_is_authorized code opt_list =
  let rec find_spi opt_list =
    match opt_list with
      [] -> None
    | h::t ->
	(match h.option with
	  AUTH(HMAC(spi)) -> Some(spi)
	| _ -> find_spi t) in
  let rec find_match digest ath off =
    let newoff = Str.search_forward angle_regexp ath 0 in
    if digest = (Str.matched_group 1 ath) then true
    else find_match digest ath newoff in
  let ccc = sha1init () in
      sha1update ccc code;
  let digres = present_digest (sha1final ccc) in
  match find_spi opt_list with
    None -> trusted_module digres
  | Some(spi) ->
      Mutex.lock spi_recv_hash_m;
      let (c1, c2, cntr, ath) = Hashtbl.find spi_recv_hash spi in
      Mutex.unlock spi_recv_hash_m;
      try
        (*Str.search_forward reeg ath 0; true*)
	find_match digres ath 0
      with Not_found -> (Log.log_msg ("Existing auth is for " ^ ath ^ "\n");
	    		 trusted_module digres)
	  
let burst_wf pkt =
  let version = get_version pkt in
  let (do_forward, _) = get_flags pkt in
  let type_id = get_type_id pkt in
  let hdr_len = get_hdr_len pkt in
  let pkt_len = get_pkt_len pkt in
  let hdr = { do_forward = do_forward; type_id = type_id } in
  let payload = get_payload pkt hdr_len pkt_len in
  let opt_list = get_options pkt 8 (hdr_len - 2) payload in
  match type_id with
    20 ->
      let (name, code, data) = get_code_data pkt hdr_len pkt_len in
      if code_is_authorized code opt_list then
	(hdr, opt_list, name, code, data)
      else
	(Log.log_msg "Attempt to load unauthorized module.\n";
	 raise Auth_failure)
  | _ -> raise Unimplemented
	
let top_thread () =
  uport_val := Some(Udp.listen_udp_port !default_anep_port);
  while not !die do
    let (hdr, len, pkt) = Udp.get_packet (uport()) in
    begin try
      let (hdr, opts, func, code, data) = burst_wf pkt in
      Mutex.lock conn_hash_m;
      begin try
	let (q, m, c) = Hashtbl.find conn_hash hdr.type_id in
	Mutex.unlock conn_hash_m;
	Mutex.lock m;
	Queue.add (hdr, opts, func, code, data) q;
	Condition.signal c;
	Safethread.delay 0.0;
	Mutex.unlock m
      with Not_found -> Mutex.unlock conn_hash_m
      |	e -> (Mutex.unlock conn_hash_m; raise e)
      end
    with Auth_failure -> (Log.log_msg "authorization failed\n";())
    | e -> raise e
    end
  done;
  Udp.unbind_udp (uport());
  uport_val := None

(* This string arg indicates the default port to use, if present *)
let wf arg =
  let port = int_of_string arg in
  if port != 0 then
    default_anep_port := port;
  top_thread ();
  ""

let _ = value := Some(Func.register "wf" wf)
