open Nat
open Dsaprimes

(* Tag definition -- includes everything we may need really *)
type tag = { protoHashType : char ;
             protoHash : string ; (* Protocol Hash, length implied *)
             protoTime : string ; (* Time message was sent *)
             cookie : string ;    (* Cookie *)
             generator : string ; (* XXX DH generator/modulus *)
             modulus : string ;   (* XXX Ignore for now *)
             pubval : string ;    (* Peer DH public value *)
             rID : string  ;      (* Sender's ID *)
             ixID : string ;      (* Initiator XID *)
             rxID : string ;      (* Responder XID *)
             authorization : string }

(* Certificate definition *)
type certificate = { signerType : char ;
                     signer : string ;
                     subjectType : char ;
                     subject : string ;
                     notValidBefore : string ;
                     notValidAfter : string ;
                     delegate : char ;
                     cerTag : tag ;
                     signatureType : char ;
                     signatureR : string ;
                     signatureS : string }

(*
  The certificate file should contain the following fields (each on a
  line by itself):
  
  DSA p value
  DSA q value
  DSA g value
  public key
  private key
  subject key
  not valid before (in seconds)
  not valid after (in seconds)
  delegate (y/n)
  remote ID
 authorization (string)
  *)

let hashTypeSHA = 'a'
let typeDSA = 'a'
    
let filename = Sys.argv.(1)
    
(* Encode a list of certificate in a single string *)
let rec encode_certificates certlist =
    match certlist with
      [] -> ""
    | cert :: tail ->
	let certstr = Printf.sprintf "#signerType=%c;signer=%s;subjectType=%c;subject=%s;notValidBefore=%s;notValidAfter=%s;delegate=%c;cerTag.protoHashType=%c;cerTag.protoHash=%s;cerTag.protoTime=%s;cerTag.cookie=%s;cerTag.generator=%s;cerTag.modulus=%s;cerTag.pubval=%s;cerTag.rID=%s;cerTag.ixID=%s;cerTag.rxID=%s;cerTag.authorization=%s;signatureType=%c;signatureR=%s;signatureS=%s;"
	    cert.signerType cert.signer cert.subjectType cert.subject
	    cert.notValidBefore cert.notValidAfter cert.delegate
	    cert.cerTag.protoHashType cert.cerTag.protoHash
	    cert.cerTag.protoTime cert.cerTag.cookie
	    cert.cerTag.generator cert.cerTag.modulus
	    cert.cerTag.pubval cert.cerTag.rID cert.cerTag.ixID
	    cert.cerTag.rxID cert.cerTag.authorization
	    cert.signatureType cert.signatureR cert.signatureS in
	certstr ^ encode_certificates(tail)

let crypto_verify (r, s) myPublicKey subject notValidBefore notValidAfter
    delegate authorization commonDSAp commonDSAq commonDSAg rid signerkey =
  let message = "SignerType" ^
    string_of_int(Char.code(typeDSA)) ^
    "Signer" ^
    myPublicKey ^
    "SubjectType" ^
    string_of_int(Char.code(typeDSA)) ^
    "Subject" ^
    subject ^
    "NotValidBefore" ^
    notValidBefore ^
    "NotValidAfter" ^
    notValidAfter ^
    "Delegate" ^
    string_of_int(Char.code(delegate)) ^
    "ProtoHashType" ^
    string_of_int(Char.code(hashTypeSHA)) ^
    "ProtoHash" ^
    "" ^
    "ProtoTime" ^
    "" ^
    "Generator" ^
    "" ^
    "Modulus" ^
    "" ^
    "Pubval" ^
    "" ^
    "rID" ^
    rid ^
    "ixID" ^
    "" ^
    "rxID" ^
    "" ^
    "Authorization" ^
    authorization ^
    "SignatureType" ^
    string_of_int(Char.code(typeDSA)) in
  if verify message (r, s) (commonDSAp, commonDSAq, commonDSAg, signerkey)
  then
    Printf.printf "Certificate verified\n"
  else
    Printf.printf "Signature doesn't verify!\n"
      
let crypto_sign myPublicKey subject notValidBefore notValidAfter
    delegate authorization commonDSAp commonDSAq commonDSAg
    myPrivateKey rid signerkey zeroval =
  let message = "SignerType" ^
    string_of_int(Char.code(typeDSA)) ^
    "Signer" ^
    myPublicKey ^
    "SubjectType" ^
    string_of_int(Char.code(typeDSA)) ^
    "Subject" ^
    subject ^
    "NotValidBefore" ^
    notValidBefore ^
    "NotValidAfter" ^
    notValidAfter ^
    "Delegate" ^
    string_of_int(Char.code(delegate)) ^
    "ProtoHashType" ^
    string_of_int(Char.code(hashTypeSHA)) ^
    "ProtoHash" ^
    "" ^
    "ProtoTime" ^
    "" ^
    "Generator" ^
    "" ^
    "Modulus" ^
    "" ^
    "Pubval" ^
    "" ^
    "rID" ^
    rid ^
    "ixID" ^
    "" ^
    "rxID" ^
    "" ^
    "Authorization" ^
    authorization ^
    "SignatureType" ^
    string_of_int(Char.code(typeDSA)) in
  let (r, s) = sign message 
      (commonDSAp, commonDSAq, commonDSAg, signerkey)
      myPrivateKey zeroval in
  (r, s)
    
let rec skip str ptr =
  let len = String.length str in
  let cc = str.[ptr] in
  if cc = '1' || cc = '2' || cc = '3' || cc = '4' ||
  cc = '5' || cc = '6' || cc = '7' || cc = '8' ||
  cc = '9' || cc = '0'
  then
    String.sub str ptr (len - ptr)
  else
    if ptr = (len - 1)
    then
      (Printf.printf "Ooops, something wrong here.\n"; "")
    else
      skip str (ptr + 1)
	
let main () =
  let filedesc = open_in filename in
  let rr = input_line filedesc in
  let commonDSAp = Safeloader.Nat.nat_of_string(skip rr 0) in
  let rr = input_line filedesc in
  let commonDSAq = Safeloader.Nat.nat_of_string(skip rr 0) in
  let rr = input_line filedesc in
  let commonDSAg = Safeloader.Nat.nat_of_string(skip rr 0) in
  let rr = input_line filedesc in
  let myPublicKey = skip rr 0 in
  let signerkey = Safeloader.Nat.nat_of_string myPublicKey in
  let rr = input_line filedesc in
  let myPrivateKey = Safeloader.Nat.nat_of_string(skip rr 0) in
  let subject = input_line filedesc in
  let rr = input_line filedesc in
  let notValidBefore = string_of_int(int_of_string rr) in
  let rr = input_line filedesc in
  let notValidAfter = string_of_int(int_of_string rr) in
  let deleg = input_line filedesc in
  let delegate = deleg.[0] in
  let rid = input_line filedesc in
  let authorization = input_line filedesc in
  let zeroval = Safeloader.Nat.nat_of_int 0 in
  let (r, s) = crypto_sign myPublicKey subject notValidBefore notValidAfter
      delegate authorization commonDSAp commonDSAq
      commonDSAg myPrivateKey rid signerkey zeroval in
  let nada = crypto_verify (r, s) myPublicKey subject notValidBefore
      notValidAfter delegate authorization commonDSAp
      commonDSAq commonDSAg rid signerkey in
  let encoded = encode_certificates [ { signerType = typeDSA;
					signer = myPublicKey;
					subjectType = typeDSA;
					subject = subject;
					notValidBefore = notValidBefore;
					notValidAfter = notValidAfter;
					delegate = delegate;
					cerTag = {
					protoHashType = hashTypeSHA;
					protoHash = "";
					protoTime = "";
					cookie = "";
					generator = "";
					modulus = "";
					pubval = "";
					rID = rid;
					ixID = "";
					rxID = "";
					authorization = authorization;
				      }	;
					signatureType = typeDSA;
					signatureR = Safeloader.Nat.string_of_nat r;
					signatureS = Safeloader.Nat.string_of_nat s;
				    } ] in
  Printf.printf "%s\n" encoded
    
let _ = main ()
