external get_time : unit -> float = "get_time"

(* Functions for calculating useful statistics from data in fp arrays *)
(* average *)
let avg a n =
  let acc = ref 0.0 in
  for i = 0 to n-1 do
    acc := !acc +. a.(i)
  done;
  acc := !acc /. (float n);
  !acc

(* cumulative total *)
let total a n =
  let acc = ref 0.0 in
  for i = 0 to n-1 do
    acc := !acc +. a.(i)
  done;
  !acc

(* median *)
let med a n =
  let l = Array.to_list (Array.sub a 0 n) in
  let sl = Sort.list (fun x y -> x < y) l in
  List.nth sl (n/2)

(* timeit f init n executes function f (a thunk) n times,
   after first calling init.  The average and median time to execute
   the function are printed as a result *)  
let timeit f init n =
  init();
  let a = Array.create n 0.0 in
  for i = 0 to n-1 do
    let start = get_time() in
    f ();
    let finish = get_time() in
    a.(i) <-
       (let raw = finish -. start in
       if raw > 0.0  then raw
       else raw +. 1000000.0) (* rollover *)
    (* Printf.printf "%d: %f seconds\n" i a.(i) *)
  done;
  Printf.printf ">>> average = %f, median = %f\n"
    (avg a n) (med a n);
  flush Pervasives.stdout

(* For use in the running program *)

let array_size = 2000
let num_arrays = 10

(* For associating names with tests *)

let test_names = Array.init num_arrays (fun _ -> String.create 0)
let test_num = ref 0
exception NoneLeft

let register_test name =
  if !test_num = num_arrays then raise NoneLeft
  else
    let num = !test_num in
    incr test_num;
    test_names.(num) <- name;
    num

let name_test name i =
  if i >= num_arrays then failwith "name_test: not that many arrays";
  test_names.(i) <- name

(* One array for each timing test: the array will have array_size elems,
   and there shall be num_arrays arrays. *)
let array_of_arrays = 
  Array.init num_arrays (fun _ -> Array.create array_size 0.0)
let array_of_indexes = Array.init num_arrays (fun _ -> (ref 0))

(* The nth mutex of this array protects the nth entries of the above arrays *)
(*let array_of_mutexes = 
  Array.init num_arrays (fun _ -> Mutex.create())*)

(*let critical m thunk =
  try
    Mutex.lock m;
    let result = thunk() in
    Mutex.unlock m;
    result
  with e -> Mutex.unlock m; raise e*)

(* For reporting the results of a particular test *)
let report_fun = ref (print_endline)

let register_report_fun f =
  report_fun := f

(* Assumes the mutex is held *)
let report idx =
  let arr = array_of_arrays.(idx) in
  let arr_idx = !(array_of_indexes.(idx)) in
  let num_vals = if arr_idx <> 0 then arr_idx else array_size in 
  let (t,a,m) = (total arr num_vals, avg arr num_vals,med arr num_vals) in
  (!report_fun) (Printf.sprintf 
		   ">>> Test %-10s: %d trials, total %f, avg %f, med = %f"
		   test_names.(idx) num_vals t a m)

let dump_samples idx st ed =
  let arr = array_of_arrays.(idx) in
  let arr_idx = !(array_of_indexes.(idx)) in
  let num_vals = if arr_idx <> 0 then arr_idx else array_size in
  let upper = if ed > num_vals then num_vals - 1 else ed in
  let lower = if st < upper then st else upper in
  for i = lower to upper do
    (!report_fun) (Printf.sprintf "%-10s: %d %f" test_names.(idx) i arr.(i))
  done
  
(* Give this call the start time and it will atomically write the
   record into the array *)

(*let record_time idx start_time =
  let finish = get_time() in
  critical (array_of_mutexes.(idx))
    (function _ ->
      let arr_idx = array_of_indexes.(idx) in
      let arr_idx_val = !arr_idx in
      (array_of_arrays.(idx)).(arr_idx_val) <-
	 (let raw = finish -. start_time in
	 if raw >= 0.0  then raw
	 else raw +. 1000000.0); (* rollover *)
      arr_idx := (arr_idx_val + 1) mod array_size;
      if !arr_idx = 0 then report idx)*)

(* Only safe to use these if the programmer will only use one pair of
   timing calls per array (prevents different threads from making the
   same calls at the same time) *)

let start_time idx = 
  (array_of_arrays.(idx)).(!(array_of_indexes.(idx))) <- get_time()

let end_time idx =
  let finish = get_time() in
  let arr_idx = array_of_indexes.(idx) in
  let arr_idx_val = !arr_idx in
  (array_of_arrays.(idx)).(arr_idx_val) <-
     (let raw = finish -. (array_of_arrays.(idx)).(arr_idx_val) in
     raw);
(*     if raw > 0.0  then raw
     else raw +. 1000000.0); (* (rollover *)*)
  arr_idx := (arr_idx_val + 1) mod array_size;
  if !arr_idx = 0 then report idx

let clear_time idx =
  array_of_indexes.(idx) := 0
